#!/usr/bin/env python

# next up: perform genotyping against a diploid genome

SNIPER_VERSION = '1.6.4'


"""sniper.py: SNP Identification using Probability of Every Read. 
 
 Version: %s (unix/mac)
 
 Daniel F. Simola (simola@mail.med.upenn.edu)
 Laboratory of Junhyong Kim
 University of Pennsylvania
 May 2011
 
 Copyright (c) 2011, Daniel F. Simola and Junhyong Kim, University of
 Pennsylvania.  All Rights Reserved.
 
 You may not use this file except in compliance with the terms of our License. 
 You may obtain a copy of the License at 
 http://kim.bio.upenn.edu/software/LICENSE
 
 Unless required by applicable law or agreed to in writing, this
 software is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
 CONDITIONS OF ANY KIND, either express or implied.  See the License
 for the specific language governing permissions and limitations
 under the License."""%(SNIPER_VERSION)



import os, re, sys, math, time, stat, time, random, subprocess
from operator import add,mul
from string import *

# GLOBAL VARIABLES AND FUNCTIONS
# -----------------------------------------------------------------
multichromtoken = ';'      # separator for --rest <chr1.chr2.....chrn> separation
nastr = '"NA"'             # represents missing value
SGr = {}                   # global current read pileup dictionary
URB = {'SE':[] , 'PE':[]}  # unique read buffer
DRB = {'SE':[], 'PE':[]}   # degenerate index read buffer
bufferlimit = 1000         # number of lines to buffer in memory (10MB)


def delete(x):
	"helper function for pileup updating"
	global SGr
	del SGr[x]
	return


def fetchLines(handler,list_of_byte_len):
	"Random access to lines in the read map files"
	for nbytes,nline in sorted(list_of_byte_len):
		handler.seek(nbytes)
		for x in xrange(nline): yield handler.next() # return a full line
		# yield handler.read(nline)


def indexFile(infile, outfile):
	
	"Records number of bytes into file for first row of each chromosome"
	
	if not os.access(infile, os.F_OK): sys.exit('Error: cannot index file %s. Check file access.'%(infile))
	
	dct = {} # keyed by chromosome
	count = 1; nbytes = 0
	fh = open(infile)
	for line in fh:
		ll = len(line)
		if line[0] == '@': nbytes += ll; continue
		
		try:
			lst = line[:-1].split('\t')
			# chrom = lst[2]
			# pos = int(lst[3])-1
			
			if lst[2] not in dct:
				dct[lst[2]] = nbytes
			nbytes += len(line)
		# exception if a line beginning with @
		except IndexError: nbytes += ll; pass
		except ValueError: nbytes += ll; pass
	fh.close()
	
	printTable(sorted(dct.items()), file=outfile)
	
	return 1


def indexDegFile(infile, outbase, filterids={}):
	"Creates a multi-alignment index and a metaindex for this file. Format allows stream-processing"
	
	if not os.access(infile, os.F_OK): sys.exit('Error: cannot index file %s. Check file access.'%(infile))
	pipeit('indexing...')
	# helper function for fast printing
	printme = lambda pipe, astr: lambda tup: pipe.write('%s\t%s\t%s\n'%(tup[0], tup[1], astr))
	
	# output file
	fidx = open('%s.index.unsort'%(infile), 'w')
	
	# temporary storage to summarize info for all deg alignments for a given hitid
	curHitid = None; curPese = None; curInfo = []; curByte = []
	count = 1; nbytes = 0 # update line by line
	
	fh = open(infile)
	for line in fh:
		if count % 1000000 == 0: pipeit('%.1e,'%(count))
		count+=1; ll = len(line)
		
		try:
			lst = line[:-1].split('\t')
			chrom = lst[2]
			if chrom == '*': nbytes += ll; continue
			
			hitid = hash(lst[0]); pos=int(lst[3])-1; seq=lst[9]; qual=lst[10]; rl = len(seq)
			
			# must record all instances of multiply-mapped reads (relevant to bounds)
			# but for each entry store first line for a new readid and number of aligments
			# this allows us to avoid the otherhits dictionary
			
			if curHitid == None:
				# So this is the first read we encounter in the file
				curHitid = hitid # name of the multiply-mapped read
				curPese = pese # store in PE or SE part of dictionary
				curInfo = [(chrom,pos,rl)] # relevant positions for this read
				curByte = [(nbytes,ll)] # minimal info for this read's alignments
				
			elif hitid == curHitid:
				# continuing with multiple alignments of the same read
				curInfo += [(chrom,pos,rl)] # add this alignment
				curByte += [(nbytes,ll)]
				
			else:
				# Then we moved on to a new read. Save previous information
				
				# don't save PCR duplicates
				if curHitid in filterids: continue
				
				# Write a single line with all of this information to each chromosome index file
				# iterate over alignments for this read
				# single str representation of all alignments for this read, wrt this chr,pos
				# ccb = ','.join(map(lambda x: '%s:%s'%(x[0],x[1]), curByte[:maxaln]))
				
				# Since alignments are grouped we should only need to save start and number of next calls
				ccb = '%s\t%s'%(curByte[0][0], len(curByte))
				# ccb = ','.join(map(lambda x: '%s:%s'%(x[0],x[1]), curByte[:maxaln]))
				
				map(printme(fidx,ccb), curInfo)
				
				# for cc,cp,clen in curInfo: # store each alignment, pointing them to the block of bytes
				# 	fidx.write('%s\t%s\t%s\n'%(cc, cp, ccb))
				
				# update
				curHitid = hitid
				curPese = pese
				curInfo = [(chrom,pos,rl)]
				curByte = [(nbytes,ll)]
			
		# exceptions if line begins with @
		except IndexError: pass
		except ValueError: pass
		nbytes += ll
	fh.close()
	fidx.close()
	
	# sort the index files spatially
	pipeit('sorting...')
	os.system('sort -T "%s" -k 1,1 -k 2,2n "%s" > "%s"'%(sorttmpdir, '%s.index.unsort'%(infile), '%s.index'%(infile)))
	os.system('rm -f "%s"'%('%s.index.unsort'%(infile)))
	
	# would be nice to sort within each chromosome only...
	
	# create a meta-index file that stores the starting bytes for each chromosome
	fh = open('%s.index'%(infile))
	dct = {}
	nbytes = 0
	for line in fh:
		chrom,pos = line[:-1].split('\t')[:2]
		if chrom not in dct: dct[chrom] = nbytes
		nbytes += len(line)
	fh.close()
	printTable(sorted(dct.items()), file='%s.metaindex'%(infile))
	return 1


def indexDegFileParts(infile, outbase, filterids={}, bufsize=5000000):
	"Creates a multi-alignment index and a metaindex for this file. Format allows stream-processing"
	
	if not os.access(infile, os.F_OK): sys.exit('Error: cannot index file %s. Check file access.'%(infile))
	
	# output file
	# fidx = open('%s.index.unsort'%(infile), 'w')
	finalindex = '%s.index'%(infile)
	fmetaindex = '%s.metaindex'%(infile)
	
	# -------------------------------------------
	# store temp file names, but clear first if exists
	outdir = infile+'.indices/'
	if os.access(outdir,os.F_OK): os.system('rm -rf "%s"'%(outdir))
	createdir(outdir)
	
	Buff = {}; cbuf = 0
	IH = {}
	fh = open(infile)
	try:
		line = fh.next()
		while line[0] == '@':
			# @SQ	SN:chr1	LN:230208
			if line[0:3] == '@SQ':
				chrom = line[:-1].split('\t')[1].split(':')[1]
				IH[chrom] = outdir+'%s.%s.index'%(outbase,chrom)
				Buff[chrom] = []
			line = fh.next()
	except StopIteration: fh.close(); os.system('rmdir "%s"'%(outdir)); return 1
	fh.close()
	# -------------------------------------------
	
	# temporary storage to summarize info for all deg alignments for a given hitid
	curHitid = None; curPese = None; curInfo = []; curByte = []
	count = 1; nbytes = 0 # update line by line
	
	fh = open(infile)
	for line in fh:
		if count % 1000000 == 0: pipeit('%.1e,'%(count))
		count+=1; ll = len(line)
		
		try:
			lst = line[:-1].split('\t')
			chrom = lst[2]
			if chrom == '*': nbytes += ll; continue
			
			hitid = hash(lst[0]); pos=int(lst[3])-1; seq=lst[9]; qual=lst[10]; rl = len(seq)
			
			# must record all instances of multiply-mapped reads (relevant to bounds)
			# but for each entry store first line for a new readid and number of aligments
			# this allows us to avoid the otherhits dictionary
			
			if curHitid == None:
				# So this is the first read we encounter in the file
				curHitid = hitid # name of the multiply-mapped read
				curPese = pese # store in PE or SE part of dictionary
				curInfo = [(chrom,pos,rl)] # relevant positions for this read
				curByte = [(nbytes,ll)] # minimal info for this read's alignments
				
			elif hitid == curHitid:
				# continuing with multiple alignments of the same read
				curInfo += [(chrom,pos,rl)] # add this alignment
				curByte += [(nbytes,ll)]
				
			else:
				# Then we moved on to a new read. Save previous information
				
				# don't save PCR duplicates
				if curHitid in filterids: continue
				
				# Write a single line with all of this information to each chromosome index file
				# iterate over alignments for this read
				# single str representation of all alignments for this read, wrt this chr,pos
				ccb = map(lambda x: '%s:%s'%(x[0],x[1]), curByte)
				for cc,cp,clen in curInfo: # store each alignment, pointing them to the block of bytes
					
					Buff[cc] += ['%s\t%s\t%s'%(cc,cp,','.join(ccb))]
					cbuf += 1
					# print >> fidx, '%s\t%s\t%s'%(cc,cp,','.join(ccb))
				
				# update
				curHitid = hitid
				curPese = pese
				curInfo = [(chrom,pos,rl)]
				curByte = [(nbytes,ll)]
			
		# exceptions if line begins with @
		except IndexError: pass
		except ValueError: pass
		nbytes += ll
		
		# flush the buffer
		if cbuf >= bufsize:
			for cc in Buff.keys(): 
				if len(Buff[cc]): 
					ftmp = open(IH[cc], 'a')
					for line in Buff[cc]: ftmp.write('%s\n'%(line))
					ftmp.close()
					Buff[cc] = []
			cbuf = 0
		
	fh.close()
	# fidx.close()
	
	# flush once more
	for cc in Buff.keys(): 
		if len(Buff[cc]): 
			ftmp = open(IH[cc], 'a')
			for line in Buff[cc]: ftmp.write('%s\n'%(line))
			ftmp.close()
			Buff[cc] = []
	
	
	# # sort the index files spatially
	# pipeit('sorting...')
	# os.system('sort -k 1,1 -k 2,2n "%s" > "%s"'%('%s.index.unsort'%(infile), '%s.index'%(infile)))
	# os.system('rm -f "%s"'%('%s.index.unsort'%(infile)))
	
	# clear final index
	ftmp = open(finalindex,'w'); ftmp.close()
	
	# sort the index files by position and concatenate to final index file in sorted order
	pipeit('sorting...')
	for chrom in sorted(IH.keys()):
		if os.access(IH[chrom], os.F_OK): os.system('sort -T "%s" -k 1,1 -k 2,2n "%s" >> "%s"'%(sorttmpdir, IH[chrom], finalindex))
	os.system('rm -rf "%s"'%(outdir))
	
	# create a meta-index file that stores the starting bytes for each chromosome
	fh = open(finalindex)
	dct = {}
	nbytes = 0
	for line in fh:
		# try:
		# chrom,pos,alns = line[:-1].split('\t')
		chrom = line[0:line.index('\t')]
		if chrom not in dct: dct[chrom] = nbytes
		# except ValueError: pass
		nbytes += len(line)
	fh.close()
	
	ftmp = open(fmetaindex, 'w')
	for chrom,line in sorted(dct.items()): ftmp.write('%s\t%s\n'%(chrom,line))
	ftmp.close()
	
	# printTable(sorted(dct.items()), file=fmetaindex)
	return 1


def loadDegIndex(idxhandle, chromosomesOfInterest, bounds={}, subsample=None, maxaln=1e10):
	"""Load the degenerate portion of the read map for any reads aligning to this chromosome."""
	
	# convert to dict
	chint = dict( map(lambda x: (x,None), chromosomesOfInterest) )
	
	Bb = {} # set up dictionary based on bounds
	for k in bounds.keys():
		for low,high,flag in filter(lambda x: nastr not in x, sorted(bounds[k])):
			if flag == '+':
				try: Bb[k] += [(low,high)]
				except KeyError: Bb[k] = [(low,high)]
	
	# gather reads of interest wrt this chrom group
	# Since read map is organized by read, really we just need to save
	# information for the first alignment along with line number in the file and number of alignments
	RdxChr = {}
	for chrom in chromosomesOfInterest: RdxChr[chrom] = {}
	
	# now collect all reads with read names of interest for this portion of genome
	count = 1
	try:
		line = idxhandle.next()
		while True:
			count % 1000000 == 0 and pipeit('%.1e,'%(count))
			count +=1 
			chrom,pos,startBytes,nlines = line[:-1].split('\t')
			pos = int(pos); startBytes = int(startBytes); nlines = int(nlines)
			if chrom in chint:
				if len(Bb) and not poverlaps(Bb[chrom], pos): line = idxhandle.next(); continue
				if subsample and random.random() > subsample: line = idxhandle.next(); continue
				
				# load up to specified number of alignments
				parsedAlns = [[startBytes,nlines>maxaln and maxaln or nlines]]
				
				# only store the first maxnumalign alignments
				# trick is that we don't need to store this at all locations...but pileup below should handle it
				try: RdxChr[chrom][pos] += [parsedAlns]
				except KeyError: RdxChr[chrom][pos] = [parsedAlns]
			else: # file is sorted so just return
				return RdxChr
			line = idxhandle.next()
	except StopIteration: return RdxChr
	return RdxChr


def loadPCRduplicates(filename):
	# load the read ids to ignore (due to PCR duplicates)
	badIDs = {}
	try:
		lh = open(filename)
		for line in lh:
			count, badid = line[:-1].split('\t')[:2]
			badIDs[hash(badid)] = None
		lh.close()
	except IOError: pipeit('Warning: Cannot access PCR dup file %s'%(filename),1)
	return badIDs


def reloadBuffer(target, handle, limit):
	try:
		for i in xrange(limit): target += [handle.next()]
	except StopIteration: return -1
	return 1
	
	# try: 
	# 	for i in xrange(limit): 
	# 		# dump += [handle.next()]
	# 		yield handle.next()
	# except StopIteration: pass
	# # return dump
	


fixNull = lambda nastr: lambda x: x=='' and nastr or x

def valueWithinRange(interval, val):
	list(interval).sort()
	if val < interval[0] or val > interval[1]: return False
	else: return True


def poverlaps(tt,p):
	return sum([valueWithinRange(t,p) for t in tt])



def pipeit(s, nl=0, pipe=sys.stdout):
	pipe.write(s+(nl==1 and '\n' or '')); sys.stdout.flush()

def isnan(x):
    return str(x) == str(1e400*0)

def unique(lst):
	"""Returns a subset of lst containing the unique elements."""
	d = {}
	for i in lst:
		if i not in d: d[i] = 0
		else: d[i] += 1
	e = d.keys(); e.sort()
	return e

def maxlenval(tup):
	mlv = 0
	for i in tup:
		x = len(str(i))
		if x>mlv:mlv = x
	return mlv

def fixspacing(item,nsp):
	v = str(item)
	if len(v) == nsp: return v
	delt = nsp-len(v)
	if delt < 0:
		v = v[0:nsp]
	else:
		for i in range(delt): v+=' '
	return v

def transpose(mat):
	"""Given a 2d array of arrays, will invert the dimensions."""
	newm = []
	for c in range(len(mat[0])):
		newrow = []
		for r in range(len(mat)):
			newrow.append(mat[r][c])
		newm.append(newrow)
	return newm

def flatten(sequence):
	def rflat(seq2):
		seq = []
		for entry in seq2:
			if seqin([entry]):
				seq.extend([i for i in entry])
			else:
				seq.append(entry)
		return seq
	
	def seqin(sequence):
		for i in sequence:
			## all sequences have '__contains__' in their dir()
			## parentheses present to aid commenting mid-condition
			if ('__contains__' in dir(i) and type(i) != str and type(i) != dict):
				return True
		return False
	
	seq = [sequence][:] # in case parameter isn't already a sequence
	while seqin(seq): seq = rflat(seq)
	return seq

def createdir(dirname):
	if not dirname: return 0
	if not os.access(dirname, os.F_OK):
		try: os.mkdir(dirname); return 1
		except OSError: pass
	return 0

def record(astr, filename, ioMethod='a'):
	if ioMethod == 'a':
		assert os.access(filename, os.F_OK), 'Cannot access '+filename
	
	pmrfh = open(filename, ioMethod)
	print >>pmrfh, astr
	pmrfh.close()
	
	return 1

def trimComments(dat):
	# parse comments
	ndat = []
	for line in dat:
		# only care about non commented lines
		if not re.match('(^\s*#.*$)|(^\s$)', line):
			# parse any commentary on this line
			cleanme = re.split('^(.*)#.*$', line)
			if len(cleanme) > 1: cleanme.pop(0)
			ndat.append(cleanme[0].strip('\n'))
	return ndat

def getFiles(indirfile, include=[], exclude=[]):
	if type(include)!=type(list()): include = [include]
	if type(exclude)!=type(list()): exclude = [exclude]
	return getContents(indirfile, typ='files', include=include, exclude=exclude)

def getDirectories(indirfile, include=[], exclude=[]):
	if type(include)!=type(list()): include = [include]
	if type(exclude)!=type(list()): exclude = [exclude]
	files = getContents(indirfile, typ='dirs', include=include, exclude=exclude)
	return map(slash, files)

def getContents(indirfile, typ='all', include=[], exclude=[]):
	if type(include)!=type(list()): include = [include]
	if type(exclude)!=type(list()): exclude = [exclude]
	keepfiles = []
	wildcards = ['-', '+', ':', '*', '.']
	if len(include) == 0: include = ['']
	try:
		mode = os.stat(indirfile)[stat.ST_MODE]
		INISADIR = False
		if stat.S_ISDIR(mode):
			INISADIR = True
			indirfile = slash(indirfile)
		files = []
		if INISADIR and os.access(indirfile, os.F_OK):
			for f in os.listdir(indirfile):
				# print indirfile, 'vs', f
				# if f[0] != '.' and f[0] != '~' and f[len(f)-3:len(f)] not in exclude:
				if f[0] != '.' and f[0] != '~':
					# do includes
					for item in include:
						# escape wild cards
						for esc in wildcards:
							item = item.replace(esc, '\\'+esc)
						
						if re.match('.*%s.*'%(item), f):
							# confirm not a directory
							mode = os.stat(indirfile+f)[stat.ST_MODE]
							if typ=='all': 
								if stat.S_ISDIR(mode): files.append(indirfile+f)
								else: files.append(indirfile+f)
							elif typ=='files' and not stat.S_ISDIR(mode):
								files.append(indirfile+f)
							elif typ=='dirs' and stat.S_ISDIR(mode):
								files.append(indirfile+f)
							# else: print indirfile+f, 'is a directory'
			for f in files:
				inexclude = False
				for item in exclude:
					for esc in wildcards:
						item = item.replace(esc, '\\'+esc)
					if re.match('.*%s.*'%(item), f): inexclude = True
				if inexclude == False:
					# confirm not a directory
					keepfiles.append(f)
					
					# try:
					# 	print 'testing', indirfile+f
					# 	mode = os.stat(indirfile+f)[stat.ST_MODE]
					# 	print 'mode', mode
					# except OSError:
					# 	print 'why?'
					# if typ=='all': 
					# 	if stat.S_ISDIR(mode): keepfiles.append(indirfile+f+'/')
					# 	else: keepfiles.append(indirfile+f)
					# elif typ=='files' and not stat.S_ISDIR(mode):
					# 	keepfiles.append(indirfile+f)
					# elif typ=='dirs' and stat.S_ISDIR(mode):
					# 	print 'got a dir'
					# 	keepfiles.append(indirfile+f+'/')
					# else:
					# 	print 'failure', stat.S_ISDIR(mode)
		elif not INISADIR: keepfiles.append(indirfile)
		else: sys.exit('cannot access dir '+indirfile)
	except OSError: pass
	return keepfiles

def printList(lst, title="", delim="\n", pipe=sys.stdout, file='', sort=False, newline='\n', ioMethod='w'):
	if file: pipe = open(file,ioMethod)
	if sort: lst.sort()
	if title: 
		lst = [title]+lst#print >> pipe, title
	for i in range(len(lst)):
		if i == len(lst)-1:
			print >> pipe, str(lst[i]),
		else:
			print >> pipe, str(lst[i])+delim,
	if newline: print >> pipe, newline,
	if file: pipe.close()

def readTable(filename, header=True, rownames=True, delim="\t", comments=True, keepKey=False):
	fh = open(filename)
	table = fh.readlines()
	fh.close()
	if comments: table = trimComments(table)
	
	tabHeader = []; rowIDs = []; data = []
	if header: 
		tabHeader = table.pop(0).strip('\n').split(delim)
		if not keepKey: tabHeader.pop(0)
		# else: tabHeader = table.pop(0).strip('\n').split(delim)
	for line in table:
		sline = line.strip('\n').split(delim)
		if rownames:
			rowIDs.append(sline.pop(0))
		data += [sline]
	return data, rowIDs, tabHeader

def printTable(tab, header=[], rows=[], delim="\t", newline='\n', file='', pipe=sys.stdout, ioMethod='w'):
	
	if file: pipe = open(file,ioMethod)
	
	if header:
		printList(header, delim=delim, pipe=pipe, newline=newline)
	if len(rows):
		assert len(rows) == len(tab), "Rows and tab must have same length."
		for ri in range(len(tab)):
			r = [rows[ri]]+tab[ri]
			printList(r, delim=delim, pipe=pipe, newline=newline)
	else:
		for r in tab:
			printList(r, delim=delim, pipe=pipe, newline=newline)
	
	if file: pipe.close()

def printFormattedTable(tab, header=[], rows=[], delim=" ", newline='\n', file='', pipe=sys.stdout, nastr='"NA"', ioMethod='w', colSpacing=[]):
	if file: pipe = open(file,ioMethod)
	if tab == [] and file: pipe.close(); return
	tab2 = tab[:]
	if rows:
		for i in range(len(tab)): tab2[i] = [rows[i]]+tab[i]
	
	# need to determine for each column the max length value and space to that
	collens = colSpacing[:]
	if not len(collens):
		# print 'printformattedtable: fixing col spacing'
		temp = transpose(tab2)
		if len(header):
			for i in range(len(temp)):
				r = temp[i]
				r.append(header[i])
				collens.append(maxlenval(r))
		else:
			for r in temp: collens.append(maxlenval(r))
	
	# now create a new tuple with proper spacing
	temp = []; sheadings = []
	if len(header):
		for i in range(len(header)):
			sheadings.append(fixspacing(header[i],collens[i]))
	
	for row in tab2:
		newrow = []
		for i in range(len(row)):
			val = row[i]==nastr and '.' or row[i]
			#if val==nastr: val = '.' # SAS missing value
			newrow.append(fixspacing(val,collens[i]))
		temp.append(newrow)
		
	# pipe = open(file,ioMethod)
	printTable(temp,header=sheadings,pipe=pipe,delim=delim,newline=newline)
	if file: pipe.close()

def nts2iupac(nts):
	X = {'A':'A', 'T':'T', 'C':'C', 'G':'G', 'N':'N', 'U':'U', '-':'-'}
	
	X['GT'] = 'K'
	X['AC'] = 'M'
	X['CGT'] = 'B'
	X['ACG'] = 'V'
	X['CG'] = 'S'
	X['AT'] = 'W'
	X['AGT'] = 'D'
	X['CT'] = 'Y'
	X['AG'] = 'R'
	X['ACT'] = 'H'
	
	X['AA'] = 'A'
	X['TT'] = 'T'
	X['CC'] = 'C'
	X['GG'] = 'G'
	
	nts = ''.join(sorted(map(lambda x: x.upper(), nts)))
	return X[nts]
	
	

def iupac2nts(a,double=True):
	X = {'A':'A', 'T':'T', 'C':'C', 'G':'G', 'N':'N', 'U':'U', '-':'-'}
	
	X['K'] = 'GT'
	X['M'] = 'AC'
	X['B'] = 'CGT'
	X['V'] = 'ACG'
	X['S'] = 'CG'
	X['W'] = 'AT'
	X['D'] = 'AGT'
	X['Y'] = 'CT'
	X['R'] = 'AG'
	X['H'] = 'ACT'
	
	if double:
		X['A'] = 'AA'
		X['T'] = 'TT'
		X['C'] = 'CC'
		X['G'] = 'GG'
	
	return X[a.upper()]

def readFasta(faname, usealtid=False, split='^>', idPat='', addins=[], TOUPPER=False, VERBOSE=False):
	"""Parses a multiple fasta file into a dictionary object keyed by fasta identifier. 
	Option to key the dictionary by an alternative id found in the fasta header: altid=True|False.
	If directory of several fasta files is provided as faname, loads them into single dictionary.
	"""
	fadict = {} # return dictionary of fasta entries
	faj = '' # load all information into single data string, faj
	files = getFiles(faname)
	
	for f in files:
		if VERBOSE: print '- Loading %s'%(f)
		fh = open(f)
		fa = fh.readlines()
		fh.close()
		# remove comment lines
		faj += ''.join(filter(lambda x: not re.match(' *#.*', x) and True, fa))
	
	if faj == '': raise IOError
	
	# parse by entry (>)
	getentries = re.compile(split, re.MULTILINE)
	faentries = re.split(getentries, faj)
	
	# for some reason the first entry is the null string
	faentries.pop(0)
	
	# parse individual entries
	for entry in faentries:
		# first  has format >name other-header-info
		(header, seq) = entry.split("\n", 1)
		# trim off the '>' character
		theid = ""
		altid = ""
		info = ""
		# further split up the info - for use of alternative identifier
		pattern = '^(\S+) (.*)$'
		if re.match(pattern, header):
			(crap, theid, info, crap) = re.split(pattern, header)
			info = info.strip(' ').strip('\t')
			if len(info) > 0:
				#print "\""+info+"\""
				(crap, altid, info, crap) = re.split('(\S+)(.*)', info)
				info = info.strip(' ').strip('\t')
		else:
			theid = header
			altid = ""
		
		# remove newlines so the sequence data is contiguous
		seq = re.sub("\n", "", seq)
		# remove terminal whitespace
		seq = seq.strip(' ')
		
		if idPat != '':
			crap, theid, crap = re.split(idPat, theid)
			#print 'theid', theid
		
		for pre,suf,pat in addins:
			if re.match(pat, theid):
				# print 'theid', theid, 'pat', pat
				theid = pre+theid+suf
				# print 'new', theid
		
		if TOUPPER:
			seq = seq.upper()
		
		# build the entry
		# add a little check to see if there are repeat entries
		if fadict.has_key(theid):
			print theid, ": This key has already been used to store another fasta entry!"
		else:
			# build the entry for this id
			# allow choice of primary dict key
			if theid and altid and usealtid:
				#print "OK", theid, altid
				fadict[altid] = {'id':theid, 'altid':altid, 'info': info, 'seq': seq}
			else:
				fadict[theid] = {'id':theid, 'altid':altid, 'info': info, 'seq': seq}
	return fadict

def list2fasta(seqs=[], ids=[]):
	dct = dict()
	uid = 0
	for i in range(len(seqs)):
		theid = str(uid)
		try: theid = ids[i]
		except IndexError: pass
		dct[theid] = {'id':theid, 'seq':seqs[i], 'info':''}
		uid += 1
	return dct

def printFasta(fadict, file='', pipe=sys.stdout, key='seq', usealtid=False):
	"""Print a dictionary of fasta entries to a file in mfasta format"""
	keys = fadict.keys()
	keys.sort()
	
	if file: pipe = open(file,'w')
	
	for theid in keys:
		if 'altid' in fadict[theid] and not usealtid:
			header = '>'+theid+' '+fadict[theid]['altid']+' '+fadict[theid]['info']
		elif 'id' in fadict[theid]:
			header = '>'+theid+' '+fadict[theid]['id']+' '+fadict[theid]['info']
		else:
			header = '>'+theid+' '+fadict[theid]['info']
		print >> pipe, header
		print >> pipe, fadict[theid][key]
		print >> pipe
	
	if file: pipe.close()

def fasta2Phylip(dct={}, file='phylipi.txt', infile='', verbose=False):
	"""Convert a fasta file into Phylip interleaved format."""
	# print infile
	d = dct
	
	if infile: d = readFasta(infile)
	
	names = d.keys()
	# fix length
	maxlen = 10+1
	shorts = ['' for i in range(len(names))]
	for i in range(len(names)): 
		shorts[i] = fixspacing(names[i], maxlen)
	blank = fixspacing('',maxlen)
	taxa = len(names)
	# print d
	# print 'test', d[d.keys()[0]]
	alilen = len(d[d.keys()[0]]['seq'])
	seqs = ['' for i in range(taxa)]
	
	if verbose: print shorts
	if verbose: print taxa, alilen
	if verbose: print map(len, seqs)
	enc = 'ascii'
	for i in range(taxa): seqs[i] = d[names[i]]['seq'].encode(enc)
	
	# print out an interleaved phylip formatted file
	pos = 0
	fh = open(file, 'w')
	# alilen = 715
	fh.write(str(taxa)+' '+str(alilen)+'\n'.encode(enc))
	
	for i in range(taxa):
		fh.write(shorts[i]+seqs[i][pos:pos+maxlen]+' '+seqs[i][pos+maxlen:pos+2*maxlen]+' '+seqs[i][pos+2*maxlen:pos+3*maxlen]+' '+seqs[i][pos+3*maxlen:pos+4*maxlen]+'\n'.encode(enc))
	pos = 2*maxlen
	fh.write('\n'.encode(enc))
	
	while pos < alilen:
		for i in range(taxa):
			fh.write(blank+seqs[i][pos:pos+maxlen]+' '+seqs[i][pos+maxlen:pos+2*maxlen]+' '+seqs[i][pos+2*maxlen:pos+3*maxlen]+' '+seqs[i][pos+3*maxlen:pos+4*maxlen]+'\n'.encode(enc))
		pos += 4*maxlen
		fh.write('\n'.encode(enc))
	
	fh.close()
	
	return 1

def argmax(v):
	"""Return argmax of a list"""
	mi = 0
	mv = v[0]
	for i in range(1,len(v)):
		if float(v[i]) > float(mv):
			mv = float(v[i])
			mi = i
	return mi

def slash(astr):
	if not len(astr): return '/'
	elif astr[-1] !='/': astr+='/'
	elif astr[len(astr)-2:len(astr)] == '//': astr = astr[:-1]
	return astr

def makeContigs(intervals):
	intervals.sort()
	# print intervals[0:100]
	# combine overlaps
	coverage = 0 # total bases mapped
	contigs = []
	curr = 1
	while curr < len(intervals):
		bak = curr
		start = intervals[curr-1][0]; stop = intervals[curr-1][1]
		
		while curr < len(intervals) and intervals[curr][0] <= stop:
			if intervals[curr][1] > stop: stop = intervals[curr][1]
			curr += 1
		
		contigs += [(start,stop)]
		coverage += abs(start-stop)
		curr += 1
	# print 'contigs', contigs
	return {'contigs':contigs, 'coverage':coverage}


eq = lambda s: lambda i: str(s)!=str(i)
filterstr = lambda s: lambda v: filter(eq(s), v)
roundna = lambda e: lambda x: ((x==nastr or e==nastr) and nastr) or round(float(x),int(e))
floatna = lambda x: x == nastr and nastr or float(x)
intna = lambda x: (x == nastr or x == '' or x == 'NA' and nastr) or int(x)

def argmin(v):
	"Return argmin of a list"
	mi = 0
	mv = v[0]
	for i in range(1,len(v)):
		if v[i] < mv:
			mv = v[i]
			mi = i
	return mi


def factit(x):
	"""Iterative implementation of factorial function."""
	if x <= 1: return 1
	a = [0 for i in range(x+1)]
	a[0] = 1; a[1] = 1
	for i in range(2,x+1): a[i] = i*a[i-1]
	return a[x]

nchoosek = lambda n,k: factit(n)/(factit(k)*factit(n-k)) # binomial coefficient
prod = lambda lst: reduce(mul, lst) # product function

# basic statistics
def avg(lst):
	try: return reduce(add, lst)/float(len(lst))
	except TypeError: return nastr

def harmonicMean(v, minval=1e-20):
	if len(v) == 0: return 0
	# prevent 0 division
	v = map(lambda x: x==0 and minval or x, v)
	return len(v) / sum(map(lambda x: 1./x, v))


ssd = lambda lst, mu: reduce(add, map(lambda x: (x-mu)*(x-mu), lst))
def variance(lst):
	try: return divna(ssd(lst,avg(lst)), float((len(lst)-1)))
	except TypeError: return nastr

def sd(lst):
	try: return math.sqrt(variance(lst))
	except TypeError: return nastr

def percentile(v=[], k=0.5):
	"""
	Return the value of array that corresponds to the kth percentile of the data.
	"""
	if len(v) <= 0: return 0.0
	temp = map(float, v)
	temp.sort()
	n = len(temp)
	idx = float(k)*(n+1)
	lidx = math.floor(idx)
	hidx = math.ceil(idx)
	if idx < 1.0:
		return float(temp[0])
	elif idx > n-1:
		return float(temp[n-1])
	elif not (idx == lidx or idx == hidx):
		return float(temp[int(lidx)] + temp[int(hidx)])/2.0
	else:
		return float(temp[int(idx)])

median = lambda v: percentile(v,.5)
getQuants = lambda dat: ','.join([str(q)+':'+str(round(percentile(dat, q),4)) for q in [0.05, 0.25, 0.5, 0.75, 0.95]])

# list operations
head = lambda x: x[0]
tail = lambda x: x[len(x)-1]
car = lambda x: x[0]
cdr = lambda x: x[len(x)-1]

# arithmetic operations with missing values
subna = lambda m, v: ((m==nastr or v==nastr or float(v)==0.0) and nastr) or float(m) - float(v)
multna = lambda m, v: ((m==nastr or v==nastr or float(v)==0.0) and nastr) or float(m) * float(v)
divna = lambda m, v: ((m==nastr or v==nastr or float(v)==0.0) and nastr) or float(m) / float(v)
roundna = lambda e: lambda x: ((x==nastr or e==nastr) and nastr) or round(float(x),int(e))

def makeRange(xmin,xmax,interval=None,n=None):
	if n == None: n = int(abs(xmax-xmin)/float(interval))
	else: 
		interval = (xmax-xmin)/float(n)
		interval += interval/float(n-1)
	return [xmin+interval*i for i in range(n)]

def histogram(dat, nbins=100, bins=[], categorical=False, yfreq=False, logscale=False, logbase=10, shift=0, dtype=None):
	import math
	if logscale: dat = map(lambda x: math.log(x,logbase), dat)
	
	if len(bins) > 0: 
		# print 'problem?'
		nbins = len(bins)
		givenbins = True
	else: givenbins = False
	
	# need to create a histogram
	filt = filterstr(nastr)(dat)
	if dtype: filt = map(dtype, filterstr(nastr)(dat))
	dat = None
	if not len(filt): return [[0,0]]
	
	counts = [0 for i in range(nbins)]
	if not givenbins and not categorical: bins = makeRange(min(filt), max(filt), n=nbins)
	elif not givenbins: bins = unique(filt)
	
	# minor bin correction for plotting
	if shift != 0: bins = map(lambda x: x+shift, bins)
	
	# this is like O(N2) too slow
	if not categorical:
		for j in range(len(filt)):
			k = 0
			# print j, len(filt)
			while k < nbins and float(filt[j]) > bins[k]: 
				k +=1
			if k == nbins: k -= 1
			counts[k] += 1
	else:
		bindct = {}
		for b in bins: bindct[b] = 0
		bk = bindct.keys()
		for j in range(len(filt)): bindct[filt[j]] += 1
		# convert back to a histogram
		bins = []; counts = []
		for key in bindct.keys(): bins += [key]; counts += [bindct[key]]
	
	tmp = []
	intnbins = int(nbins)
	# counts or frequency?
	if yfreq:
		tot = float(sum(counts))
		if logscale: tmp = [[logbase**bins[k],counts[k]/tot] for k in range(intnbins)]
		else: tmp = [[bins[k],counts[k]/tot] for k in range(intnbins)]
	if logscale: tmp = [[logbase**bins[k],counts[k]] for k in range(intnbins)]
	else: tmp = zip(bins,counts)
	
	if categorical: tmp = sorted(tmp)
	
	return tmp


# dictionary keyed by ascii character returning phred q-score
# ascii 33 to 126 - full range
fastqchars = ['!', '"', '#', '$', '%', '&', "'", '(', ')', '*', '+', ',', '-', '.', '/', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9', ':', ';', '<', '=', '>', '?', '@', 'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', '[', '\\', ']', '^', '_', '`', 'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z', '\{', '\|', '\}', '~']
A2Q = dict(zip(fastqchars, range(len(fastqchars))))

# print sorted(A2Q.items())

# string conversion to integer (potentially a long)
encodeNt = lambda x: x=='A'and'1' or x=='T'and'2' or x=='C'and'3' or x=='G'and'4' or x=='N'and'5'
encodeSeq = lambda seq: int(''.join(map(encodeNt, seq)))
decodeNt = lambda x: x=='1'and'A' or x=='2'and'T' or x=='3'and'C' or x=='4'and'G' or x=='5'and'N'
decodeSeq = lambda val: ''.join(map(decodeNt, str(val)))

encodeQ = dict(zip(fastqchars, map(str,range(10,len(fastqchars)+10)))) # every character is 2 integer digits
decodeQ = dict(zip(map(str,range(10,len(fastqchars)+10)), fastqchars))
encodeQual = lambda qual: int(''.join(map(lambda x: encodeQ[x], qual)))
pairUP = lambda qual: map(lambda x: qual[x-1]+qual[x] ,range(1,len(qual),2))
decodeQual = lambda qual: ''.join(map( lambda x: decodeQ[x], pairUP(str(qual)) ))

P2Q = lambda p: -10*math.log(p,10)

qccheck = lambda line: len(line[:-1].split('\t'))==11 and True

def getChromsFromHeader(mapfile):
	chrm = {}
	SEEN = False
	fh = open(mapfile)
	BADFILE = False
	count = 0
	for line in fh:
		if count > 1000000: 
			BADFILE = True; break
		
		if line[0:3]=='@SQ':
			crap,scaf,ct = line[:-1].split('\t')
			scaf = scaf[3:]
			ct = int(ct[3:])
			chrm[scaf] = ct
			SEEN = True
		elif SEEN: break
		count += 1
	fh.close()
	if BADFILE: sys.exit('Error: Provided map file lacks proper sam header: %s'%(mapfile))
		
	return chrm


def sniper(readfiles, GR, outfile='', saveall='', thesnpdir='', tmpdir='', expPhred=30, maxreadlen=150, Lambda=.5, prior='resequence', theta=0.001, verbose=False, mindepth=1, maxdepth=1000, bounds={}, ploidy=2, LIPieces=25):
	"""Bayesian genome SNP calling utilizing multiply aligned reads bounded by k mismatches and d alignments per read."""
	
	USECCODE = False
	try: import snipercore; USECCODE = True; pipeit(' - Sniper using snipercore.so for faster likelihood calculations.',1)
	except ImportError: pipeit(' - NB, Sniper did not find snipercore.so module.',1)
	
	global SGr, URB
	
	totalLoci = 0
	numHypotheses = 0 # count up number of hypotheses we make
	
	
	readmaph = {}
	for k in readfiles.keys(): 
		try: readmaph[k] = open(readfiles[k])
		except IOError: pass
	
	if GR == None:
		pipeit(' - Loading reference genome... ')
		try: GR = readFasta(bfaref, TOUPPER=True)
		except IOError: sys.exit("CANNOT READ REFERENCE FILE %s"%(bfaref))
		flg = sum(map(lambda x: len(x['seq']), GR.values()))
		pipeit('N=%s loci\n'%(flg),1)
	else:
		# count total genomic positions
		flg = sum(map(lambda x: len(x['seq']), GR.values()))
	
	# store length of each chromosome/scaffold
	GRlengths = dict( map(lambda x: (x, len(GR[x]['seq'])), GR.iterkeys()) )
	
	
	# Internal methods
	# ---------------------------------------------------------------
	ln = lambda x: math.log(x, math.e)
	log = lambda x: math.log(x, 10)
	lowestQ = int(-10*log(1e-16))
	
	# convert between phred score and p-value
	Q2P = lambda e: (e == 0 and 1-1e-16) or 10**(-e/10.) # avoid 1
	P2Q = lambda p: -10*log(p)
	
	A2P = dict( [ascii, 1-Q2P(A2Q[ascii])] for ascii in A2Q.keys() ) # Probability of no error
	invA2P = dict( [ascii, Q2P(A2Q[ascii])] for ascii in A2Q.keys() ) # Probability of error
	
	# Computes mismatches between 2 sequences
	derror = lambda a,b: (a == b and 0) or sum(map(lambda x,y: x != y and 1 or 0, a,b))
	
	def valueWithinRange(interval, val):
		list(interval).sort()
		if val < interval[0] or val > interval[1]: return False
		else: return True
	
	
	def intervalWithinRange(baseint, queryint):
		"""Returns 0,1,2,3 according to these rules:
		0: queryint is not within base int
		1: queryint is within base int
		2: queryint spans the beginning of baseint
		3: queryint spans the end of baseint
		baseint/queryint need not already be sorted
		ie, [1000, 2000] and [2000,1000] will return same value
		"""
		# assumes sorted coordinates - slightly faster
		# baseint = sorted(baseint); queryint = sorted(queryint)
		
		# if the end of the element comes before the start of the base
		if (queryint[1] < baseint[0] or baseint[1] < queryint[0]):
			return 0
		# if element lies inside the base region
		elif (queryint[0] >= baseint[0] and queryint[1] <= baseint[1]):
			return 1
		# if we span a region
		elif (queryint[0] < baseint[0] and queryint[1] >= baseint[0]):
			return 2
		else:
			return 3
		
	
	
	def overlaps(tt,q):
		return sum([intervalWithinRange(t,q) for t in tt])
	
	
	def poverlaps(tt,p):
		return sum([valueWithinRange(t,p) for t in tt])
	
	
	def loadUniqueReads(tmpfile, fromChromosomes, filterids={}, bounds={}):
		# initialize D with chrom names
		D = dict( map(lambda x: (x,{}), fromChromosomes) )
		
		Bb = {} # set up dictionary based on bounds
		for k in bounds.keys():
			for low,high,flag in sorted(bounds[k]):
				if flag == '+' and low!=nastr and high!=nastr:
					try: Bb[k] += [(low,high)]
					except KeyError: Bb[k] = [(low,high)]
					
		if not os.access(tmpfile, os.F_OK): 
			pipeit('Warning: Cannot access unique file %s'%(tmpfile))
			return D
		
		fh = open(tmpfile)
		count = 1; nbytes = 0
		for line in fh:
			if count % 1000000 == 0: pipeit('%s,'%(count))
			count+=1; ll = len(line)
			
			try:
				lst = line[:-1].split('\t')
				chrom = lst[2]
				pos = int(lst[3])-1
				
				if chrom == '*' or (chrom not in D) or lst[0] in filterids: nbytes += ll; continue
				elif chrom in Bb and overlaps(Bb[chrom], (pos,pos+len(lst[10]))) == 0: nbytes += ll; continue
				
				# store bytes in file to this line
				if pos in D[chrom]: D[chrom][pos] += [(nbytes,ll)]
				else: D[chrom][pos] = [(nbytes,ll)]
				nbytes += ll
				
			# exception if a line beginning with @
			except IndexError: nbytes += ll; pass
			except ValueError: nbytes += ll; pass
		fh.close()
		return D # map(lambda pos: pos in D and D[pos] or [], range(totallen))
	
	
	def loadDegReads(readfiles, chromosomesOfInterest, bounds={}, subsample=None):
		"""Load the degenerate portion of the read map for any reads aligning to this chromosome."""
		
		# convert to dict
		chint = dict( map(lambda x: (x,None), chromosomesOfInterest) )
		
		Bb = {} # set up dictionary based on bounds
		for k in bounds.keys():
			for low,high,flag in sorted(bounds[k]):
				if flag == '+':
					try: Bb[k] += [(low,high)]
					except KeyError: Bb[k] = [(low,high)]
					
		
		# gather reads of interest wrt this chrom group
		# Since read map is organized by read, really we just need to save
		# information for the first alignment along with line number in the file and number of alignments
		RdxChr = {'PE':{}, 'SE':{}}
		# chrKeep = {}
		for chrom in GR.keys(): # needs all chrom keys
			# chrKeep[chrom] = None
			RdxChr['PE'][chrom] = {}
			RdxChr['SE'][chrom] = {}
			
		
		# now collect all reads with read names of interest for this portion of genome
		for pese in ['PE', 'SE']:
			degfile = readfiles[pese+'deg']
			if os.access(degfile, os.F_OK):
				# temporary storage to summarize info for all deg alignments for a given hitid
				curHitid = None; curPese = None; curInfo = []; curByte = []
				# curByteStart = 0; curNBytes = 0
				
				count = 1; nbytes = 0 # update line by line
				fh = open(degfile)
				for line in fh:
					if count % 1000000 == 0: pipeit('%.1e,'%(count))
					count+=1; ll = len(line)
					
					# allow coverage subsampling
					if subsample and random.random() > subsample: nbytes += ll; continue
					
					try:
						lst = line[:-1].split('\t')
						chrom = lst[2]
						if chrom == '*': nbytes += ll; continue
						
						hitid = hash(lst[0]); pos=int(lst[3])-1; seq=lst[9]; qual=lst[10]; rl = len(seq)
						
						# must record all instances of multiply-mapped reads (relevant to bounds)
						# but for each entry store first line for a new readid and number of aligments
						# this allows us to avoid the otherhits dictionary
						
						if curHitid == None:
							# So this is the first read we encounter in the file
							curHitid = hitid # name of the multiply-mapped read
							curPese = pese # store in PE or SE part of dictionary
							curInfo = [(chrom,pos,rl)] # relevant positions for this read
							curByte = [(nbytes,ll)] # minimal info for this read's alignments
							
						elif hitid == curHitid:
							# continuing with multiple alignments of the same read
							curInfo += [(chrom,pos,rl)] # add this alignment
							curByte += [(nbytes,ll)]
							
						else:
							# then we moved on to a new read. Save previous information
							# to the non-unique read pileup... if it is relevant
							
							# Currently, only want alignments for reads that occur on chromosomes of interest
							# so if none of these alignments meets this criterion, toss them all
							meetsCriterion = False
							for cc,cp,clen in curInfo:
								if cc in chint or (cc in Bb and overlaps(Bb[cc], (cp,cp+clen)) > 0):
									meetsCriterion = True; break
									
							if meetsCriterion:
								# to recover all of the alignments for this read
								# we only need the starting byte number and total bytes to read
								# stuff = (curByte[0][0], sum([curByte[i][1] for i in xrange(1,len(curByte))])) # would be nice if this worked...
								
								# print 'storing for %s pos'%(len(curInfo)), curByte
								# print fetchLines(open(degfile), curByte)
								
								# only store first maxnumalign alignments
								for cc,cp,clen in curInfo: # store each alignment, pointing them to the block of bytes
									if cp in RdxChr[curPese][cc]: RdxChr[curPese][cc][cp] += [curByte[:maxnumalign]]
									else: RdxChr[curPese][cc][cp] = [curByte[:maxnumalign]]
									
							# update
							curHitid = hitid
							curPese = pese
							curInfo = [(chrom,pos,rl)]
							curByte = [(nbytes,ll)]
							
					# exceptions if line begins with @
					except IndexError: pass
					except ValueError: pass
					nbytes += ll
				fh.close()
			else: pipeit('\nWarning: cannot access degenerate file %s'%(degfile))
		return RdxChr
	
	
	
	# Initialize local variables
	# =========================================================================
	if prior == 'haploid': ploidy = 1
	if ploidy == 1: prior = 'haploid'
	
	# # if reference is diploid then we must change prior
	# diploidreference = False
	# if prior != 'resequence-diploid':
	# 	pipeit('- Examining reference for homo/heterozygosity: ')
	# 	iupacs = {'K':1,'M':1,'B':1,'V':1,'S':1,'W':1,'D':1,'Y':1,'R':1,'H':1}
	# 	for x in GR.iterkeys():
	# 		for nt in GR[x]['seq']: 
	# 			if nt in iupacs: 
	# 				diploidreference = True
	# 				break
	# 	if diploidreference == True: prior = 'resequence-diploid'
	# 	pipeit(diploidreference==True and 'heterozygous' or 'homozygous',1)
	# else:
	# 	diploidreference = True
	
	nucleotides = ['A', 'T', 'C', 'G']
	iupacs = ['K','M','B','V','S','W','D','Y','R','H'] # 2 and 3 alleles
	iupac2s = ['K','M','S','W','Y','R'] # 2 alleles only
	genotypes = None
	genoUni = None
	if ploidy == 2:
		# 10 possible genotypes that we can identify
		genotypes = [('A','A'),('A','T'),('A','C'),('A','G'),('T','T'),('T','C'),('T','G'),('C','C'),('C','G'),('G','G')]
		genoUni = {('A','A'):'AA',('A','T'):'AT',('A','C'):'AC',('A','G'):'AG',
				   ('T','T'):'TT',('T','C'):'CT',('T','G'):'GT',('C','C'):'CC',
				   ('C','G'):'CG',('G','G'):'GG'}
	else:
		genotypes = ['A', 'T', 'C', 'G']
		genoUni = {'A':'A', 'T':'T', 'C':'C', 'G':'G'}
		
	rangelengenotypes = range(len(genotypes))
	ishomozygous = [1,0,0,0, 1,0,0, 1,0, 1]
	
	
	GT = genotypes # alias
	invGT = None
	if ploidy == 2:
		# given genotype return index in genotypes list
		invGT = dict( map(lambda i: ((GT[i][0],GT[i][1]),i) , rangelengenotypes) )
		invGT2 = dict( map(lambda i: ((GT[i][1],GT[i][0]),i) , rangelengenotypes) )
		invGT.update(invGT2)
		# include heterozygous comparisons
		
		
	refhomozygous = lambda nt: nt=='A' or nt=='T' or nt=='C' or nt=='G' and True or False
	# =========================================================================
	
	
	# Parameters
	# =========================================================================
	
	# P(G_xy(l)) = pr. drawing genome at position l with genotype xy
	# --------------------------------------------------------------
	# p = log(1/10.)
	# PG = [p for i in range(lengenotypes)] # uniform prior
	
	# for homozygous reference genome we have the following event outcomes
	# h0 = homozygous reference, t1 = het, t2 = homozygous derived, h2=heterozygous derived
	
	# for diploid (possibly heterozygous) reference genome we have the following outcomes
	# hh0 = homozygous reference (no change); e.g., AA->AA
	# tt0 = heterozygous reference (no change); e.g., AG->AG
	# ht1 = gain in heterozygosity (1 change); e.g., AA->AG,AC,AT
	# ht2 = gain in heterozygosity (2 changes); e.g., AA->CT
	# th1 = loss of heterozygosity (1 change); e.g., AG->AA,GG
	# th2 = loss of heterozygosity (2 changes); e.g., AG->CC,TT
	# hh1 = not valid
	# tt1 = change in heterozygosity (1 change); e.g., AG->AC,AT,CG,TG
	# tt2 = change in heterozygosity (2 changes); e.g., AG->CT
	# hh2 = change in homozygosity (2 changes); e.g., AA->CC
	
	# for simplicity we group these cases to minimize parameters
	# assume P(hh0)=P(tt0), P(ht1)=P(th1), P(hh2), P(ht2)=P(th2)=P(tt2)
	
	# with respect to a homozygous reference, these outcomes reduce to:
	# hh0 => h0, ht1 => t1, hh2 => h2, ht2 => t2
	
	
	if prior == 'mutational':
		# reference genome is diploid (potentially heterozygous)
		#                 P(hh0)=P(tt0),   P(ht1)=P(th1), P(hh2),  P(ht2)=P(th2)=P(tt2)
		PGvals = [1-theta-0.1*theta-theta**2, theta,    0.1*theta, theta**2]
	
	elif prior == 'resequence':
		# reference genome is diploid (potentially heterozygous)
		#                 P(hh0)=P(tt0),   P(ht1)=P(th1), P(hh2),  P(ht2)=P(th2)=P(tt2)
		PGvals = [1-theta-theta**2, theta/2., theta/2., theta**2]
		
		# OLD
		# # heterozygous different is rarest
		# # h0, t1, h2, t2, where P(h2) = P(t1) and P(t2) == P(t1)^2
		# PGvals = [1-theta-theta**2, theta/2., theta/2., theta**2]
		
	elif prior == 'resequence-het':
		# h0, t1, h2, t2, where P(h2) = P(t2)
		PGvals = [1-theta-theta**2, theta, (theta**2)/2., (theta**2)/2.]
	
	elif prior == 'uniform':
		# h0, t1, h2, t2
		PGvals = [1.-(3*theta), theta, theta, theta]
	
	elif prior == 'maq':
		# Durbin/maq model: Pr{XX} = Pr{YY} > Pr{XY}
		# h0, t1, h2, t2
		PGvals = [(1-theta-theta**2)/2., theta, (1-theta-theta**2)/2., theta**2]
	
	elif prior == 'naive':
		PGvals = [.25,.25,.25,.25]
	
	elif prior == 'haploid':
		# h0, t1, h2, t2, where P(t1) and P(t2) = 0
		PGvals = [1-theta, theta]
	
	else:
		sys.exit('snpper: Do not understand prior argument: %s'%(prior))
	
	
	# OLD: homozygous reference
	# def PGlamb(i,ref):
	# 	if genotypes[i][0]==ref and genotypes[i][1]==ref: return PGvals[0]
	# 	elif (genotypes[i][0]==ref and genotypes[i][1]!=ref) or (genotypes[i][0]!=ref and genotypes[i][1]==ref): return PGvals[1]
	# 	elif genotypes[i][0]==genotypes[i][1] and genotypes[i][0]!=ref: return PGvals[2]
	# 	elif genotypes[i][0]!=genotypes[i][1] and genotypes[i][0]!=ref and genotypes[i][1]!=ref: return PGvals[3]
	# 
	
	# heterozygous reference
	# if ploidy == 2 and diploidreference == True:
	def PGlamb(i,ref):
		# first determine edit distance (0,1,2)
		ham = derror(genotypes[i],iupac2nts(ref))
		if ham == 0: return PGvals[0]
		elif ham == 1: return PGvals[1]
		elif ham == 2:
			# determine if hh2
			if refhomozygous(ref) and ishomozygous[i]: 
				# print 'we have hh2', ref, genotypes[i]
				return PGvals[2]
			else: 
				# print 'we have th2, tt2, th2', ref, genotypes[i]
				return PGvals[3]
		else:
			sys.exit('Sniper.py::PGlamb: sequence edit distance is odd: %s-%s(%s)=%s'%(genotypes[i],ref,iupac2nts(ref),ham))
	
	
	if ploidy == 1:
		# maps a genotype given reference to 0,1 = same as ref, or different
		def PGlamb(i,ref):
			if genotypes[i]==ref: return PGvals[0]
			else: return PGvals[1]
		
	
	# make into a lookup dictionary
	PG = {}
	for i in rangelengenotypes:
		for nt in nucleotides+iupac2s: PG[i,nt] = PGlamb(i,nt)
	
	thetastr = ', Theta = %s'%(theta)
	if prior == 'uniform': thetastr = ''
	print ' - Using global prior probabilities: Prior model = %s%s'%(prior, thetastr)
	# =========================================================================
	
	
	# Precompute nt probability scores
	# =========================================================================
	# PriS = pr. sequencing read ri from template sequence S
	# --------------------------------------------------------------
	maxphred = 40
	# generalize this for all read lengths - 0 to 200
	binCoef = [[nchoosek(rl,d) for d in range(rl+1)] for rl in xrange(maxreadlen+1)]
	# call: binomialP(readlength)(mismatches, Perror)
	# tune between read/position-specific error vs a global/position specific error
	binomialP = lambda rl: lambda d,p: binCoef[rl][d] * p**d * (1-p)**(rl-d)
	expError = 10**(-expPhred/10.)
	PriSbin = [[Lambda * binomialP(rl)(d, expError) for d in range(rl+1)] for rl in xrange(maxreadlen+1)]
	
	# index by [rl][offset]
	offsetRange = [map(lambda offset: range(offset)+range(offset+1, rl), xrange(maxreadlen+1)) for rl in xrange(maxreadlen+1)]
	
	ntPriS = None
	opLambda = 1. - Lambda
	
	
	# for heterozygous matching make a direct comparison dictionary 
	# of each nucleotide against each possible iupac reference character
	# this is the fastest way to do a True/False comparison
	# for comp[A][B], 1 := A is subset of B; 0 := A not in B
	comp = {}
	comp['A'] = {'A':1, 'T':0, 'U':0, 'C':0, 'G':0, 'K':0, 'M':1, 'B':0, 'V':1, 'S':0, 'W':1, 'D':1, 'Y':0, 'R':1, 'H':1}
	comp['T'] = {'A':0, 'T':1, 'U':1, 'C':0, 'G':0, 'K':1, 'M':0, 'B':1, 'V':0, 'S':0, 'W':1, 'D':1, 'Y':1, 'R':0, 'H':1}
	comp['U'] = {'A':0, 'T':1, 'U':1, 'C':0, 'G':0, 'K':1, 'M':0, 'B':1, 'V':0, 'S':0, 'W':1, 'D':1, 'Y':1, 'R':0, 'H':1}
	comp['C'] = {'A':0, 'T':0, 'U':0, 'C':1, 'G':0, 'K':0, 'M':1, 'B':1, 'V':1, 'S':1, 'W':0, 'D':0, 'Y':1, 'R':0, 'H':1}
	comp['G'] = {'A':0, 'T':0, 'U':0, 'C':0, 'G':1, 'K':1, 'M':0, 'B':1, 'V':1, 'S':1, 'W':0, 'D':1, 'Y':0, 'R':1, 'H':0}
	comp['N'] = {'A':0, 'T':0, 'U':0, 'C':0, 'G':0, 'K':0, 'M':0, 'B':0, 'V':0, 'S':0, 'W':0, 'D':0, 'Y':0, 'R':0, 'H':0}
	comp['-'] = {'A':0, 'T':0, 'U':0, 'C':0, 'G':0, 'K':0, 'M':0, 'B':0, 'V':0, 'S':0, 'W':0, 'D':0, 'Y':0, 'R':0, 'H':0}
	
	
	# binomial + quality model
	if Lambda < 1.0:
		print ' - Using tuned binomial+quality likelihood model with lambda=%s, global error=%s'%(Lambda, expPhred)
		
		def ntPriS(S, R, Q, offset, rl):
			# localize variables for speed
			Pb = PriSbin
			lA2P = A2P
			linvA2P = invA2P
			nucs = nucleotides
			
			oR = offsetRange
			d = 0 # number of mismatches in the alignment
			p = 1 # read/position specific likelihood
			# avoid the offset position
			# the best
			for i in oR[rl][offset]:
				# heterozygous comparison (R[i] may be IUPAC) using custom function
				p *= (comp[S[i]][R[i]] and lA2P[Q[i]]) or linvA2P[Q[i]]
				# original code
				# p *= (S[i] == R[i] and lA2P[Q[i]]) or linvA2P[Q[i]]
				d += int(S[i] != R[i])
			
			# C code
			# p,d = maths.prishelper(S,R,Q,offset,rl)
			
			return dict( map(lambda nt: (nt, Pb[rl][d+int(S[offset]!=nt)] + opLambda*p*(S[offset]==nt and lA2P[Q[offset]] or linvA2P[Q[offset]])), nucs) )
		
		
	else:
		print ' - Using binomial only likelihood model with global error=%s'%(expPhred)
		# binomial only model when lambda = 1
		# this will be faster
		
		def ntPriS(S, R, Q, offset, rl):
			# localize variables for speed
			oR = offsetRange
			Pb = PriSbin
			nucs = nucleotides
			
			# original
			# d = reduce(lambda d,i: int(S[i]!=R[i]) + d, oR[rl][offset], 0)
			
			# heterozygous comparison
			d = reduce(lambda d,i: int(1-comp[S[i]][R[i]]) + d, oR[rl][offset], 0)
			
			# d = 0 # number of mismatches in the alignment
			# # avoid the offset position
			# 
			# # for i in range(offset)+range(offset+1,rl):
			# for i in oR[rl][offset]:
			# 	d += int(S[i] != R[i])
			
			# complete with offset position
			return dict( map(lambda nt: (nt, Pb[rl][d+int(S[offset]!=nt)]), nucs) )
		
		
	# switches between C code and Python code
	compLik = lambda S,R,Q,O,RL,L,E: USECCODE and snipercore.ntPriS(S,R,Q,O,RL,L,E) or ntPriS(S,R,Q,O,RL)
	
	# DONE CHOOSING METHOD
	if USECCODE: crap = snipercore.initprefact()
	
	# PriSj = pr. that Sj was the template out of all SGri templates
	# --------------------------------------------------------------
	maximumcoverage = 100001 # this will crash if any locus has greater read depth
	PriSj = [nastr]+map(lambda n: 1/float(n), range(1,maximumcoverage)) # uniform, precomputed
	
	# indicator variable that returns 1 if genotype is homozygous and 2 if heterozygous
	# used because we are only considering the 10 unique genotypes
	I2xHet = {}
	if ploidy == 1:
		for X in genotypes: 
			I2xHet[X] = 1.
	else:
		for X,Y in genotypes:
			if X==Y: I2xHet[(X,Y)] = 1.
			else: I2xHet[(X,Y)] = 2.
	
	
	# control plot for phred to Pvalue transformation
	# lst = [1-1/10**(x/10.) for x in range(34)
	# plot.scatter(zip(range(34),lst), style='linespoints', xlabel='phred', ylabel='1-error rate')
	
	# Set up bounds for subset of genome to process
	# =========================================================================
	GRNA = {}
	
	if len(bounds.keys()):
		flg = 0
		pipeit(' - Applying genotyping bounds...',1)
		null = [nastr, 'NA']
		for tmpk in GR.keys():
			master = GR[tmpk]['seq'][:]
			SEEN = False
			for k in [tmpk, tmpk.lower(), tmpk.upper()]:
				if k in bounds and not SEEN:
					SEEN = True
					lastpos = 0
					# may be many positions of interest
					for low,high,flag in sorted(bounds[k]):
						if flag == '+':
							# remove positions not specified in a range
							if low not in null and high not in null:
								# we have a range
								tmps = master
								tmp = ''.join(map(lambda x: 'N', tmps[0:low]))
								tmp += tmps[low:high+1]
								tmp += ''.join(map(lambda x: 'N', tmps[high+1:len(tmps)]))
								master = tmp
							elif low not in null:
								# we have a single position of interest
								# since we've sorted these positions, 
								
								# erase between lastpos and low
								tmp = master[0:lastpos+1]
								tmp += ''.join(map(lambda x: 'N', master[lastpos+1:low]))
								tmp += master[low]
								tmp += master[low+1:len(master)]
								master = tmp
								lastpos = low
							# else just leave the entire chromosome intact
						elif flag == '-':
							# set ranges to NA
							if low not in null and high not in null:
								tmps = master
								tmp = tmps[0:low]
								tmp += ''.join(map(lambda x: 'N', tmps[low:high+1]))
								tmp += tmps[high+1:len(tmps)]
								master = tmp
								
					# now I can finish erasing the end
					if lastpos:
						master = master[0:lastpos] + ''.join(map(lambda x: 'N', master[lastpos:len(master)]))
					
					# write back
					GRNA[tmpk] = {'id':tmpk, 'seq':master}
					flg += len(master)
	else: GRNA = GR
	
	# Initialize variables for likelihood computation
	# =========================================================================
	outfh = None
	if outfile: outfh = open(outfile, 'w')
	alloutfh = None
	if saveall: alloutfh = open(saveall, 'w')
	
	# blank template list for likelihood calculations below (copy from this)
	LLg0 = [0 for i in genotypes] # if doing likelihood sums
	
	# now proceed to genotype each reference genomic position
	chromosomesOfInterest = sorted(GRNA.keys())
	nchroms = len(chromosomesOfInterest)
	
	# split genome into min(LIPieces,|chromosomes|) loading groups
	chromgroups = []
	copydir = chromosomesOfInterest[:]
	if LIPieces > nchroms: LIPieces = max(nchroms,1)
	itemspergroup = math.ceil(nchroms / float(LIPieces))
	
	while len(copydir):
		currgrp = []
		while len(currgrp) < itemspergroup and len(copydir):
			currgrp += [copydir.pop(0)]
		chromgroups += [currgrp]
		
	globRu = {'PE':{}, 'SE':{}}
	globRd = {}
	
	# Open pointer to sorted unique files
	# reopen for each chromosome because sam sorted order might differ from scan order
	readhandle = {}
	if len(readfiles['SEuni']) and 'SEuni' in readmaph: readhandle['SE'] = readmaph['SEuni']
	if len(readfiles['PEuni']) and 'PEuni' in readmaph: readhandle['PE'] = readmaph['PEuni']
	RHK = readhandle.keys()
	
	# Load the index for this file
	mapindex = {}
	for pese in RHK:
		mapindex[pese] = dict(readTable(readfiles[pese+'uni']+'.index', header=0, rownames=0)[0])
		for k in mapindex[pese].keys(): mapindex[pese][k] = int(mapindex[pese][k])
	
	
	# Open pointer to sorted deg index
	dreadhandle = {}
	if len(readfiles['SEdeg']) and 'SEdeg' in readmaph: 
		if not PESECOMBO: dreadhandle['SE'] = open(btmapdir+sampleID+'.SE.degenerate.map.index')
		else: dreadhandle['SE'] = open(btmapdir+sampleID+'.degenerate.map.index')
	if len(readfiles['PEdeg']) and 'PEdeg' in readmaph: 
		if not PESECOMBO: dreadhandle['PE'] = open(btmapdir+sampleID+'.PE.degenerate.map.index')
		# else: dreadhandle['PE'] = None
	dRHK = dreadhandle.keys()
	
	
	USEDEGINDEX = False; degmapindex = {}
	if os.access(readfiles['SEdeg']+'.metaindex', os.F_OK) or os.access(readfiles['PEdeg']+'.metaindex', os.F_OK):
		USEDEGINDEX = True
		# Load the deg index for this file
		for pese in dRHK:
			if not os.access(readfiles[pese+'deg']+'.metaindex', os.F_OK): continue
			degmapindex[pese] = dict(readTable(readfiles[pese+'deg']+'.metaindex', header=0, rownames=0)[0])
			for k in degmapindex[pese].keys(): degmapindex[pese][k] = int(degmapindex[pese][k])
	
	
	lipct = 0
	for chromosomesOfInterest in chromgroups:
		lipct += 1
		
		# load the relevant reads for this chromosome group
		# -------------------------------------------------
		PESEs = ['SE', 'PE']
		if not USESINGLETONS: PESEs = ['PE']
		
		# unique reads
		# for pese in PESEs:
		# 	pipeit(' - Indexing unique %s reads | loading group %s/%s... '%(pese, lipct, LIPieces))
		# 	globRu[pese] = loadUniqueReads(readfiles[pese+'uni'], chromosomesOfInterest, bounds=bounds)
		# 	pipeit('Done.',1)
		
		globRd = None
		if not USEDEGINDEX:
			# index degenerate reads
			pipeit(' - Indexing degenerate map | loading group %s/%s...'%(lipct, LIPieces))
			globRd = loadDegReads(readfiles, chromosomesOfInterest, bounds=bounds)
			pipeit('Done.',1)
		
		# placeholder if we want to filter PCR duplicates and such
		filterids = {}
		
		for chrom in chromosomesOfInterest:
			# reset variables per chromosome
			GRseq = GR[chrom]['seq'] # want actual sequence for proper derror calculations
			chromlen = len(GRseq)
			
			# print status update every 10% per chromosome/scaffold
			printcutoff = chromlen > 10000 and int(chromlen/10.) or 10000
			
			# clear read pileup and marked list every new chromosome
			SGr.clear() # stores read pileup - this is a global variable
			prevGi = -maxreadlen # remember the last position Gi we visited
			
			REACHEDENDOFFILE = {'SE':False, 'PE':False}
			dREACHEDENDOFFILE = {'SE':False, 'PE':False}
			
			# set file pointers at first row for this chromosome
			for pese in RHK:
				try: readhandle[pese].seek(mapindex[pese][chrom])
				except KeyError: REACHEDENDOFFILE[pese] = True
			
			for pese in dRHK:
				try: dreadhandle[pese].seek(degmapindex[pese][chrom])
				except KeyError: dREACHEDENDOFFILE[pese] = True
			
			# if we have a deg index file, we don't need to worry about LIPieces, just load what we need per chromosome
			# using the meta index, load the relevant index for the chromosome group
			if USEDEGINDEX and not DEGREADSTREAM:
				pipeit(' - Loading degenerate index for %s...'%(chrom))
				globRd = {}
				for pese in dRHK:
					globRd[pese] = loadDegIndex(dreadhandle[pese], [chrom], bounds=bounds, maxaln=maxnumalign)
				pipeit('done.',1)
				
			
			# save the previous line read
			curLine = {'SE':[0,0,0,0], 'PE':[0,0,0,0]}
			dcurLine = {'SE':[0,0,0,0], 'PE':[0,0,0,0]}
			
			# ITERATE OVER REFERENCE GENOMIC POSITIONS
			# ----------------------------------------
			for Gi in xrange(chromlen):
				if GRNA[chrom]['seq'][Gi] == 'N': continue
				
				ref = GRseq[Gi]
				totalLoci += 1
				
				Gi%printcutoff==0 and \
					pipeit('   %s %s/%s = %s (Hypos=%s | %.2f'%(chrom, Gi,chromlen, ref, numHypotheses, 100*(totalLoci/float(flg)))+'% complete)',1)
				
				# GET READ PILEUP FOR THIS POSITION
				# ---------------------------------------------------------------
				# slight code duplication this way but fewer if statements
				if DEGREADSTREAM:
					# Stream from a read buffer to balance speed and memory usage
					for pos in xrange(max(0, (Gi-maxreadlen)+1, prevGi+1), min(Gi+1, chromlen)):
						if len(SGr) >= maxdepth: 
							# pipeit('Notice: pileup at %s:%s exceeded maximum read depth, %s > %s'%(chrom,pos,len(SGr), maxdepth),1)
							# randomly trim?
							# thek = SGr.keys()
							while len(SGr) > maxdepth: del SGr[random.choice(SGr.keys())]
							# break
						
						# FETCH CORRESPONDING UNIQUE READS
						# --------------------------------
						for pese in RHK:
							# having group-indexed the reads by starting position we can do this
							# each key of SGr is the read id, returning a list of pos,seq,phred triples
							
							# first check previous line
							# hlo = hash(curLine[pese][0])
							hlo = curLine[pese][0]
							if curLine[pese][2]==chrom and int(curLine[pese][3])-1 == pos and hlo not in filterids:
								#                         pos         seq    qual     readlength
								SGr[(hlo,pese)] = [(int(curLine[pese][3])-1, curLine[pese][9], curLine[pese][10], len(curLine[pese][9]))]
								
							if not REACHEDENDOFFILE[pese]:
								try:
									while curLine[pese][2] < chrom or (curLine[pese][2]==chrom and int(curLine[pese][3])-1 <= pos):
										try:
											curLine[pese] = URB[pese].pop(0)[:-1].split('\t') # fetch the next line from the buffer
											# curLine[pese] = readhandle[pese].next()[:-1].split('\t') # fetch the next line from the buffer
											
											# no longer need to hash the read id? - yes because of filterids...
											# try without
											hlo = curLine[pese][0]
											if curLine[pese][2]==chrom and int(curLine[pese][3])-1 == pos:
												
											# check chrom, position, and filter file
											# hlo = hash(curLine[pese][0])
											# if curLine[pese][2]==chrom and int(curLine[pese][3])-1 == pos and hlo not in filterids:
												#                         pos         seq    qual     readlength
												SGr[(hlo,pese)] = [(int(curLine[pese][3])-1, curLine[pese][9], curLine[pese][10], len(curLine[pese][9]))]
										except IndexError: 
											# reload buffer
											retval = reloadBuffer(URB[pese], readhandle[pese], bufferlimit)
											if retval == -1: break # end of file
								except StopIteration: REACHEDENDOFFILE[pese] = True
						
						# FETCH CORRESPONDING DEGENERATE READS
						# ------------------------------------
						# stream-read the deg index files
						# after we fetch a line, then we need to fetch the sequence and quality information from the actual SAM file
						for pese in dRHK:
							# first check previous line
							if dcurLine[pese][0]==chrom and int(dcurLine[pese][1]) == pos:
								# parse this line: chrom,pos,alns
								# glob = [map(lambda x: map(int,x.split(':')), dcurLine[pese][2].split(','))]
								
								startBytes = int(dcurLine[pese][2])
								nlines = int(dcurLine[pese][3])
								glob = [[[startBytes,nlines>maxnumalign and maxnumalign or nlines]]]
								
								# now fetch information from SAM file
								for lineblock in map(lambda bytelst: fetchLines(readmaph['%sdeg'%(pese)],bytelst), glob):
									altaln = []; ehitid = None
									for line in lineblock:
										lst = line[:-1].split('\t')
										# ohitid = hash(lst[0]); 
										
										ochrom = lst[2]; opos=int(lst[3])-1; oseq=lst[9]; oqual=lst[10]
										ohitid = lst[0]
										if opos == pos:
											ehitid = ohitid
											SGr[(ehitid,pese)] = [(opos, oseq, oqual, len(oseq))]
										else: altaln += [(opos, ochrom, oseq, oqual)]
									try: SGr[(ehitid,pese)] += altaln
									except KeyError: pass # left over - there is no alignment that overlaps this position
							
							if not dREACHEDENDOFFILE[pese]:
								try:
									while dcurLine[pese][0] < chrom or (dcurLine[pese][0]==chrom and int(dcurLine[pese][1]) <= pos):
										try:
											dcurLine[pese] = DRB[pese].pop(0)[:-1].split('\t')
											# dcurLine[pese] = dreadhandle[pese].next()[:-1].split('\t')
											
											if dcurLine[pese][0]==chrom and int(dcurLine[pese][1]) == pos:
												# parse this line: chrom,pos,alns
												# glob = [map(lambda x: map(int,x.split(':')), dcurLine[pese][2].split(','))]
												
												startBytes = int(dcurLine[pese][2])
												nlines = int(dcurLine[pese][3])
												glob = [[[startBytes,nlines>maxnumalign and maxnumalign or nlines]]]
												
												# now fetch information from SAM file
												for lineblock in map(lambda bytelst: fetchLines(readmaph['%sdeg'%(pese)],bytelst), glob):
													altaln = []; ehitid = None
													for line in lineblock:
														lst = line[:-1].split('\t')
														# ohitid = hash(lst[0]); 
														ohitid = lst[0]
														ochrom = lst[2]; opos=int(lst[3])-1; oseq=lst[9]; oqual=lst[10]
														if opos == pos:
															ehitid = ohitid
															SGr[(ehitid,pese)] = [(opos, oseq, oqual, len(oseq))]
														else: altaln += [(opos, ochrom, oseq, oqual)]
													try: SGr[(ehitid,pese)] += altaln
													except KeyError: pass # left over - there is no alignment that overlaps this position
										except IndexError: 
											# reload buffer
											retval = reloadBuffer(DRB[pese], dreadhandle[pese], bufferlimit)
											if retval == -1: break # end of file
								except StopIteration: dREACHEDENDOFFILE[pese] = True
								
				else:
					for pos in xrange(max(0, (Gi-maxreadlen)+1, prevGi+1), min(Gi+1, chromlen)):
						if len(SGr) >= maxdepth: 
							# pipeit('Notice: pileup at %s:%s exceeded maximum read depth, %s > %s'%(chrom,pos,len(SGr), maxdepth),1)
							# randomly trim
							# thek = SGr.keys()
							while len(SGr) > maxdepth: del SGr[random.choice(SGr.keys())]
							# break
						
						
						# FETCH CORRESPONDING UNIQUE READS
						# --------------------------------
						for pese in RHK:
							# having group-indexed the reads by starting position we can do this
							# each key of SGr is the read id, returning a list of pos,seq,phred triples
							
							# first check previous line
							# hlo = hash(curLine[pese][0])
							hlo = curLine[pese][0]
							if curLine[pese][2]==chrom and int(curLine[pese][3])-1 == pos and hlo not in filterids:
								#                         pos         seq    qual     readlength
								SGr[(hlo,pese)] = [(int(curLine[pese][3])-1, curLine[pese][9], curLine[pese][10], len(curLine[pese][9]))]
								
							if not REACHEDENDOFFILE[pese]:
								try:
									while curLine[pese][2] < chrom or (curLine[pese][2]==chrom and int(curLine[pese][3])-1 <= pos):
										try:
											curLine[pese] = URB[pese].pop(0)[:-1].split('\t') # fetch the next line from the buffer
											# curLine[pese] = readhandle[pese].next()[:-1].split('\t') # fetch the next line from the buffer
											
											# no longer need to hash the read id? - yes because of filterids...
											# try without
											hlo = curLine[pese][0]
											if curLine[pese][2]==chrom and int(curLine[pese][3])-1 == pos:
												
											# check chrom, position, and filter file
											# hlo = hash(curLine[pese][0])
											# if curLine[pese][2]==chrom and int(curLine[pese][3])-1 == pos and hlo not in filterids:
												#                         pos         seq    qual     readlength
												SGr[(hlo,pese)] = [(int(curLine[pese][3])-1, curLine[pese][9], curLine[pese][10], len(curLine[pese][9]))]
										except IndexError: 
											# reload buffer
											retval = reloadBuffer(URB[pese], readhandle[pese], bufferlimit)
											if retval == -1: break # end of file
								except StopIteration: REACHEDENDOFFILE[pese] = True
						
						# FETCH CORRESPONDING DEGENERATE READS
						# ------------------------------------
						# non streaming version
						for pese in dRHK:
							if pos not in globRd[pese][chrom] or not readmaph['%sdeg'%(pese)]: continue
							
							for lineblock in map(lambda bytelst: fetchLines(readmaph['%sdeg'%(pese)],bytelst), globRd[pese][chrom][pos]):
								altaln = []; ehitid = None
								for line in lineblock:
									lst = line[:-1].split('\t')
									# ohitid = hash(lst[0]); 
									ohitid = lst[0]
									ochrom = lst[2]; opos=int(lst[3])-1; oseq=lst[9]; oqual=lst[10]
									if opos == pos:
										ehitid = ohitid
										SGr[(ehitid,pese)] = [(opos, oseq, oqual, len(oseq))]
									else: altaln += [(opos, ochrom, oseq, oqual)]
								try: SGr[(ehitid,pese)] += altaln
								except KeyError: pass # left over - there is no alignment that overlaps this position
						
				
				# final check?
				# while len(SGr) > maxdepth: del SGr[random.choice(SGr.keys())]
				
				# DONE OBTAINING READ PILEUP
				prevGi = Gi # Update previous position we visited for next pileup loading
				
				####################################################################
				
				# Compute the sum log likelihood of read pileup across genotypes
				LL = LLg0[:]                             # read-sum log likelihoods
				AF = {'A':0, 'T':0, 'C':0, 'G':0, 'N':0} # allele frequency distribution
				multiplicity = 0 # number of reads that are degenerate (have >1 alignment; some k < n)
				totalmultiplicity = [] # total number of degenerate alignments (d-n)
				marked = [] # list of readids marked to be deleted
				
				for readidpese,alignments in SGr.iteritems():
					(pos,seq,qual,curRL) = alignments[0]
					altAlignments = alignments[1:]
					
					# first alignment overlaps current position on chrom
					offset = Gi-pos   # what is nt of interest wrt this read?
					
					# if offset beyone read length, read doesn't actually overlap current locus
					if offset >= curRL: marked += [readidpese]; continue
					
					AF[seq[offset]] += 1 # update allele counts
					
					# what is the size of SG(l), the k-bounded string set 
					# for the resequenced genome SG at position l
					# that is, total number of degenerate alignments + alignment at this position
					# This is d+1 = |alignments|
					
					# However, it is important to note that PE reads no longer carry /1 and /2,
					# so we need to distinguish different ends by quality
					# each mate will have one of 2 qualities (reverse)
					# this is actually an assumption in sniper - we break up the degenerate paired end reads
					if len(altAlignments) > 0 and readidpese[1] == 'PE':
						# that is, degenerate paired end reads
						# we only want the group of alignments carrying the alignment of interest
						groups = {}
						for tup in altAlignments:
							cq = tup[-1]; rcq = cq[::-1]
							if cq in groups: groups[cq] += [tup]
							elif rcq in groups: groups[rcq] += [tup]
							else: groups[cq] = [tup]
							
						# now which group do we want here?
						if qual in groups: altAlignments = groups[qual]
						else: 
							try: altAlignments = groups[qual[::-1]]
							except KeyError: altAlignments = [] # no deg alignments of interest for this read
							
					# final update to number of templates
					numTemplates = len(altAlignments)+1 # = d+1, where d is multiplicity for this read
					
					# now have d+1 alignments, which are all degenerate
					multiplicity += (numTemplates>1 and 1) or 0 # are there any degenerate alignments?
					totalmultiplicity += [numTemplates-1] # add d = len(alignments)
					
					#    likelihood = (1/d+1)P(ri|Gl=x) + (1/d+1)Pdeg1 + ... + (1/d+1)Pdegd
					#               = (1/d+1)[P(ri|Gl=x) + Pdeg1 + ... + Pdegd]
					#               = (1/d+1)[P(ri|Gl=x) + sum_d(Pdeg)]
					#               ~ (1/d+1)P(ri|Gl=x) + (1/d+1)d x Pdeg
					#    so we are just computing a weighted average
					
					# now the 1/d+1 term is the probability of one genomic segment 
					# serving as the template for sequencing: P(s | G)
					# here this is uniformly distributed across reads
					# below is an implementation where P(s | G) ~ P(s | ri, G)
					# where we solve in a Bayesian fashion:
					# P(s | ri, G) = P(ri | s, G)P(s,G) / Sum P(ri | s, G)P(s,G)
					# with uniform prior this reduces to
					# P(s | ri, G) = P(ri | s, G) / Sum P(ri | s, G)
					# which is a weighting based on relative strength of the read aligning to 
					# that genomic region
					# basically this says that for degenerate reads, give the most weight 
					# to the location with fewer mismatches or higher quality
					
					# if ploidy = 2, then really have 2d + 2 alignments for this read
					Pcond = ploidy == 1 and PriSj[numTemplates] or PriSj[2*numTemplates]
					# Note, we could formulate Pcond in a more read-specific manner...
					
					# the following is the main speed bottleneck by far
					# so this is implemented in C
					# compute the likelihood for this read given the offset position
					# Prlo =      ntPriS(seq,GRseq[pos:pos+curRL],qual,offset,curRL)
					# Prl = maths.ntPriS(seq,GRseq[pos:pos+curRL],qual,offset,curRL,Lambda,expError)
					
					Prl = compLik(seq,GRseq[pos:pos+curRL],qual,offset,curRL,Lambda,expError)
					
					
					# if Prl != Prlo:
					# 	print 'Compare', Gi
					# 	print 'orig', Prlo
					# 	print 'new ', Prl
					
					# sum up the degenerate probability mass: info = (opos, ochrom, seq, oqual)
					# consider variant reference genotypes at degenerate positions - same offset as reference
					# Pdeg = sum( map(lambda info: max(ntPriS(info[2],GR[info[1]]['seq'][info[0]:info[0]+curRL],info[3],offset,curRL).values()), altAlignments) )*Pcond
					# C code
					# Pdeg = sum( map(lambda info: max(maths.ntPriS(info[2],GR[info[1]]['seq'][info[0]:info[0]+curRL],info[3],offset,curRL,Lambda,expError).values()), altAlignments) )*Pcond
					
					Pdeg = sum( map(lambda info: max(compLik(info[2],GR[info[1]]['seq'][info[0]:info[0]+curRL],info[3],offset,curRL,Lambda,expError).values()), altAlignments) )*Pcond
					
					twoPdeg = 2*Pdeg # saves on computation time
					
					LL = ploidy == 1 and map(lambda g: LL[g] + log(Pcond*Prl[GT[g]] + Pdeg), rangelengenotypes) or \
					                 map(lambda g: LL[g] + log(Pcond*(Prl[GT[g][0]] + Prl[GT[g][1]]) + twoPdeg), rangelengenotypes)
				
				# done loop over reads
				map(delete,marked) # delete marked reads from SGr
				
				# ===============================================
				# # pileup plots
				# if Gi == 310:
				# 	x = [(tup[0],tup[1],tup[-1]) for readid,tup in SGr.iteritems()]
				# 	x.sort()
				# 	print 'pileup at %s (n=%s reads)'%(Gi, len(SGr))
				# 	for pos,seq,rl in x:
				# 		print ''.join([' ' for zz in xrange(rl-(Gi-pos))]), seq, pos
				# ===============================================
				
				# what is final read depth at position Gi in resequenced genome?
				depth = len(SGr)
				# don't print out info if position has no depth
				if depth < mindepth: continue
				
				# need to scale down the likelihoods so that natural scale values don't become 0
				# we are currently on log scale so values are large negative exponents
				if sum(LL) < 1e-250: # that is some really small value
					# to avoid numerical underflow due to exponentiation of large negative LL values
					# need to find a positive constant C such that math.e^LL+C is computable, eg < 1000
					
					# precursor step is to do blind rescaling to bring values near computable range
					# if you don't do this, you will inevitably set many positions values to equality
					# whereas they could be quantified just fine
					maxll = max(LL)
					while maxll < -200:
						c = -maxll/4. # this can't become positive
						LL = map(lambda x: x+c, LL)
						# print >> sys.stderr, 'Step:', lik
						maxll = max(LL)
						
					badvalue = -100000000000000000000000000000000 # token indicator of too small value
					LL = map(lambda x: (x < -250 and badvalue) or x, LL)
					
					# confirm that not all have badvalue
					if len(filter(lambda x: x!=badvalue,LL)) == 0:
						print >> sys.stderr, 'Warning: All likelihoods are 0 at position', Gi
						# then all likelihoods have the same value
						LL = map(lambda x: -250, LL)
					else:
						idx = argmax(LL) # get the largest likelihood
						candC = -LL[idx] - 100 # makes the max possible log likelihood = -100: 10**-100 = 1e-100
						# all other liklihoods will be smaller than this
						LL = map(lambda x: (x == badvalue and -250) or (x+candC), LL)
				# DONE CORRECTION
				
				
				# Return likelihoods to natural scale
				LL = map(lambda x: 10**x, LL)
				
				# Allele frequency
				# ----------------
				AFstr = ','.join( map(lambda nt: nt+':'+str(round(AF[nt]/float(depth),2)), nucleotides) )
				
				# Posterior probabilities
				# ----------------------------------------------
				# we now have a length-10 vector of likelihood values 
				# for all reads at Gi given each genotype
				# calculate the posterior probability of each possible genotype
				# given the read set
				
				# we must calculate the PGxyR for all possible genomic string configurations C(SGxy(l))
				# assuming genotype at position l is xy
				# so sum up prob over the C(SGxy(l)) for each genotype xy
				
				# so basic assumption is G = GR so that C(SGxy(l)) is simply the k-bounded
				# set of the reference: C(SGxy(l)) = C(SGRxy(l))
				# this way there is no summing involved
				
				# Multiplicity ratio: d-n / n (or Ndeg / N), ranging from 0 to inf
				# aka the average number of additional alignments per read / locus
				# Sum(d_read) / Sum(reads) per locus
				# --------------------------------------------------
				udalpha = round(harmonicMean(totalmultiplicity), 3)
				
				PGxyR = None
				try:
					# numerator: P(r1...rn | Gxy(Gi)) x P(Gxy(Gi))
					if ploidy == 1:
						PGxyR = map(lambda g: LL[g]*PG[g,ref], rangelengenotypes)
					else:
						# take maximum log likelihood among non-reference alleles
						dnt = {'A':None,'T':None,'C':None,'G':None, ref:None}; del dnt[ref]
						dL = max(map(lambda refH: max(map(lambda alt: LL[invGT[(refH,alt)]], dnt.keys())), iupac2nts(ref,double=False)))
						
						homweight = max(1e-10, 1.-(udalpha+1)*theta) # so minimum P(RR) = e-10
						hetweight = udalpha*theta
						
						# fix this to allow for bi-allelic reference character
						
						# PGxyR = None
						# try: # catch heterozygous reference alleles
						PGxyR = map(lambda g: GT[g][0]==ref and GT[g][0]==GT[g][1] and \
					                          LL[g]*homweight + dL*hetweight or \
					                          LL[g]*PG[g,ref], rangelengenotypes)
						# except KeyError:
						# 	# we know reference is heterozygous so default to no correction
						# 	PGxyR = map(lambda g: LL[g]*PG[g,ref], rangelengenotypes)
							
					# denominator: Sum_xy{ P(r1...rn | Gxy(Gi)) x P(Gxy(Gi)) }
					Ptotal = float(sum(PGxyR))
				except OverflowError:
					print 'snpper(overflow error):\nLL=',LL#, '\nprior=', [PG[i,ref] for i in rangelengenotypes]
					continue
					
					
				# Final posterior probabilities for all genotypes at position Gi
				# --------------------------------------------------------------
				# double up the heterozygous probabilities, e.g. P[]
				PGxyR = map(lambda i: I2xHet[genotypes[i]]*PGxyR[i], rangelengenotypes)
				Ptotal = float(sum(PGxyR))
				PP = map(lambda x: x/Ptotal, PGxyR)
				
				# Determine MAP genotype and posterior
				# ------------------------------------
				# take argmax, but multiple genotypes may have similar or identical PP. 
				# This would happen if the pileup is highly degenerate. 
				idx = argmax(PP)
				geno = genoUni[genotypes[idx]]
				MAP = PP[idx]>1 and 1 or PP[idx] # force to 1
				Q = nastr
				try: Q = int(-10*log(1-MAP)) # convert to phred score
				# this occurs if MAP = 1, so force to some arbitrary minimum pvalue
				except ValueError: Q = lowestQ # this will be Q = 160
				# occurs if MAP is < about 1e-16
				except OverflowError: Q = lowestQ # this will be Q = 160
				
				# Record statistics
				info = [chrom, str(Gi), ref, geno, str(Q), str(depth), str(multiplicity), str(sum(totalmultiplicity)), '%.3f'%(udalpha), 'N', AFstr]
				# if verbose: print info
				
				# fixed haploid/diploid issue
				genoref = (ploidy > 1 and iupac2nts(ref)) or ref
				if MAP == 0: info[3] = genoref # default to ancestor
				# print 'genoref', genoref, 'ref', ref, 'geno', geno
				# Assess nominal significance (P<0.05 and writes on the fly)
				# ---------------------------
				numHypotheses += 1 # we are evaluating a hypothesis
				
				
				# or only print nominally significant snps
				if Q>=13 and genoref != geno:
					info[9] = 'Y'
					info = '\t'.join(info)
				else: info = '\t'.join(info)
				
				
				if saveall: 
					alloutfh.write(info+'\n')#, print >>alloutfh, info # if printing for all loci
				# if saveall: print >>alloutfh, info # if printing for all loci
				elif outfile and Q>=QvalueRange[0] and genoref!=geno: print >>outfh, info # if printing significant SNPs only
				
				
		# DONE LOOP OVER CHROMOSOME GROUPS
	# DONE GENOME
	
	if outfile: outfh.close()
	if saveall: alloutfh.close()
	pipeit(' - Finished (n=%s loci).'%(numHypotheses),1)
	
	if not IDENTIFYSIGNIFICANTS: return numHypotheses
	pipeit(' - Identifying significant SNPs for %s...'%(sampleID),1)
	filterSignificantSNPs(sampleID+boundslabel, saveall, ADAPTIVEQ=ADAPTIVEQ, QvalueRange=QvalueRange, prior=prior)
	
	for k in readmaph.keys(): readmaph[k].close()
	for k in dreadhandle.keys(): dreadhandle[k].close()
	
	return numHypotheses


numHypos = 0
def filterSignificantSNPs(basename, files, ADAPTIVEQ=True, QvalueRange=[20], prior='resequence'):
	# sampleID must be sampleID+boundslabel
	global numHypos
	numHypos = 0
	if type(files) != type([]): files = [files]
	
	# first check if we need to recompute anything
	checkups = []
	
	if ADAPTIVEQ:
		# multiple-test corrected snp variants
		snpfile = snpdir+basename+'.aQ%s,%s.txt'%(ADAPTIVEQ,int(readlength))
		access = os.access(snpfile, os.F_OK)
		if access and not REDOFINALSNPCALLING: return 1
		if access: os.system('rm "%s"'%(snpfile))
		checkups += [(ADAPTIVEQ,snpfile)]
	else:
		for minPvalue in QvalueRange:
			# multiple-test corrected snp variants
			snpfile = snpdir+basename+'.Q%s.txt'%(minPvalue)
			access = os.access(snpfile, os.F_OK)
			if access and not REDOFINALSNPCALLING: return 1
			if access: os.system('rm "%s"'%(snpfile))
			checkups += [(minPvalue,snpfile)]
	
	checkups.sort()
	
	if len(checkups) > 0:
		# first get total number of hypotheses from mother and store data
		for fn in files:
			label = fn.split('/')[-1]
			infh = None
			try: infh = open(fn)
			except IOError: pipeit('Cannot access: %s...SKIP.'%(fn),1); continue
			pipeit(' - Counting number of genomic positions evaluated from %s...'%(label))
			# numHypos = sum([qccheck(line) and 1 or 0 for line in fh])
			def addup(v):
				global numHypos
				numHypos += v
			
			[addup(int(qccheck(line))) for line in infh]
			infh.close()
			
			# if mother file contained no information
			if numHypos == 0: numHypos = 1
			pipeit('n=%s hypotheses.'%(numHypos),1)
			
			refilt = {}
			if REDUNDANCYFILTER and os.access(REDUNDANCYFILTER, os.F_OK):
				# print 'REFILTERING'
				fh = open(REDUNDANCYFILTER)
				for line in fh:
					entries = map(lambda x: x.split(','), line[:-1].split('\t'))
					# what are all targets
					offset = 0
					tlist = map(lambda x: [x[0],int(x[1])+offset], entries)
					for c,p,l in entries: refilt[c,int(p)+offset] = tlist
				fh.close()
			
			# now we can post-filter the significant set 
			# based on the actual number of hypotheses we made
			pipeit(' - Filtering SNPs...',1)
			for Qvalue,snpfile in checkups:
				# generate list of significant genotypes
				outfh = open(snpfile, 'a')
				print >>outfh, '# Total positions tested (%s): %s'%(label, numHypos)
				keepers = []
				jointmaybes = {}
				nsig = 0
				infh = open(fn)
				ct = 0
				for line in infh:
				# for i in xrange(len(data)):
					# row = data[i][:-1].split('\t')
					ct += 1
					if not qccheck(line): 
						sys.stderr.write('Warning: Bad line (%s): %s\n'%(ct, line[:-1])); sys.stderr.flush()
						continue
					row = line[:-1].split('\t')
					tchrm = row[0]; tpos = int(row[1])
					Q = (row[4]==nastr or row[4]=='NA' and nastr) or float(row[4])
					ref = row[2]; geno = row[3]; alpha = float(row[8])
					# genoref = ref
					genoref = (prior!='haploid' and iupac2nts(ref)) or ref
					
					# print prior, 'cf', genoref, 'vs cons', geno, 'ref', ref, 'iupac', iupac2nts(ref)
					
					if genoref != geno and Q != nastr:
						# check for SNP candidacy
						ISCANDIDATE = False
					
						if ADAPTIVEQ: 
							# compute adaptive Q threshold based on number of correlated hypotheses
							# ie, number of loci with shared reads ~ 2 read len + (num deg alignments / read len)
							numDegReads = int(row[6])
							
							# this is arithemetic mean - more appropriate
							numDegAlnPerRead = int(row[7])/float(row[5])
							
							# udalpha is too lenient for this multiple correction
							# numDegAlnPerRead = alpha # this is udalpha or mean deg alignments ~ number of additional correlated loci
							# APthresh = ADAPTIVEQ / (2*readlength + numDegReads * numDegAlnPerRead * readlength)
						
							AQthresh = P2Q( ADAPTIVEQ / (2*readlength * (1 + numDegAlnPerRead)) )
							# AQthresh = P2Q(APthresh)
							# print 'needed', AQthresh, readlength, numDegAlnPerRead, row[5], row[7]
							if Q >= AQthresh: ISCANDIDATE = True
						
						elif (not MAXALPHA and Q >= Qvalue) or (MAXALPHA and alpha <= MAXALPHA and alpha >= MINALPHA and Q >= Qvalue):
							ISCANDIDATE = True
					
					
						# postprocessing and postfilter are deprecated
						elif POSTPROCESSING:
							# basically, if we have high depth and at most 10:1 deg to unique, 
							# then ignore the quality scores at this position
							# moreover if AAFD is close to 50:50 or 100:0 but MAP genotype does not
							# reflect this, then fix the genotype
							AAFD = sorted(map(lambda y: (float(y[1]),y[0]), map(lambda x: x.split(':'), row[10].split(','))))
							depth = int(row[5])
							alpha = float(row[8])+1e-10
						
							# basically this says ignore the P-value if the depth and degeneracy are ok
							# realistically this is just like using maq except we have degeneracy values
							if (depth >= FDEPTH[0] and depth/alpha >= FDEG[0] and Q >= FQ[0]) or (depth >= FDEPTH[1] and depth/alpha >= FDEG[1] and Q >= FQ[1]):
								ISCANDIDATE = True
							
						if POSTFILTER:
							depth = int(row[5]); ddeg = int(row[6])
							if float(ddeg)/depth > DDEGCUTOFF: ISCANDIDATE = False
					
						# find SNPs that have template redudndancy at this Qvalue
						if ((not MAXALPHA and Q >= Qvalue) or (MAXALPHA and alpha <= MAXALPHA and alpha >= MINALPHA and Q >= Qvalue)) and (tchrm,tpos) in refilt: 
							ISCANDIDATE = False
							
							# reactivate if desired
							# jointmaybes[(tchrm,tpos)] = [row,i]
						
						# FINALLY SAVE SNP IF STILL GOOD
						if ISCANDIDATE:
							row[9] = 'Y'
							print >>outfh, '\t'.join(row)
							nsig += 1
							# since this is largest Qvalue, we continue filtering rows...
							# this will accelerate future scans
							# reactivate if desired
							# keepers += [data[i]]
				infh.close()
				
				# update keepers with subset of jointmaybes
				# this is a list of the positions - cluster these into their redundancy groups
				# then choose the best 1 from each group
				candfile = snpdir+basename+'.candidate.Q%s.txt'%(Qvalue)
				grouping = {}
				groupkeeps = 0
				# at each position, check for snps at redundant positions
				for tchrm,tpos in jointmaybes.keys():
				
					# first, we want to isolate positions with the same snp call
					tmpdct = {} # group redundant snps by genotype
					for x in refilt[(tchrm,tpos)]:
						try: 
							tmp = jointmaybes[tuple(x)]
							entry = (int(tmp[0][4]), float(tmp[0][8]), tmp)
							try: tmpdct[tmp[0][3]] += [entry]
							except KeyError: tmpdct[tmp[0][3]] = [entry]
						except KeyError: pass
				
					# treat each genotype separately - we want to avoid replicate calls
					for cons,tmpobj in tmpdct.iteritems():
						tmplist = tmpobj[:] # list of positions with same genotype
						q = None; row = None; i = None
					
						# group calls by quality value
						qdct = {}
						for q,alpha,(row,i) in tmplist:
							try: qdct[q] += [(alpha,row,i)]
							except KeyError: qdct[q] = [(alpha,row,i)]
						maxq = sorted(qdct.keys())[-1] # get maximum q group
					
						# group snps of max quality by alpha
						adct = {}
						for alpha,row,i in qdct[maxq]:
							try: adct[alpha] += [(row,i)]
							except KeyError: adct[alpha] = [(row,i)]
						# keep the snp if there is a single best alpha
						minalpha = sorted(adct.keys())[0]
					
						tmplist = []
						for ttlist in adct.values(): tmplist += ttlist
					
						# temporarily disabled
						if len(tmplist) == 1:
							row,i = tmplist[0]
							row[9] = 'Y'
							print >>outfh, '\t'.join(row)
							keepers += [data[i]]
							groupkeeps += 1
						# else:
						# 	# return the remainder
						# 	if len(tmplist): print >> outfh2, tmplist
					
				# nsig = len(keepers)
				# data = keepers
				
				outfh.close()
				if ADAPTIVEQ:
					pipeit(' - %s/%s (%.3f) significant SNP variants at FWER < %s.'%(nsig, numHypos, nsig/float(numHypos), ADAPTIVEQ),1)
				else:
					pipeit(' - %s/%s (%s.3f) significant SNP variants at Q > %s (P < %.1e).'%(nsig, numHypos, nsig/float(numHypos), Qvalue, 10**(-Qvalue/10.)),1)
	


# -----------------------------------------------------------------
# -----------------------------------------------------------------


# CLI VARIABLES
# -------------------------------------------------------------------
outdir = slash(os.getcwd())

btdir = ''
btdirs = []
for x in btdirs:
	if os.access(x, os.F_OK): btdir = x; break

btmapdir = ''; SETMAPDIR = False
MAPCUT = False
findir = ''
mapfile = ''
bfaref = ''
zkey = 'sniper_results' # Default folder name for processed samples


########################### SNIPER PARAMETERS ###########################
#########################################################################

# PART I: ALIGNMENT PARAMETERS
# ----------------------------
RUNBOWTIE = False               # align reads to reference genome using bowtie
REALIGN = False                 # realign reads
# nprocs = 4                      # number of processors per computer for bowtie assembly

STRATEGY = 'all'                # broad mapping strategy
                                # one of: all, unique, best, bestno [default: all]
                                # unique: genotype using uniqely aligned reads only (--bestm -m 1)
                                # best: genotype using best-guess aligned reads only (--bestk -m 1)
                                # if UNIQ = BEST = False, align using --bestk/m -m d
                                # best-noguess: exclude best reads if multiply best alignments exist

MAPWITHQUALITY = False          # quality-blind or quality-aware mapping

alignmentPolicy = 'bestm'       # basic, best, all, bestk, bestm, maq-bestm, bestm-stratum, best-stratum
                                # highly recommended to use bestm for plant/vertebrate genomes
mismatches = {'SE':2, 'PE':2}   # number of mismatches for bowtie alignment (-v mismatches[x])
maxnumalign = 50                # maximum number of alignments a read could have, 
                                # subject to -mm mismathces (ONLY FOR policies bestm and bestk)
                                # set this to the expected read coverage
readfileformat = 'fastq'        # or fasta
COUNTREADS = False              # count number of mapped reads from bowtie output
# qualstr = '--phred64-quals'     # default is 64-base phred scores for Illumina GA 1.3+
qualstr = '--phred33-quals'     # default is 33-base phred scores for recent Illumina
trim5 = ''                      # trim 5' end of reads by trim5 bases
trim3 = ''                      # trim 3' end of reads by trim3 bases
COLORSPACE = False              # use bowtie colorspace
minins = 100; maxins = 400      # mate pair insertion size, default lower and upper bounds
mininsCLI = None; maxinsCLI = None # if user provides this by command line
GETINSERTSIZE = False           # estimate PE insert size distributions
insstat = '.99'                 # take as insert bounds the inner insstat percentiles
btcustom = ''                   # user-specified string fed directly to bowtie

# PART II: READ FILE ORGANIZATION
# -------------------------------
ORGANIZEREADS = False           # organize bowtie ouput into 4 read files per sampleID: unique/deg/PE/SE
REORGANIZEREADS = False         # redo this read organization
PESECOMBO = True                # save PE and SE alignments to single map file
EXORGANIZE = False              # organize a preexisting map file
KEEPBOWTIE = False              # retain the bowtie output in addition to organized map files...
CLEARBOWTIE = False
INDEXMAPFILES = False           # sort unique map files by chrom,pos
INDEXUNI = False                # only index unique reads
INDEXDEG = False                # only index deg reads
RESORT = False                  # resort unique index
INDEXONLY = False               # assume map is already sorted
DEGREADSTREAM = True           # stream read the degenerate map instead of bulk pre-loading by scanning the file for each LIP groups
LINKUNIQUE = False              # if you only have a unique map, this will link file name to what sniper wants

# PART III: SNP CALLING
# ---------------------
SNPCALLING = False              # whether or not to perform genotyping
REDOSNPCALLING = False          # whether to redo SNP calling on a sampleID which has already been typed
IDENTIFYSIGNIFICANTS = False    # whether to filter for significant SNPs
REDOFINALSNPCALLING = False     # correct for multiple testing to return final significant snps
SAVENOMINAL = False             # only print out nominally significant SNPs while processing genome

USESINGLETONS = True            # genotype using both PE and SE reads
boundsfile = ''                 # file containing chrom\tlow\thigh rows for restricting genotyping
singlesampleid = []             # name of single sampleID to genotype
LOWSID = None; HIGHSID = None   # if user wants to bound which samples are genotyped in a map file
maxreadlen = 150                # currently global hardwired string length of all reads
Lambda = 0.67                   # weighting factor; larger values up-weight global 
                                # binomial pr. vs. read-specific phred prob.
mindepth = 1                    # minimum reads to evaluate a nucleotide locus
maxdepth = 500                  # maximum reads to evaluate a nucleotide locus (if there are more only the first n will be used)

expPhred = 35                   # Expected per-nucleotide sequencing error for global binomial likelihood model
ESTIMATEGLOBALERROR = False     # estimates the value for expPhred
prior = 'resequence'            # resequence, maq, uniform, haploid
theta = 0.001                   # expected human heterozygosity
colorspacetheta = (3/4.)*theta  # overall haploid divergence for sample--reference genome for bowtie colorspace conversion
SETINTQUALS = False             # set bowtie's --integer-quals flag

# Default range of Q values for significant SNP identification
# QvalueRange = [0, 10, 13, 20, 40, 80, 100, 160]
QvalueRange = [40]# [13, 20, 40, 80, 120]

# use --fwer P,RL
ADAPTIVEQ = 0.05
readlength = 50 #36

# PARALLEL PROCESSING AND MEMORY USAGE
# ------------------------------------
LIPieces = 1                    # Break loading of unique reads into this number of pieces
                                # Primary determinant of memory usage
nsubprocs = 1                   # number of processors for SNP computation (true multiprocessing)
ISSUBPROC = False # if this sniper call is a subprocess call from a master sniper process
USESGE = False
SGEMEM = '-l mem_free=4G'       # less required for non-mammalian genome
xflag = 'groupchrom'


# SNP POSTPROCESSING SETTINGS
# ---------------------------
MAXALPHA = None                 # don't call any SNPs that have alpha > 10
MINALPHA = 0.0
POSTFILTER = False # similar to maxalpha but looks at num deg reads / num reads at locus
                   # instead of num deg alignments vs num reads at locus
DDEGCUTOFF = 0.9   # if more than 90% of depth is due to degenerate reads then can't be a SNP
REDUNDANCYFILTER = None         # Can optionally include a file of redunant loci to mask from SNP calling

POSTPROCESSING = False        # use a maq-like quality filter
                              # allows 2 criteria to call a SNP
FDEPTH = [10,50]              # if depth at least x
FDEG = [5,1]                  # if depth/degeneracy ratio at least x
FQ = [13,13]                  # if posterior prob (as phred) at least x

# OTHER PARAMETERS
# ----------------
REDOALL = False               # redo everything
DIAGNOSTICS = False           # generate summary statistics for genotype results
SAVEASSEMBLY = False          # save new genome using new SNP calls
SKIPHIST = False
specialReferenceFiles = None
mapsuffix = '.sam'
SHUFFLEBOUNDS = False         # 

MERGESIDS = None              # list of sampleIDs to merge parts into single mother file per sampleID
REMOVEPREMERGED = False       # whether to delete the pre-merged parts after the merge
MERGESORT = False
JOINPIECES = False
CLEANFILE = None

# Internal
# READLENBYUSER = False
DRYRUN = False                # dryrun the pipeline
callprefix = ''               # specify extra path information if sniper.py is not in global $PATH

##########################################################################
##########################################################################


# CLI ARGS
# -------------------------------------------------------------------
help = """
sniper.py v%s

Full help menu: sniper.py --help/-h
Description and requirements: sniper.py --version/-v
Web page: http://kim.bio.upenn.edu/software/sniper.shtml
Latest version: http://kim.bio.upenn.edu/software/sniper/sniper_pkg.zip

Usage: sniper.py -r ref.fa (-m mapfile.txt|-id samplename) [options]

Genotyping with an existing read map file:
   The read map file should be named <mapdir>/samplename.map, where samplename
   is provided in SampleID field of the map file (see below).
   
   If the read map file contains unique reads only (if already sorted, 
   include --nosort to avoid resorting):
   > sniper.py -r ref.fa -i <mapdir> --link --index -id samplename -s
   
   If the read map file contains both unique and non-unique reads, use 
   --exO to partition the file into unique and non-unique portions:
   > sniper.py -r ref.fa -i <mapdir> --exO --index -id samplename -s

Read mapping and genotyping:
   > sniper.py -r ref.fa -m mapfile.txt -z
   
   This will align raw NGS reads, organize and index reads, perform 
   genotyping, and provide diagnostics and new assemblies for all 
   samples listed in the map file. Results will be stored in a new 
   directory 'sniper_results' in your current working directory.
   
   Note, -z is equivalent to writing 
   "-a -O --index -s -d --saveass -Q 40 --fwer 0.05,50"

Use sniper.py -h for a complete description of arguments.


Standard arguments
==================
   -m <STR>:       map file (see below for format)
   -r <STR>:       reference genome (required, fasta format)
   -i <STR>:       directory to find pre-existing map file(s)
   -o <STR>:       directory in which to save results 
                   (default: current working directory ./)
  --name <STR>:    name of the output/results directory 
                   (default: sniper_results)
   -z:             perform/redo all procedures (equivalent to 
                   -a -O --index -s -d --saveass -Q 40 --fwer 0.05,50)
   -a/-ra:         (re)align reads (generate sam-format read maps with bowtie)
   -O/-rO:         (re)organize read maps into 2 files: 
                     sampleid.unique.map, sampleid.degenerate.map
                   then build an index for each file
  --splitmap       Separate single-end and paired-end reads into different map files
  --link           if you only have a unique map, this links to that file 
                   directly rather than organizing new files
   -exO:           organize a pre-existing read map into the 4 files above 
                   (Default: sampleID.sam, where sampleID is in mapfile)
   -s/-rs:         perform/redo SNP calling (and significance filtering)
   -so/-rso <Q>    perform/redo SNP calling, reporting only SNPs with stringency >= Q
  --sig:           perform significance filtering on existing genotypes
  --resig:         redo significance filtering on existing genotypes
  --diag:          generate diagnostics file summarizing SNP calls
  --saveass:       create a new fasta file of the reference genome modified 
                   by significant SNP calls (IUPAC format). NB, by default
                   -Q 40 is used, but you can return multiple assemblies: 
                   Ex: --saveassembly -Q 20,30,40,100

Key parameters
==============
EX: sniper.py -m <map_file.txt> -r <ref_genome.fasta> -z -k 2 
    -d 10 -e 30 -l 0.5 --prior resequence --theta 0.01

Arguments:
  -k <1,2,3>:      maximum number of alignment mismatches (default: 2)
  -d <INT>:        maximum number of alignments per read (default: 50)
  -e <INT>:        global sequencing error rate (specified as phred value; 
                   E.g. Q=30 == P<0.001) (default: 35)
  -l <FLOAT>:      reweight the likelihood model towards read-specific or 
                   global binomial probability (default: 0.67 towards binomial)
  --prior <STR>:   prior probability distribution (default: resequence)
                   Options: 
                   resequence:     h0:1-theta-theta^2, t1:theta/2, h2:theta/2, t2:theta^2
                   resequence-het: h0:1-theta-theta^2, t1:theta, h2:theta^2/2, t2:theta^2/2
                   uniform:        h0:1-3*theta, t1:theta, h2:theta, t2:theta
                   maq:            h0:(1-theta-theta^2)/2, t1:theta, h2:(1-theta-theta^2)/2, t2:theta^2
                   naive:          h0:0.25, t1:0.25, h2:0.25, t2:0.25
                   haploid:        h0:1-theta, t1:theta
  --theta <FLOAT>: expected divergence (heterozygous and homozygous) from 
                   reference genome (default: 0.001)

Read map indexing
=================
EX: sniper.py -m <map_file.txt> -r <reference_genome.fasta> --index

Arguments:
  --index:         sorts and indexes the unique SE and PE map files
  --reindex:       replaces an existing index
  --(re)indexuni:  index unique maps only
  --(re)indexdeg:  index degenerate maps only
  --nosort:        index only, assuming map files are already sorted

MEMORY / RUNTIME / SCHEDULING options
=====================================
EX: sniper.py -m <map_file.txt> -r <reference_genome.fasta> -s -x 5
For each sampleID, break the genome into 5 pieces, then launch and maintain 
up to 5 parallel processes on your local machine. The total number of jobs 
equals 5 times the number of sampleIDs in the map file. If --sge is added 
schedule jobs across CPUs using Sun Grid Engine.

EX: sniper.py -m <map_file.txt> -r <reference_genome.fasta> -s -x 15 
    --rest <restriction_file.txt>

This will launch and maintain up to 15 parallel processes, where the total 
number of jobs is equal to the number of sampleIDs in the map file times the 
rows in <restriction_file.txt>.
   -x <INT>:    distribute jobs over <INT> processes on 1 machine (multi-core)
  --groupchrom: group multiple chromosomes into a single job (default for -x 1)
  --subchrom:   split each chromosome into -x jobs
  --bychrom:    process each chromosome as a separate file, regardless of -x
                (allows you to add more processors mid-run or recover from
                 errors more quickly.)
  --ids <STR[,STR,...,STR>: restrict genotyping to a subset of SampleIDs
 
  --stream:     stream the degenerate index (Default: True)
  --buffer <INT>: sets streaming buffer to approx. <INT> MB (Default: 10)
 
  --sge:        distribute jobs using Sun Grid Engine. If -x > 1 launch 
                <INT> jobs per sampleID, otherwise launch one job for each 
                different sample to be processed
  --lip <INT>:  partition map into <INT> chunks to lower memory requirements
                (NB, requires <INT> scans of the map file, increasing runtime)
  --sgemem:     Required memory to launch process on SGE (default: 16G)
  --rest:       Specify a region or file containing a list of regions over 
                which to perform genotyping (default: none)
                EX: "--rest chr1"; genotype the entirety of chr1 only
                EX: "--rest chr1.chr5"; genotype chrom1 and chr5 only
                EX: "--rest chr1,0,1000"; genotype positions 0 to 1000 
                    of chr1, inclusive
                EX: "--rest chr1,0,1000,-"; genotype all of chr1 excluding 
                    positions 0 to 1000
                A restriction file should be a tab-delimited list of one or 
                more regions similarly described, replacing commas with tabs.
                Execute this shell command to generate one:
                echo -e "chrom1\\nchrom2\\nchrom3" > myrest.txt, or
                echo -e "chrom1\\t0\\t1000\\nchrom5\\t500\\t50000" > myrest.txt


Other notable arguments
=======================
  --path <DIR>:   optional directory path to sniper.py file
  --mindepth <INT>: minimum number of reads to evaluate a nucleotide locus
  --maxdepth <INT>: maximum number of reads to evaluate a nucleotide locus.
                    If > <INT> reads only the first <INT> will be used.
  --fasta:       read files do not contain quality scores
  --qual33:      quality scores follow a phred 33 scale instead of default 
                 phred 64 scale
  --peonly:      only use paired end reads for SNP identification
  --all:         use all reads for mapping (based on --policy bestm with 
                 maximum of d alignments)
  --uniq:        only use uniquely mapping reads for SNP identification. Also
                 if provided with -a for mapping
                 only return an alignment if it is unique in the reference 
                 genome given specified maximum mismatches
                 bowtie arguments: '-v mismatches -a -m 1 --best
  --best:        generate a best-guess read map, returning an alignment if it
                 is the one with the fewest mismatches
                 given specified maximum mismatches
                 bowtie arguments: '-v mismatches -k 1 --best'
  --bestno:      similar to --best but discard reads which have multiple 
                 equally best alignments
  --mapqual      use read quality values during mapping (bowtie -n mode with
                 default parameters)
  --colorspace:  input files are in ABI SOLiD colorspace format 
                 (suffix .csfasta and .qual)
                 NB, currently only works with single-end reads
  --ins <FLOAT>: estimate insert size range for paired-end reads as the 
                 <FLOAT> percentile of the empirical distribution 
                 (default: 0.99) (see below for more details)
  --globalerror: estimate expected per-nucleotide sequencing error rate from 
                 your read map
  --cstheta <FLOAT>: Haploid proportion of divergence between sample and 
                     reference genomes (default: 3/2 theta)
  --qual-int:    sets the --integer-quals flag in bowtie
  -Q <INT[,INT,...,INT]>: Comma-delimited list of phred-quality scores for
                          which to compute significance (Default: 40)
  -fwer <FLOAT,INT>: Filter significant SNPs using a family-wise error rate 
                     of <FLOAT> and an average read length of <INT>. This 
                     corrects the posterior probability for each genotyped 
                     locus for multiple testing of all loci that utilize the 
                     same set of NGS alignments. The Q stringency required for
                     significance is -10 log( <FLOAT> / (2 <INT> dbar) ), 
                     where <FLOAT> is the error rate (e.g. 0.01), <INT> is the
                     read length (e.g. 50), and dbar is the estimated mean 
                     number of total alignments for all reads that overlap the
                     current locus of interest (Default: 0.05,50).
  --nominal:     Instead of waiting until the whole genome is processed before
                 returning significant SNPs, this updates a file 
                 sampleID.nominal.txt with potential significant SNPs
                 concurrent with processing.
  --suffix <STR>:    Change default map file suffix to FILENAME.<STR> [Default: sam]

Bowtie mapping options
======================
  --bowtie/-btd: Specify the base directory of your bowtie installation 
                 (e.g. --btd /usr/bin/bowtie-0.12.5) (default: '')
  --keepbowtie/--kb: Do not delete the original bowtie SAM file following 
                     read organization
  --btx <INT>:   number of processors used for mapping
  --policy:      Specify Bowtie alignment policy (default: bestk)
                 Options:
                 * 'basic': 'bowtie -n <mismatches> -m 1 (Maq-like)
                 * 'all': 'bowtie -v <mismatches> -a (end-to-k all alignments)
                 * 'bestk': -v <mismatches> -k <max num alignments> --best
                 * 'bestm': -v <mismatches> -a -m <max num alignments> --best
  --kpe: Permit a maximum number of alignment mismatches for paired-end reads
  --kse: Permit a maximum number of alignment mismatches for single-end reads
  --trimreads n5,n3: For alignment, consider read substring from 1+n5...|r|-n3
  --realign/--ra: perform new Bowtie alignment, replacing existing alignments
  --integer-quals: invokes Bowtie flag of same name (default: NA)
  --btcustom:      arbitrary custom string fed directly to bowtie (default: NA)

Map file format
===============
The map file is a standard tab-delimited text file indicating how each raw 
sequence reads file should be processed by sniper. The format is organized as 
one row per sequenced lane and requires 7 columns of information row, as 
follows:

#!Run Lane SampleID       Alias RefGenome           PairedEnd Path
  1   1    myfirstgenome  foo   /path/to/ref_genome 0         /path/to/run1/
  2   5    mysecondgenome bar   /path/to/ref_genome 1         /path/to/run2/

(If you copy/paste this, make sure to convert spaces to tabs.)

In this example, the first sample was run on lane 1 and contains single-end 
reads. The second sample was sequenced on lane 5 and contains paired-end reads.
The standard mapper used with sniper.py is bowtie, so the specified reference 
is the name of the respective index file, excluding file suffix.

The first line shows the headers but is treated as a comment; 
comments are discarded by Sniper.

Raw sequence read files should contain as a substring the label indicated 
in the Lane column and placed in the /path/to/reads/directory/ directory. 

For example the first sample file may be named 's_1_sequence.txt' or 
'testing1.txt'. The second sample should have 2 files since it is 
paired-end, named 's_5_1_sequence.txt' and 's_5_2_sequence.txt'. 
Colorspace files should take the format 's_1_sequence.csfasta' 
and 's_1_sequence.qual' for the fasta and quality files, respectively. 

Additional fields may be added to the map file after the required columns 
(e.g. concentration, date sequenced, author, etc.).

To specify the prior model and theta parameter value in the map file, 
include headings "Prior" and "Theta" after "Path" and include the 
particular prior and theta values for each sample. Missing values are 
tolerated, in which case those samples will be processed with the 
default values.

Estimating insert size distribution for paired-end data
=======================================================
Two-step execution for paired-end read data:
1. sniper.py -m <map_file.txt> -r <reference_genome.fasta> --ins
2. sniper.py -m <map_file.txt> -r <reference_genome.fasta> -z

By default paired-end reads are mapped using an insert size range of 
0 to 350 nt. Users may specify custom min and max bounds using

  --minins: Specify the minimium insert size for all samples
  --maxins: Specify the maximum insert size for all samples

Alternatively use '--ins <FLOAT>' to estimate the empirical insert size 
distribution for each sample specified in the map file:

  --ins <FLOAT>: Estimate insert size distributions for all samples specified 
                 in the map file using the percentile <FLOAT>[0,1] of the 
                 insert size distribution for alignment (default: 0.99)

Note that alignments must first be generated using '-a'. 
A file 'PE_insert_size.txt' will be created in the specified output directory 
that Sniper will use by default if samples are mapped again 
(using '--realign/--ra'). 
"""%(SNIPER_VERSION)


version = """\nsniper.py, version %s
March 2011
Daniel F. Simola, Ph.D.
Laboratory of Junhyong Kim, Ph.D.
University of Pennsylvania
http://kim.bio.upenn.edu/software/sniper.shtml

Sniper (SNP Identification using Probability of Every Read) is an algorithm 
that estimates the posterior probability of a SNP compared to a reference 
genome at every position in the reference genome, utilizing single or 
paired-end NGS reads that are singly or multiply-mapped. Notably, Sniper is 
compatible with variable read length data sets (by default), and NGS fasta 
files that lack quality scores can be processed as well.

Requires Unix/Mac OS X/CYGWIN with Python 2.5 or 2.6 
(preferably 64-bit for memory allocation > 4GB)

Copyright (c) 2011, Daniel F. Simola and Junhyong Kim, University of
Pennsylvania. All Rights Reserved.

You may not use this file except in compliance with the terms of our License. 
You may obtain a copy of the License at 
http://kim.bio.upenn.edu/software/LICENSE

Unless required by applicable law or agreed to in writing, this software is 
distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, 
either express or implied.  See the License for the specific language governing
permissions and limitations under the License.

"""%(SNIPER_VERSION)


nhelps = 0; helplimit = 0
args = sys.argv[:]
argstr = ''.join(args)
ai = 1
userformat = False

if len(args)==1: sys.exit('\n'.join(help.split('\n')[1:36]))

while ai < len(args):
	arg = args[ai].strip('-').strip('--')#.lower()
	try: val = args[ai+1]
	except IndexError: val = ''
	
	if re.match('out|^o$', arg): outdir = slash(val)
	elif arg == 'path': callprefix = val
	elif re.match('^bowtie$|^btd$', arg): btdir = slash(val)
	elif re.match('^map$|^m$', arg): mapfile = val
	elif re.match('^readmapdir$|^btmapdir$|^i$', arg): SETMAPDIR = True; btmapdir = val
	elif re.match('^mapcut$|^mc$', arg): MAPCUT = True; ai-=1
	elif arg == 'fasta': readfileformat = 'fasta'; ai-=1
	elif re.match('^name$', arg): 
		zkey = val
		if zkey[len(zkey)-1] == '/': zkey = zkey[0:len(zkey)-1]
	
	elif arg == 'sge': USESGE = True; ai-=1
	elif arg == 'memsge': SGEMEM = '-l mem_free=%s'%(val)
	# elif re.match('^btprocs$|^btx$', arg): nprocs = int(val)
	elif re.match('^procs$|^x$', arg): nsubprocs = int(val)
	# elif re.match('^threads$|^p$', arg): nthreads = int(val) # deprecated
	elif re.match('^issubproc$', arg): ISSUBPROC = True; ai-=1
	elif re.match('^subchrom$', arg): xflag = 'subchrom'; ai-=1
	elif re.match('^bychrom$', arg): xflag = 'bychrom'; ai-=1
	elif re.match('^groupchrom$', arg): xflag = 'groupchrom'; ai-=1
	elif re.match('^bysampleid$|^bysample$', arg): xflag = 'bysampleid'; ai-=1
	
	elif arg == 'merge': MERGESIDS = val.split(','); REMOVEPREMERGED = False
	elif arg == 'merges': MERGESIDS = val.split(','); REMOVEPREMERGED = False; MERGESORT = True
	elif arg == 'merge+': MERGESIDS = val.split(','); REMOVEPREMERGED = True
	elif arg == 'merges+': MERGESIDS = val.split(','); REMOVEPREMERGED = True; MERGESORT = True
	elif arg == 'join': JOINPIECES = True; ai-=1
	elif arg == 'clean': CLEANFILE = val
	
	elif re.match('^ref$|^b$|^r$', arg): bfaref = val
	elif re.match('policy|^pol$', arg): alignmentPolicy = val
	elif re.match('^mcap$|^d$', arg): maxnumalign = int(val)
	elif re.match('^qual33$', arg): qualstr = '--phred33-quals'; ai-=1
	elif re.match('^trimreads$', arg): trim5,trim3 = map(int, val.split(','))
	elif re.match('^colorspace$|^C$', arg): COLORSPACE = True; ai-=1
	
	elif re.match('align|^a$', arg): RUNBOWTIE = True; ai-=1
	elif re.match('realign|^ra$', arg): REALIGN = True; ai-=1
	# elif re.match('countreads$', arg): COUNTREADS = True; ai-=1
	
	elif re.match('usesingletons|pese|sepe|^g$', arg): USESINGLETONS = True; ai-=1
	elif re.match('peonly|^pe$', arg): USESINGLETONS = False; ai-=1
	
	# mapping strategy selection
	elif re.match('^all$|^total$', arg): STRATEGY = 'all'; ai-=1
	elif re.match('^uniq$|^unique$', arg): 
		# only return an alignment if it is unique in the reference genome
		# given specified maximum mismatches
		STRATEGY = 'unique'; alignmentPolicy = 'bestm'; maxnumalign = 1; ai-=1
	elif re.match('^best$|^bestguess$', arg): 
		# return an alignment if it is the one with the fewest mismatches
		# given specified maximum mismatches; quasi-randomly break ties
		STRATEGY = 'best'; alignmentPolicy = 'bestk'; maxnumalign = 1; ai-=1
	elif re.match('^bestno$|^bestnoguess$', arg):
		# return an alignment if it is the one with the fewest mismatches
		# but discard read if multiple equally best alignments exist
		STRATEGY = 'bestno'; alignmentPolicy = 'bestk'; maxnumalign = 2; ai-=1
		# technically we map allowing 2 alignments and then SNP call with
		# the unique maps only
	elif re.match('^mapqual$', arg): MAPWITHQUALITY = True; ai-=1
	
	# mismatches
	elif arg == 'k': mismatches = {'SE':int(val), 'PE':int(val)}
	elif arg == 'kpe': mismatches['PE'] = int(val)
	elif arg == 'kse': mismatches['SE'] = int(val)
	
	elif arg == 'ins': GETINSERTSIZE = True; insstat = str(val); RUNBOWTIE = True; mininsCLI=0; maxinsCLI=1000
	elif re.match('minins', arg): mininsCLI = int(val); 
	elif re.match('maxins', arg): maxinsCLI = int(val)
	
	# read organization
	elif re.match('organize|^O$', arg): ORGANIZEREADS = True; ai-=1
	elif re.match('reorganize|^rO$', arg): ORGANIZEREADS = True; REORGANIZEREADS = True; RESORT = True; KEEPBOWTIE = True; ai-=1
	elif arg == 'combo': PESECOMBO = True; ai-=1
	elif arg == 'splitmap': PESECOMBO = False; ai-=1
	elif re.match('exorganize|^exO$', arg): ORGANIZEREADS = True; EXORGANIZE = True; ai-=1
	elif re.match('^keepbowtie$|^kb$', arg): KEEPBOWTIE = True; ai-=1
	elif re.match('^clearbowtie$|^cb$', arg): CLEARBOWTIE = True; ORGANIZEREADS = True; ai-=1
	elif arg == 'skiphist': SKIPHIST = True; ai-=1
	elif arg == 'index': INDEXMAPFILES = True; ai-=1
	elif arg == 'reindex': INDEXMAPFILES = True; RESORT = True; ai-=1
	elif arg == 'indexuni': INDEXUNI = True; ai-=1
	elif arg == 'reindexuni': INDEXUNI = True; RESORT = True;ai-=1
	elif arg == 'indexdeg': INDEXDEG = True; ai-=1
	elif arg == 'reindexdeg': INDEXDEG = True; RESORT = True;ai-=1
	elif arg == 'nosort': INDEXONLY = True; ai-=1
	elif arg == 'link': LINKUNIQUE = True; ai-= 1
	
	elif arg == 'stream': DEGREADSTREAM = True; ai -= 1
	elif arg == 'nostream': DEGREADSTREAM = False; ai -= 1
	elif arg == 'buffer': bufferlimit = int(val)*100
	
	# SNP calling
	elif re.match('^snp$|^s$', arg): SNPCALLING = True; IDENTIFYSIGNIFICANTS = True; ai-=1
	elif re.match('^snponly$|^so$', arg): SNPCALLING = True; IDENTIFYSIGNIFICANTS = True; SAVENOMINAL = True; QvalueRange = [float(val)]
	elif arg == 'sig': IDENTIFYSIGNIFICANTS = True; ai-=1
	elif arg == 'nosig': IDENTIFYSIGNIFICANTS = False; ai-=1
	elif re.match('^redosnp$|^resnp$|^rs$', arg): 
		SNPCALLING = True; REDOSNPCALLING = True; IDENTIFYSIGNIFICANTS = True; REDOFINALSNPCALLING = True; ai-=1
	elif re.match('^redosnponly$|^resnponly$|^rso$', arg): 
		SNPCALLING = True; REDOSNPCALLING = True; IDENTIFYSIGNIFICANTS = True
		REDOFINALSNPCALLING = True; SAVENOMINAL = True; QvalueRange = [float(val)]
	elif re.match('^redocorrection$|^rc$|^resig$', arg): IDENTIFYSIGNIFICANTS = True; REDOFINALSNPCALLING = True; ai-=1
	elif re.match('^redocorrectionpost$|^rcf$', arg): SNPCALLING = True; IDENTIFYSIGNIFICANTS = True; REDOFINALSNPCALLING = True; POSTPROCESSING = True; ai-=1
	elif re.match('^loadinpieces$|^lip$', arg): LIPieces = int(val)
	
	# Important parameters
	elif re.match('^maxdepth$|^mx$', arg): maxdepth = int(val)
	elif re.match('^readlen$|^rl$', arg): maxreadlen = int(val)#; READLENBYUSER = True
	elif re.match('^lambda$|^l$', arg): Lambda = float(val)
	elif re.match('^expphred$|^e$', arg): expPhred = int(val)
	elif re.match('^globalerror$|^ge$', arg): ESTIMATEGLOBALERROR = True; ai-=1
	elif re.match('^prior$', arg): prior = val
	elif re.match('^theta$', arg): theta = float(val)
	elif re.match('^colorspacetheta$|^cstheta$', arg): colorspacetheta = float(val)
	elif arg == 'qual-int': SETINTQUALS = True; ai-=1
	elif re.match('^stringency$|^Q$', arg): QvalueRange = map(int, val.split(',')); ADAPTIVEQ = False
	elif re.match('^fwer$', arg): ADAPTIVEQ,readlength = map(float,val.split(','))
	elif arg == '^nofwer$': ADAPTIVEQ = False; ai-=1
	
	# restricting genotyping for a map file with multiple samples
	elif re.match('^restrict|^rest$', arg): boundsfile = val
	elif re.match('^shufflerest$', arg): SHUFFLEBOUNDS = True; ai -= 1
	elif re.match('^sampleid$|^id$', arg): singlesampleid = [val]
	elif re.match('^sampleids$|^ids$', arg): singlesampleid = map(lambda x: x.strip(), val.split(','))
	elif re.match('^subset|^sub$', arg): LOWSID,HIGHSID = map(int,val.split(','))
	
	# postprocessing settings
	elif re.match('^maxalpha$|^mxa$', arg): MAXALPHA = float(val)
	elif re.match('^minalpha$|^mna$', arg): MINALPHA = float(val)
	elif re.match('^postfilter$', arg): POSTFILTER = True; ai -=1
	elif re.match('^pfcutoff$', arg): DDEGCUTOFF = float(val)
	elif re.match('^repfilter$|^rf$', arg): REDUNDANCYFILTER = val
	
	# other settings
	elif re.match('redoall|^z$', arg): REDOALL = True; ai-=1
	elif re.match('^diagnostics$|^diag$', arg): DIAGNOSTICS = True; ai-=1
	elif re.match('^noass$|^na$', arg): SAVEASSEMBLY = False; ai-=1
	elif re.match('^saveass$|^sa$', arg): SAVEASSEMBLY = True; ai-=1
	elif re.match('^srefs$|^sr$', arg): specialReferenceFiles = val # for --saveassembly only
	
	# elif re.match('^nominal$|^noms$', arg): SAVENOMINAL = True; ai -=1
	
	elif arg == 'suffix': mapsuffix = '.%s'%(val)
	
	elif re.match('^help|h$', arg.lower()): sys.exit(help)
	elif re.match('^version|v$', arg.lower()): sys.exit(version)
	
	else:
		astr = "=> The specified argument \""+arg+"\" does not parse. Try --help for more information."
		sys.exit(astr)
	ai += 2

# -------------------------------------------------------------------

# if nsubprocs > 1: xflag = 'bychrom'

if ISSUBPROC: pipeit('\n'+' '.join(args)+'\n',1)

# contingency of flags
# --------------------
if REDOALL:
	RUNBOWTIE = REALIGN = True
	ORGANIZEREADS = REORGANIZEREADS = True
	INDEXMAPFILES = True; RESORT = True
	SNPCALLING = REDOSNPCALLING = REDOFINALSNPCALLING = True
	IDENTIFYSIGNIFICANTS = True
	DIAGNOSTICS = True
	SAVEASSEMBLY = True

# if ORGANIZEREADS and not bfaref: 
# 	sys.exit('Erorr: reference fasta file must be provided with -O using --ref/-r <REF.FASTA>.')

if SETINTQUALS:
	qualstr += ' --integer-quals'

# declaration of directories
# --------------------------
zdir = outdir+zkey+'/'
if not SETMAPDIR: btmapdir = zdir+'read_maps_all'
unmapdir = zdir+'unmapped_reads'
pmadir = zdir+'alignment_mismatches'
if STRATEGY=='unique':
	if not SETMAPDIR: btmapdir = zdir+'read_maps_uni'
	unmapdir = zdir+'unmapped_reads_uni'
	pmadir = zdir+'alignment_mismatches_uni'
if STRATEGY=='best':
	if not SETMAPDIR: btmapdir = zdir+'read_maps_best'
	unmapdir = zdir+'unmapped_reads_best'
	pmadir = zdir+'alignment_mismatches_best'
if STRATEGY=='bestno':
	if not SETMAPDIR: btmapdir = zdir+'read_maps_bestno'
	unmapdir = zdir+'unmapped_reads_bestno'
	pmadir = zdir+'alignment_mismatches_bestno'

if MAPWITHQUALITY: 
	btmapdir += 'q'
	unmapdir += 'q'
	pmadir += 'q'

btmapdir += '/'
unmapdir += '/'
pmadir += '/'
sorttmpdir = zdir+'tmp/'

zinsdir = zdir+'PE_insert_size/'

tmpstr = 'ALL'
if STRATEGY=='unique': tmpstr = 'UNI'
elif STRATEGY=='best': tmpstr = 'BEST'
elif STRATEGY=='bestno': tmpstr = 'BESTNO'
if MAPWITHQUALITY: tmpstr += 'Q'
resdir = zdir+'results-L%s,E%s,%s,%s/'%(Lambda, expPhred, USESINGLETONS and 'PESE' or 'PE', tmpstr)

findir = resdir+'assembled_genomes/'
snpdir = resdir+'snps/'
phylodir = resdir+'SNPstrings/'
tmpdir = resdir+'tmp/'


# file suffix depending on map type
ubamap = '-%s'%(STRATEGY)
if MAPWITHQUALITY: ubamap += 'q'

# create new directories
# ----------------------
createdir(outdir)
createdir(zdir)
createdir(btmapdir)
createdir(sorttmpdir)

# save a log file with the user's input arguments
logfile = zdir+'log_file_%s.txt'%(tmpstr)
iomethod = 'w'
if os.access(logfile, os.F_OK): iomethod = 'a'
fh = open(logfile, iomethod)
print >> fh, ' '.join(args)
fh.close()


# add sort option here also
if CLEANFILE:
	curscaf = None
	seen = {} # postitions
	fh = open(CLEANFILE)
	CLEANFILETARGET = CLEANFILE+'.cleaned'
	fho = open(CLEANFILETARGET, 'w')
	for line in fh:
		row = line[:-1].split('\t')
		chrm,pos = row[:2]
		if curscaf != chrm: 
			curscaf = chrm
			seen = {}
			# print >> sys.stderr, 'scaffold', chrm
			
		if pos not in seen: fho.write(line)
		# else: print >> sys.stderr, 'SEEN:', line[:-1]
		seen[pos] = None
	fh.close()
	fho.close()
	sys.exit('Done')


# parse map file and group replicates
# ----------------------------------------
seqRuns = {} # keyed by sampleID
ct = 0
totalsamples = 0
ssPrior = {} # sample-specific priors
ssTheta = {} # sample-specific theta

if not mapfile: pipeit('No map file provided.',1)

if not mapfile and not singlesampleid:
	# sys.exit('Map file must be provided with --map/-m <MAPFILE>. Use sniper.py --help for more details.')
	sys.exit('Either --map/-m <MAPFILE> or --id <sample1,...,sampleN> must be provided. Use sniper.py --help for more details.')
elif not mapfile:
	seqRuns = dict(map(lambda x: (x,{(0,0):[0, 0, 0]}) , singlesampleid))
	ssPrior = dict(map(lambda x: (x,prior), singlesampleid))
	ssTheta = dict(map(lambda x: (x,theta), singlesampleid))
else:
	# parse map file and group replicates
	# ----------------------------------------
	fh = open(mapfile)
	runs = []; mapd = []; head = []; postcombos = []
	ct = 0
	for line in fh:
		# store header
		if line[0:2] == '#!': head = map(lambda x: x.strip(), line[2:-1].split('\t'))
	
		if line[0] != '#':
			# trim past any place a # hash is present
			cline = line[:-1]
			try: cline = cline[:cline.index('#')]
			except ValueError: pass
			tmp = cline.split('\t')
		
			if len(tmp) > 1:
				runs += [tmp.pop(0)]
				mapd += [tmp]
	
		ct += 1
		
	prioridx = None; thetaidx = None
	if 'Prior' in head: prioridx = head.index('Prior') - 1
	if 'Theta' in head: thetaidx = head.index('Theta') - 1
	
	for i in range(len(runs)):
		ct += 1
		if MAPCUT and (ct < LOWSID or ct > HIGHSID): continue
		totalsamples += 1
		run,lane,sampleID,alias,refGenome,paired,path = runs[i], mapd[i][0], mapd[i][1], mapd[i][2], mapd[i][3], int(mapd[i][4]), mapd[i][5]
		path = path.strip()
		# print 'loading', path, 'vs', mapd[i]
		if path == '': path = '/'
		else: path = slash(path)
		
		# make absolute path if local directory specified
		if path[0] != '/' or path == '/':
			abspath = slash(os.getcwd())
			# if this does not end at the current local path, add path
			if slash(abspath.split('/')[-2]) != path: abspath += path
			path = abspath
		
			path = slash(path)
	
		# sample-specific prior probability
		ssPrior[sampleID] = prior
		try:
			if prioridx: ssPrior[sampleID] = mapd[i][prioridx]
		except IndexError: pipeit('Cannot read prior for sample %s... defaulting to %s'%(sampleID,prior),1)
		except ValueError: pipeit('Cannot read prior for sample %s... defaulting to %s'%(sampleID,prior),1)
	
		# sample-specific theta
		ssTheta[sampleID] = theta
		try:
			if thetaidx: ssTheta[sampleID] = float(mapd[i][thetaidx])
		except IndexError: pipeit('Cannot read theta for sample %s... defaulting to %s'%(sampleID,theta),1)
		except ValueError: pipeit('Cannot read theta for sample %s... defaulting to %s'%(sampleID,theta),1)
	
	
		try:
			seqRuns[sampleID][(run,lane)] = [refGenome, paired, path]
		except KeyError:
			seqRuns[sampleID] = {(run,lane):[refGenome, paired, path]}

	if nsubprocs == 1:
		pipeit('- Read %s samples from map file'%(totalsamples),1)

# ---------------------------------------------------------
sortseqruns = seqRuns.keys()
sortseqruns.sort()

if not MAPCUT:
	keeps = dict(map(lambda x: (x,None), sortseqruns[LOWSID:HIGHSID]))
	for sid in sortseqruns:
		if sid not in keeps:
			del seqRuns[sid]
			

if len(singlesampleid):
	sr2 = {}
	for sid in singlesampleid:
		if sid in seqRuns:
			bak = seqRuns[sid]
			sr2[sid] = bak
		else: print 'Warning: sample %s not found in map file.'%(sid)
	seqRuns = sr2

sortseqruns = seqRuns.keys()
sortseqruns.sort()

if LOWSID or HIGHSID: print '- Processing samples %s'%(sortseqruns[LOWSID:HIGHSID])
else: print '- Processing %s sampleIDs'%(len(sortseqruns))
# ---------------------------------------------------------

# load bounds file
boundslabel = ''
seqbounds = {}
if boundsfile and os.access(boundsfile, os.F_OK):
	d,r,c = readTable(boundsfile, header=False)
	for i in range(len(d)):
		tmp = []
		try: tmp += [intna(d[i][0])]
		except IndexError: tmp += [nastr]
		except ValueError: tmp += [nastr]
		try: tmp += [intna(d[i][1])]
		except IndexError: tmp += [nastr]
		except ValueError: tmp += [nastr]
		try: tmp += [d[i][2]] # flag - + or -
		except IndexError: tmp += ['+']
		keyname = r[i].strip()
		try: seqbounds[keyname] += [tmp]
		except KeyError: seqbounds[keyname] = [tmp]
		
	parts = boundsfile.split('/')
	crap, boundslabel, crap = re.split('(.+)\..+$', parts[len(parts)-1])
	boundslabel = '.'+boundslabel
	
elif boundsfile:
	# this is a string formatted in a few possible ways
	# ex1: chr1[.chr2. ... chrx]
	# ex2: chr1,500,1000,+
	# ex3: chr1,500,NA,+
	# ex4: chr1
	
	splt = boundsfile.split(multichromtoken)
	if len(splt) > 1:
		# then we have ex1
		for loc in splt:
			seqbounds[loc] = [(nastr, nastr, '+')]
		boundslabel = '.%s--%s'%(splt[0],splt[-1])
	else:
		# ex 2 or 3, or 4
		splt = boundsfile.split(',')
		if len(splt) == 1:
			loc = splt[0]
			seqbounds[loc] = [(nastr, nastr, '+')]
			# boundslabel = '.%s,%s,%s,%s'%(loc,nastr,nastr,'+')
			boundslabel = '.%s'%(loc)
		elif len(splt) == 3:
			loc, low, high = map(lambda x: x.strip(' '), splt)
			seqbounds[loc] = [(intna(low), intna(high), '+')]
			boundslabel = '.%s,%s,%s,%s'%(loc,low,high,'+')
		elif len(splt) == 4:
			loc, low, high,flag = map(lambda x: x.strip(' '), splt)
			seqbounds[loc] = [(intna(low), intna(high), flag)]
			boundslabel = '.%s,%s,%s,%s'%(loc,low,high,flag)
		else:
			sys.exit('Restriction flag not specified properly: '%(boundsfile))



# merge motther files if split
# ----------------------------
if MERGESIDS:
	for sid in MERGESIDS:
		filesToMerge = getFiles(snpdir, include='%s.mother'%(sid), exclude='%s.mother%s.txt'%(sid,boundslabel))
		filesToMerge.sort()
		if not len(filesToMerge): pipeit('Warning: no files for %s found to merge. SKIP.'%(sid))
		mergedFile = snpdir+'%s.mother%s.txt'%(sid,boundslabel)
		fh = open(mergedFile,'w'); fh.close()
		
		if MERGESORT:
			# first sort the contents of each partial file
			pipeit('- Sorting %s partial files...'%(len(filesToMerge)),1)
			for fx in xrange(len(filesToMerge)):
				pipeit('  %s) %s'%(fx+1, filesToMerge[fx].split('/')[-1]),1)
				# pipeit('  %s) %s'%(fx+1, '/'.join(filesToMerge[fx].split('/')[-1])),1)
				os.system('sort -T "%s" -k 1,1 -k 2,2n "%s" > "%s"'%(sorttmpdir, filesToMerge[fx], filesToMerge[fx]+'.sort'))
			
			pipeit('- Merging %s files (in above order) into %s'%(len(filesToMerge),mergedFile),1)
			for afile in filesToMerge:
				os.system('cat "%s" >> "%s"'%(afile+'.sort',mergedFile))
			
			# remove originals
			if REMOVEPREMERGED:
				for afile in filesToMerge: os.system('rm -f "%s"'%(afile)); os.system('rm -f "%s"'%(afile+'.sort'))
				
		else:
			pipeit('- Merging %s files (in order) into %s'%(len(filesToMerge),mergedFile),1)
			for fx in xrange(len(filesToMerge)): pipeit('  %s) %s'%(fx+1,filesToMerge[fx]),1)
			for afile in filesToMerge:
				os.system('cat "%s" >> "%s"'%(afile,mergedFile))
			
			# remove originals
			if REMOVEPREMERGED:
				for afile in filesToMerge: os.system('rm -f "%s"'%(afile))
	sys.exit('Done merging sampleIDs: %s'%(','.join(MERGESIDS)))

# DONE INITIALIZATION OF VARIABLES


# =========================================================
# Part I: Align reads to reference genome
# =========================================================
if RUNBOWTIE or REALIGN:
	print '\nGENERATING READ MAPS USING BOWTIE...'
	
	createdir(zinsdir)
	createdir(unmapdir)

	parts = mapfile.split('/')
	peinsfile = zinsdir+'PE_insertsize_table-%s'%(parts[len(parts)-1])
	peinsize = {} # dict specifying PE insert size statistics, keyed by sampleID
	if os.access(peinsfile, os.F_OK):
		d,r,c = readTable(peinsfile, header=True)
		for i in range(len(d)):
			peinsize[r[i]] = {}
			for j in range(len(c)):
				peinsize[r[i]][c[j]] = (d[i][j] != '' and floatna(d[i][j])) or nastr
	
	
	# call bowtie on all samples in mapfile
	for sampleID in sorted(seqRuns.keys()):
		
		# set parm str each time here because insert size may be sample specific
		try:
			minins1 = peinsize[sampleID]['min'+insstat]
			maxins1 = peinsize[sampleID]['max'+insstat]
			if minins1 not in [nastr, '']: minins = minins1
			if maxins1 not in [nastr, '']: maxins = maxins1
		except KeyError: pass
		
		if mininsCLI: minins = mininsCLI
		if maxinsCLI: maxins = maxinsCLI
		
		print '- ID %s: if paired-end, min/max insert size = [%s,%s]'%(sampleID, minins, maxins)
		
		parmstr = '-t --pairtries 300 %s '%(qualstr)
		sinparmstr = '-t %s '%(qualstr)
		if readfileformat == 'fasta': parmstr += ' -f '; sinparmstr += ' -f '
		
		if trim5: 
			astr = '--trim5 %s '%(trim5)
			parmstr += astr
			sinparmstr += astr
		
		if trim3: 
			astr = '--trim3 %s '%(trim3)
			parmstr += astr
			sinparmstr += astr
		
		
		# APPLY MISMATCHES
		# default mapping is v mode which is quality-blind
		mmstr = ' -v %s '%(mismatches['PE'])
		sinmmstr = ' -v %s '%(mismatches['SE'])
		if MAPWITHQUALITY:
			# use quality values for mapping
			# current is default seed length -l and max qual -e
			mmstr = ' -n %s '%(mismatches['PE'])
			sinmmstr = ' -n %s '%(mismatches['SE'])
		
		parmstr += mmstr
		sinparmstr += sinmmstr
		
		
		# APPLY STRATEGY
		if alignmentPolicy == 'basic':
			# for the maq-like policy
			parmstr += '--minins %s --maxins %s -m 1 --best'%(minins, maxins)
			sinparmstr += ' -m 1' # for singleton samples
		
		elif alignmentPolicy == 'all':
			# end-to-k policy (-v) with all valid alignments reported (-a)
			parmstr += '--minins %s --maxins %s -a'%(minins, maxins)
			sinparmstr += '-a' # for singleton samples
		
		elif alignmentPolicy == 'bestk':
			parmstr += '--minins %s --maxins %s -k %s --best'%(minins, maxins, maxnumalign)
			sinparmstr += '-k %s --best'%(maxnumalign) # for singleton samples
		
		elif alignmentPolicy == 'bestm':
			parmstr += '--minins %s --maxins %s -a -m %s --best'%(minins, maxins, maxnumalign)
			sinparmstr += '-a -m %s --best'%(maxnumalign) # for singleton samples
		
		elif alignmentPolicy == 'maq-bestm':
			parmstr += '--minins %s --maxins %s -a -m %s --best'%(minins, maxins, maxnumalign)
			sinparmstr += '-a -m %s --best'%(maxnumalign) # for singleton samples
		
		elif alignmentPolicy == 'bestm-stratum':
			parmstr += '--minins %s --maxins %s -a -m %s --best --strata'%(minins, maxins, maxnumalign)
			sinparmstr += '-a -m %s --best --strata'%(maxnumalign) # for singleton samples
		
		elif alignmentPolicy == 'best-stratum':
			parmstr += '--minins %s --maxins %s -a --best --strata'%(minins, maxins)
			sinparmstr += '-a --best --strata' # for singleton samples
		
		# output in sam format
		parmstr += ' --sam'; sinparmstr += ' --sam'
		
		# add custom options
		parmstr += ' %s'%(btcustom)
		sinparmstr += ' %s'%(btcustom)
		
		for run,lane in seqRuns[sampleID].keys():
			refGenome,paired,path = seqRuns[sampleID][(run,lane)]
			UID = '%s-%s-%s'%(sampleID,run,lane)
			
			# find the file that is indicated in the map file
			fastqFiles = sorted(getFiles(path, include=lane))
			
			if paired:
				# remove extraneous files that do not have the _1 or _2 in them
				fastqFiles = filter(lambda x: ('_1' in x or '_2' in x) and True, fastqFiles)
				
				if len(fastqFiles) > 2:
					print >> sys.stderr, 'Warning: Found more than 2 matching read files. Trying default s_lane_x_sequence.fastq format.'
					tmpfiles = ['s_%s_1_sequence'%(lane), 's_%s_2_sequence'%(lane)]
					fastqFiles = sorted(getFiles(path, include=tmpfiles, exclude='pdf'))
				
				if len(fastqFiles) > 2:
					print >> sys.stderr, 'Warning: Found more than 2 matching read files. Trying default s_lane_x_sequence.txt format.'
					fastqFiles = [path+'s_%s_1_sequence.txt'%(lane), path+'s_%s_2_sequence.txt'%(lane)]
				elif len(fastqFiles) < 2:
					print >> sys.stderr, 'Check read file names or path. Did not find 2 files for sample %s in path "%s": %s'%(lane, path, fastqFiles)
					continue
				
				if not os.access(fastqFiles[0], os.F_OK) or not os.access(fastqFiles[1], os.F_OK): 
					print >> sys.stderr, 'Warning: could not find some raw read files. Check paths.'
					continue
				
				
				call = '%sbowtie %s -p %s "%s" -1 "%s" -2 "%s" "%s%s.sam" --un "%s%s_unfq.fq" 2>> "%s"'%(btdir,parmstr, nsubprocs,  refGenome, fastqFiles[0], fastqFiles[1], btmapdir, UID, btmapdir, UID, zdir+'bowtie_mapping.log')
				
				# call = '%sbowtie %s -p %s "%s" -1 "%ss_%s_1_sequence.txt" -2 "%ss_%s_2_sequence.txt" "%s%s.map" --un "%s%s_unfq.fq"'%(btdir,parmstr,nprocs, refGenome,path,lane,path,lane,btmapdir,UID,btmapdir,UID)
				
				pipeit('- Mapping paired end data for %s...'%(UID),1)
				printList([call], file=zdir+'bowtie_mapping.log', ioMethod='a')
				
				if not DRYRUN and (not os.access(btmapdir+UID+'.sam', os.F_OK) or REALIGN): 
					os.system(call)
				
				# set this up to process the remaining singletons as well, and then merge the output files...
				# outfile is _unfq.fq
				if 1: # USESINGLETONS
					pipeit('- Mapping left-over singleton data for %s...'%(UID),1)
					for lane in [1,2]:
						call2 = '%sbowtie %s -p %s "%s" "%s" "%s" --un "%s" 2>> "%s"'%(btdir, parmstr, nsubprocs, refGenome, btmapdir+UID+'_unfq_%s.fq'%(lane), btmapdir+UID+'_singles_%s.sam'%(lane), unmapdir+UID+'_unmapped_%s.fq'%(lane), zdir+'bowtie_mapping.log')
						printList([call2], file=zdir+'bowtie_mapping.log', ioMethod='a')
						if not DRYRUN and (not os.access(btmapdir+UID+'_singles_%s.sam'%(lane), os.F_OK) or REALIGN):
							os.system(call2)
						
						
						# remove the input fq file, contents of which are in singles_map and unmapped
						os.system('rm -f "%s"'%(btmapdir+UID+'_unfq_%s.fq'%(lane)))
						
				else:
					# rename the --un files from the PE run as unmapped and move them into the unmapped folder
					for lane in [1,2]:
						call = 'mv %s %s'%(btmapdir+UID+'_unfq_%s.fq'%(lane), unmapdir+UID+'_unmapped_%s.fq'%(lane))
						os.system(call)
			else:
				if not COLORSPACE:
					if len(fastqFiles) > 2:
						print >> sys.stderr, 'Warning: Found more than 2 matching read files. Using default s_lane_x_sequence.txt format.'
						fastqFiles = [path+'s_%s_1_sequence.txt'%(lane), path+'s_%s_2_sequence.txt'%(lane)]
					elif len(fastqFiles) != 1:
						print >> sys.stderr, 'Check read file names or path. Did not find 1 file for sample %s in dir "%s": %s'%(lane, path, fastqFiles)
						
				# len(fastqFiles) != 1:
				# 	sys.exit('Check read file names. Did not find 1 file for sample %s: %s'%(lane, fastqFiles))
				
				call = '%sbowtie %s -p %s "%s" "%s" "%s%s.sam" --un "%s%s_unmapped.fq" 2>> "%s"'%(btdir, sinparmstr, nsubprocs, refGenome, fastqFiles[0], btmapdir, UID, unmapdir, UID, zdir+'bowtie_mapping.log')
				
				# call = '%sbowtie %s -p %s "%s" "%ss_%s_sequence.txt" "%s%s.map" --un "%s%s_unmapped.fq"'%(btdir, sinparmstr, nprocs, refGenome, path, lane, btmapdir, UID, unmapdir, UID)
				if COLORSPACE:
					# trim any .txt files
					fastqFiles = filter(lambda x: x.count('.csfasta') or x.count('.qual'), fastqFiles)
					
					if len(fastqFiles) != 2:
						print >> sys.stderr, 'Check read file names or path. Did not find 2 files (.csfasta and .qual) for sample %s in file "%s": %s'%(lane,path, fastqFiles)
					
					# call = '%sbowtie %s -C -p %s "%s" -f "%ss_%s_sequence.csfasta" -Q "%ss_%s_sequence.qual" "%s%s.map" --un "%s%s_unmapped.fq" --snpfrac %s'%(btdir, sinparmstr, nprocs, refGenome, path, lane, path,lane, btmapdir, UID, unmapdir, UID, colorspacetheta)
					call = '%sbowtie %s -C -p %s "%s" -f "%s" -Q "%s" "%s%s.sam" --un "%s%s_unmapped.fq" --snpfrac %s 2>> "%s"'%(btdir, sinparmstr, nsubprocs, refGenome, fastqFiles[0], fastqFiles[1], btmapdir, UID, unmapdir, UID, colorspacetheta, zdir+'bowtie_mapping.log')
					
				
				if not DRYRUN and (not os.access(btmapdir+UID+'.sam', os.F_OK) or REALIGN): 
					pipeit('- Assembling original singleton data for %s...'%(UID),1)
					printList([call], file=zdir+'bowtie_mapping.log', ioMethod='a')
					os.system(call)
			
		print '=> DONE', sampleID
	print 'DONE READ ALIGNMENT FOR ALL SAMPLES'
	print

# =========================================================
# Part IB: Assay raw read coverage / Insert size stats
# =========================================================
# if COUNTREADS:
# 	print '- Loading reference genome...'
# 	G = readFasta(bfaref)
# 	# chromosome total lengths (bp)
# 	chromlengths = dict( map(lambda k: (k, len(G[k]['seq'])), G) )
# 	master_chroms = chromlengths.keys()
# 	
# 	readbphead = ['SampleID', 'Run', 'Lane', 'PE', 'Reads', 'Bases', 'Coverage']
# 	readbptab = []
# 	
# 	print 'Computing number of reads for samples...'
# 	for sampleID in sorted(seqRuns.keys()):
# 		for run,lane in seqRuns[sampleID].keys():
# 			refGenome,paired,path = seqRuns[sampleID][(run,lane)]
# 			# count up reads and bases
# 			if paired:
# 				for pe in [1]: # don't need the second because it has same number of reads
# 					call = "wc -l %s > %s"%(path+'s_'+lane+'_'+str(pe)+'_sequence.txt', outdir+'tmpread.txt')
# 					print 'call', call
# 					os.system(call)
# 					line = readList(outdir+'tmpread.txt')
# 					nreads, crap = line[0].split(' ')
# 					nreads = float(nreads)/2.0 # each read takes up 2 rows in file
# 					row = [sampleID, run, lane, pe,'%.5e'%(float(nreads)), '%.5e'%(float(nreads)*readlen), '%.2f'%(float(nreads)*readlen/float(2*sum(chromlengths.values())))]
# 					readbptab += [row]
# 				os.system('rm "%s"'%(outdir+'tmpread.txt'))
# 			else:
# 				print 'have not implemented single end raw read counting yet...'
# 	printFormattedTable(readbptab, readbphead, file=zdir+'Raw reads summary-%s.txt'%(zkey))
# 	
# 	sys.exit('Done counting raw reads. Exiting...')
# print

if GETINSERTSIZE:
	# save an insert size histogram for each sampleID and a summary table providing .95, .99, and min/max stats
	# this only works in bowtie saved output as sam maps
	
	print 'Computing PE insert size distributions...'
	peinsize = []
	
	for sampleID in sorted(seqRuns.keys()):
		insvalues = []
		
		for run,lane in seqRuns[sampleID].keys():
			refGenome,paired,path = seqRuns[sampleID][(run,lane)]
			UID = '%s-%s-%s'%(sampleID,run,lane)
			
			# only used for paired end reads
			if paired:
				# get data from the sam/bam file
				fh = open(btmapdir+UID+'.sam')
				for line in fh.readlines():
					if line[0]=='@': continue
					
					row = line.strip('\n').split('\t')
					# print 'row has', len(row)
					# print row
					insvalue = abs(int(row[8]))
					if insvalue != 0: insvalues += [insvalue]
				fh.close()
		
		insvalues = filter(lambda x: x!=nastr and True, insvalues)
		
		if len(insvalues) == 0: print '- Unable to read read map for %s'%(sampleID)
		else:
			print '- %s has %s values'%(sampleID, len(insvalues))
			print '  saving histogram...'
			hist = histogram(insvalues, nbins=25, yfreq=True)
			printTable(hist, ['Size', 'Count'], file=zinsdir+sampleID+' histogram.txt')
		
			print '  calculating statistics...'
			ansd = sd(insvalues)
			meanins = avg(insvalues)
			sdL = meanins - ansd; sdR = meanins + ansd
		
			ninefiveL = percentile(insvalues, .025)
			ninefiveR = percentile(insvalues, .975)
		
			ninenineL = percentile(insvalues, .005)
			ninenineR = percentile(insvalues, .995)
		
			mn = min(insvalues); mx = max(insvalues)
		
			perow = [sampleID, meanins, sdL, sdR, ninefiveL, ninefiveR, ninenineL, ninenineR, mn, mx]
		
			print perow
		
			peinsize += [perow]
	
	header = ['ID', 'Mean', 'minSD', 'maxSD', 'min.95', 'max.95', 'min.99', 'max.99', 'min', 'max']
	printTable(peinsize, header, file=peinsfile)
	
	print '=> Finished estimating insert size distributions.\n\nRerun sniper.py using the -ra flag to realign reads using these estimated insert size bounds.\n'
	sys.exit()

# ==========================================================
# Part II: generate final read files grouped by multiplicity
# ==========================================================
if ORGANIZEREADS:
	pipeit('\nREAD ORGANIZATION',1)
	
	createdir(pmadir)
	
	# ------------------------------------------
	# if --bestno map type is specified, we want to filter the SE and PE degenerate maps
	# to remove reads with multiple equally good alignments
	# then concatenate these "unique" alignments to the SE and PE unique map files
	# this is done right before saving to SE/PE.unique/degenerate.map files below
	# ------------------------------------------
	
	# touch new summary tables
	pmafile = zdir+'Summary-Perfect vs mismatch alignments%s.txt'%(ubamap)
	if not os.access(pmafile, os.F_OK) or REORGANIZEREADS:
		header=['Sample', 'Perfect alignments', 'Mismatched alignments', 'Total alignments', 'Frac perfect', 'Frac mismatch']
		record('\t'.join(header), pmafile, 'w')
	
	almfile = zdir+'Summary-alignment multiplicity%s.txt'%(ubamap)
	if not os.access(almfile, os.F_OK) or REORGANIZEREADS:
		record('Sample: SE and PE multiplicity distributions (alignments, reads)', almfile, 'w')
	
	crdfile = zdir+'Summary-chromosome read distribution.txt'
	if not os.access(crdfile, os.F_OK) or REORGANIZEREADS:
		record('SampleID\tChrom\tUnique SE\tUnique PE\tNon-unique SE\tNon-unique PE', crdfile, 'w')
		
	currentsample = 0
	for sampleID in sorted(seqRuns.keys()):
		
		sys.stdout.write('- Looking at sample: %s\n'%(sampleID)); sys.stdout.flush()
		
		if (not REORGANIZEREADS) and os.access(btmapdir+sampleID+'.SE.unique.map', os.F_OK):
			sys.stdout.write('  Already processed.\n'); sys.stdout.flush()
			continue
		
		# record read multiplicity - number of times each read appears
		# ------------------------------------------------------------
		pipeit('- Assessing read multiplicity...',1)
		
		# per sampleID variables of interest
		npmms = 0; nrows = 0 # count up total rows and total rows with mismatches
		totreads = {'SE':0, 'PE':0, 'Total':0}
		totaln = {'SE':0, 'PE':0, 'Total':0}
		
		sidHist = {'SE':{}, 'PE':{}, 'Total':{}} # aggregate multiplicity histogram
		
		ulinesat = {'PE':{}, 'SE':{}}
		dlinesat = {'PE':{}, 'SE':{}}
		isizes = []
		
		# for each sampleID save new SAM-format read files
		# Partition reads into 2 multiplicity groups
		# funi = {'SE':open(btmapdir+sampleID+'.SE.unique.map', 'w'), 'PE':open(btmapdir+sampleID+'.PE.unique.map', 'w')}
		# fdeg = {'SE':open(btmapdir+sampleID+'.SE.degenerate.map', 'w'), 'PE':open(btmapdir+sampleID+'.PE.degenerate.map', 'w')}
		funi = {}
		fdeg = {}
		
		if PESECOMBO:
			stmp = open(btmapdir+sampleID+'.unique.map', 'w')
			funi = {'SE':stmp, 'PE':stmp}
			stmp = open(btmapdir+sampleID+'.degenerate.map', 'w')
			fdeg = {'SE':stmp, 'PE':stmp}
		else:
			# Partition reads into 2 multiplicity groups
			funi = {'SE':open(btmapdir+sampleID+'.SE.unique.map', 'w'), 'PE':open(btmapdir+sampleID+'.PE.unique.map', 'w')}
			fdeg = {'SE':open(btmapdir+sampleID+'.SE.degenerate.map', 'w'), 'PE':open(btmapdir+sampleID+'.PE.degenerate.map', 'w')}
		
		HEADER = {'SE':None,'PE':None}
		
 		# the list of (run,lane) pairs of samples corresponding to this sampleID
		thesamples = seqRuns[sampleID].keys(); thesamples.sort()
		
		
		for run,lane in thesamples:
			# multiplicity counting by read
			UID = '%s-%s-%s'%(sampleID,run,lane)
			
			# if user wants to organize a preexisting map
			if EXORGANIZE: UID = sampleID
			
			refGenome,paired,path = seqRuns[sampleID][(run,lane)][:3]
			rlp = '%s,%s-'%(run,lane) # run,lane prefix for all reads
			currentsample += 1
			
			# what map files are relevant for this sampleID?
			fpairs = []
			if paired == 1:
				# note the PE goes last so that the SAM header is correctly saved to the map files
				# filenames for singleton leftovers from PE
				fpairs += [('SE', btmapdir+UID+'_singles_1'+mapsuffix)]
				fpairs += [('SE', btmapdir+UID+'_singles_2'+mapsuffix)]
				# filename for standard PE run
				fpairs += [('PE', btmapdir+UID+mapsuffix)]
			else:
				# filename for singleton run
				fpairs += [('SE', btmapdir+UID+mapsuffix)]
			
			# special nounique map to store discarded best-noguess alignments
			createdir(unmapdir)
			fbestno = open(unmapdir+UID+'.bestno_discard.sam', 'w')
			
			# for each map file, sort then partition into organized map files
			for pese,fname in fpairs:
				
				# Stream-organization method for read organization
				# because the reads are sorted by id, we can easily count number of alignments
				# based on number of rows
				
				keyid = None # id of read under consideration
				keypese = None # if this read maps as SE or PE
				ALN = [] # store alignment rows for a single hitid
				MMS = [] # store mismatches per alignment row
				
				# histogram summarizing multiplicity distribution
				RIDm = dict(map(lambda x: (x,0), xrange(1,maxnumalign+1)))
				
				if os.access(fname, os.F_OK):
					parts = fname.split('/')
					pipeit('- %s Counting alignments for %s...'%(currentsample, parts[len(parts)-1]))
					fh = open(fname)
					count = 0
					
					for row in fh:
						if count % 5000000 == 0: pipeit('%.1e,'%(count))
						count += 1
						
						# reprint sam-format headers, but only once
						if (HEADER[pese]==None or HEADER[pese]==fname) and (row[0:3]=='@PG' or row[0:3]=='@HD' or row[0:3]=='@SQ'):
							
							# don't print header twice
							if PESECOMBO:
								if pese == 'PE':
									HEADER[pese] = fname
								else:
									print>>funi[pese], row[:-1]
									print>>fdeg[pese], row[:-1]
									HEADER[pese] = fname
							else:
								print>>funi[pese], row[:-1]
								print>>fdeg[pese], row[:-1]
								HEADER[pese] = fname
							continue
							
						elif row[0]=='@': continue
						
						try:
							lst = row[:-1].split('\t')
							hitid = lst[0]; chrom = lst[2]
							rem = ' '.join(lst[11:len(lst)])
							hasmate = lst[6]; insertsize=int(lst[8])
						except Exception, e:
							pipeit('Error parsing SAM line %s; does SAM file have content?',1)
							continue
							
						prelinen = None
						if hasmate != '*' and (hitid[-2:] == '/1' or hitid[-2:] == '/2'):
							hitid = hitid[:-2]
							lst[0] = hitid
							prelinen = ''.join([rlp,'\t'.join(lst)])
						else: prelinen = ''.join([rlp,row[:-1]]) # prepend run,lane information
						
						# toss unmapped reads
						if chrom == '*': continue
						
						# don't double count, so only record positive values
						# only keep a fraction of them to save on memory
						if hasmate != '*' and insertsize >= 0 and random.random() < 0.01: 
							isizes += [insertsize]
						
						mmtmp = re.split('.*NM:i:(\d+).*', rem)
						nmismatches = (len(mmtmp)>1 and int(mmtmp[1])) or 0
						
						# sam format now eliminates /1 and /2 for PE reads that map as PE
						# thus we need to use flag info for this
						# However, older bowtie versions keep the /1 and /2, so trim
						
						if keyid == None:
							keyid = hitid
							keypese = hasmate == '*' and 'SE' or 'PE'
							ALN = [prelinen]
							MMS = [nmismatches]
						# distinguish between single and paired
						elif keyid == hitid:
							ALN += [prelinen]
							MMS += [nmismatches]
						elif keyid != hitid:
							# we encountered the next read
							KEEPREAD = True
							
							# save previous read/alignments to the appropriate file
							# -----------------------------------------------------
							# if PE then write to PE.unique if there are only 2 entries
							if keypese == 'PE' and len(ALN) == 2:
								print >>funi[pese], ALN[0]
								print >>funi[pese], ALN[1]
								# get our own line count for proper sam file format
								try: ulinesat[pese][chrom] += 2
								except KeyError: ulinesat[pese][chrom] = 2
							elif keypese == 'SE' and len(ALN) == 1:
								print >>funi[pese], ALN[0]
								# get our own line count for proper sam file format
								try: ulinesat[pese][chrom] += 1
								except KeyError: ulinesat[pese][chrom] = 1
							else:
								
								# determine if one alignment has fewest mismatches
								# if not, then toss all of them
								# this works for SE reads
								if STRATEGY=='bestno':
									# what about PE reads? 
									# need to group mismatch pairs by reads
									# this assumes mate pairs are interleaved
									omms = MMS
									if keypese == 'PE':
										MMS = map(lambda x,y: (x,y), MMS[:len(MMS)/2],MMS[len(MMS)/2:])
										
									if MMS.count(min(MMS)) > 1: KEEPREAD = False
									
									if KEEPREAD:
										# store the single best read (by mismatches)
										print >>funi[pese], ALN[argmin(MMS)]
										# get our own line count for proper sam file format
										try: ulinesat[pese][chrom] += 1
										except KeyError: ulinesat[pese][chrom] = 1
									else:
										for li in xrange(len(ALN)):
											print >>fbestno, '\t'.join([ALN[li],str(omms[li])])
										
								else:
									# always keep read - store all alignments here
									for linen in ALN:
										print >>fdeg[pese], linen
										# get our own line count for proper sam file format
										try: dlinesat[pese][chrom] += 1
										except KeyError: dlinesat[pese][chrom] = 1
							
							if KEEPREAD:
								# record histogram information
								# if PE then add 2 reads each category
								if keypese == 'PE': 
									try: RIDm[len(ALN)/2] += 2
									except KeyError: RIDm[len(ALN)/2] = 2
								else: 
									try: RIDm[len(ALN)] += 1
									except KeyError: RIDm[len(ALN)] = 1
								
								# record mismatches
								for nmm in MMS:
									# npmms += (len(mmtmp)>1 and int(mmtmp[1])>0 and 1) or 0
									npmms += nmm > 0 and 1 or 0
									nrows += 1
								
							# -----------------------------------------------------
							
							# create new target
							keyid = hitid
							keypese = hasmate == '*' and 'SE' or 'PE'
							ALN = [prelinen]
							MMS = [nmismatches]
					
					# finally, spit out the last read(s) in the file (copy of above code)
					# -------------------------------------------------------------------
					
					# we encountered the next read
					KEEPREAD = True
					
					# save previous read/alignments to the appropriate file
					# -----------------------------------------------------
					# if PE then write to PE.unique if there are only 2 entries
					if keypese == 'PE' and len(ALN) == 2:
						print >>funi[pese], ALN[0]
						print >>funi[pese], ALN[1]
						# get our own line count for proper sam file format
						try: ulinesat[pese][chrom] += 2
						except KeyError: ulinesat[pese][chrom] = 2
					elif keypese == 'SE' and len(ALN) == 1:
						print >>funi[pese], ALN[0]
						# get our own line count for proper sam file format
						try: ulinesat[pese][chrom] += 1
						except KeyError: ulinesat[pese][chrom] = 1
					else:
						
						# determine if one alignment has fewest mismatches
						# if not, then toss all of them
						# this works for SE reads
						if STRATEGY=='bestno':
							# what about PE reads? 
							# need to group mismatch pairs by reads
							# this assumes mate pairs are interleaved
							omms = MMS
							if keypese == 'PE':
								MMS = map(lambda x,y: (x,y), MMS[:len(MMS)/2],MMS[len(MMS)/2:])
								
								
								if MMS.count(min(MMS)) > 1: KEEPREAD = False
							
								if KEEPREAD:
									# store the single best read (by mismatches)
									print >>funi[pese], ALN[argmin(MMS)]
									# get our own line count for proper sam file format
									try: ulinesat[pese][chrom] += 1
									except KeyError: ulinesat[pese][chrom] = 1
								else:
									for li in xrange(len(ALN)):
										print >>fbestno, '\t'.join([ALN[li],str(omms[li])])
								
						else:
							# always keep read - store all alignments here
							for linen in ALN:
								print >>fdeg[pese], linen
								# get our own line count for proper sam file format
								try: dlinesat[pese][chrom] += 1
								except KeyError: dlinesat[pese][chrom] = 1
					
					if KEEPREAD:
						# record histogram information
						# if PE then add 2 reads each category
						if keypese == 'PE': 
							try: RIDm[len(ALN)/2] += 2
							except KeyError: RIDm[len(ALN)/2] = 2
						else: 
							try: RIDm[len(ALN)] += 1
							except KeyError: RIDm[len(ALN)] = 1
						
						# record mismatches
						for nmm in MMS:
							# npmms += (len(mmtmp)>1 and int(mmtmp[1])>0 and 1) or 0
							npmms += nmm > 0 and 1 or 0
							nrows += 1
					# -------------------------------------------------------------------
					
					fh.close()
					sys.stdout.write('Done (n=%s).\n'%(count)); sys.stdout.flush()
				else: pipeit('Cannot access %s'%(fname),1)
				# end loop reading read info
				
				# Aggregate the multiplicity histogram information
				for val,vct in RIDm.iteritems():
					try: sidHist[pese][val] += vct
					except KeyError: sidHist[pese][val] = vct
					
					try: sidHist['Total'][val] += vct
					except KeyError: sidHist['Total'][val] = vct
					
					totreads[pese] += vct
					totaln[pese] += val*vct
				del RIDm
				del ALN
				del MMS
				
				
			# finished processing component read map file for this map file
			fbestno.close()
		# finished all map files for this sampleID
		
		# Save results for this sampleID
		# move on to next group if there are no mapped reads
		if nrows == 0: continue
		
		# save total read counts for this sampleID
		totreads['Total'] += totreads['SE']+totreads['PE']
		totaln['Total'] += totaln['SE']+totaln['PE']
		
		# record number of perfect/mismatch alignments
		perfrac = 0
		nper = nrows - npmms
		row = []
		perfrac = nper/float(nrows)
		mmfrac = npmms/float(nrows)
		row += ['There are %s/%s (%s) perfect (%s/%s) alignments'%(nper,nrows,perfrac,maxreadlen,maxreadlen)]
		row += ['There are %s/%s (%s) mismatched (<%s/%s) alignments'%(npmms,nrows,mmfrac,maxreadlen,maxreadlen)]
		printList(row, file=pmadir+'Perfect vs mismatch alignments-%s.txt'%(sampleID))
		
		# save perfect/mismatch reads summary
		info = '\t'.join(map(str, [sampleID, nper, npmms, nrows, perfrac, mmfrac]))
		record(info, pmafile, 'a')
		
		# save read and multiplicity statistics
		sestr = ', '.join('('+','.join(map(str,p))+')' for p in sorted(sidHist['SE'].items()))
		pestr = ', '.join('('+','.join(map(str,p))+')' for p in sorted(sidHist['PE'].items()))
		totstr = ', '.join('('+','.join(map(str,p))+')' for p in sorted(sidHist['Total'].items()))
		del sidHist # clean up
		if sestr == '': sestr = nastr
		if pestr == '': pestr = nastr
		
		recfile = almfile
		record(sampleID, recfile, 'a')
		record('SE: reads=%s, alignments=%s, alignments/read=%s'%(totreads['SE'], totaln['SE'], roundna(3)(divna(totaln['SE'],totreads['SE']))), recfile, 'a')
		record('Histogram: '+sestr, recfile, 'a')
		record('PE: reads=%s, alignments=%s, alignments/read=%s'%(totreads['PE'], totaln['PE'], roundna(3)(divna(totaln['PE'], totreads['PE']))), recfile, 'a')
		record('Histogram: '+pestr, recfile, 'a')
		record('Total: reads=%s, alignments=%s, alignments/read=%s'%(totreads['Total'], totaln['Total'], roundna(3)(divna(totaln['Total'], totreads['Total']))), recfile, 'a')
		record('Histogram: '+totstr, recfile, 'a')
		record('', recfile, 'a')
		
		# Done saving statistics for this sampleID
		# ----------------------------------------------------
		
		for pese in ['PE','SE']:
			funi[pese].close(); fdeg[pese].close()
		
		pipeit('- done partitioning unique and degenerate reads.',1)
		
		
		# ==============================================================
		# save the number of SE/PE reads on each chromosome
		fh = open(crdfile, 'a')
		# this prints the number of reads on each chromosome
		# for k in sorted(GRlengths.keys()):
		
		for k in sorted(unique(ulinesat['SE'].keys()+ulinesat['PE'].keys())):
			use = upe = dse = dpe = 0
			try: use = ulinesat['SE'][k]
			except KeyError: pass
			try: upe = ulinesat['PE'][k]
			except KeyError: pass
			try: dse = dlinesat['SE'][k]
			except KeyError: pass
			try: dpe = dlinesat['PE'][k]
			except KeyError: pass
			print >>fh, '\t'.join(map(str, [sampleID, k, use, upe, dse, dpe]))
		fh.close()
		# ==============================================================
		
		createdir(zinsdir)
		sys.stdout.write('- Saving empirical insert size distribution...'); sys.stdout.flush()
		hist = histogram(isizes[0:250000], bins=makeRange(minins,maxins+2,2))
		printTable(hist, header=['Size', 'Count'], file=zinsdir+'%s_insert_size_distribution.txt'%(sampleID))
		sys.stdout.write('Done.\n'); sys.stdout.flush()
		
		# clean up
		del ulinesat; del dlinesat; del isizes
		
		# finally delete the precursor bowtie files
		# -----------------------------------------
		if not KEEPBOWTIE:
			for run,lane in seqRuns[sampleID].keys():
				UID = '%s-%s-%s'%(sampleID,run,lane)
				refGenome,paired,path = seqRuns[sampleID][(run,lane)]
				thefiles = [btmapdir+UID+'.sam', btmapdir+UID+'_singles_1.sam', btmapdir+UID+'_singles_2.sam']
				# thefiles += [btmapdir+UID+'.sam.sort', btmapdir+UID+'_singles_1.sam.sort', btmapdir+UID+'_singles_2.sam.sort']
				for fn in thefiles: os.system('rm -f "%s"'%(fn))
	
	if CLEARBOWTIE:
		pipeit('- Clearing precusor bowtie map files...',1)
		for sampleID in sorted(seqRuns.keys()):
			for run,lane in seqRuns[sampleID].keys():
				UID = '%s-%s-%s'%(sampleID,run,lane)
				refGenome,paired,path = seqRuns[sampleID][(run,lane)]
				thefiles = [btmapdir+UID+'.sam', btmapdir+UID+'_singles_1.sam', btmapdir+UID+'_singles_2.sam']
				for fn in thefiles: os.system('rm -f "%s"'%(fn))
	
	pipeit('DONE READ ORGANIZATION.\n',1)


# link filenames if user supplies their own unique map file
if LINKUNIQUE:
	for sampleID in sortseqruns:
		sourcefile = None
		tmp = '%s%s%s'%(btmapdir,sampleID,mapsuffix)
		if os.access(tmp, os.F_OK): sourcefile = tmp
		
		if not sourcefile: 
			pipeit('Error: could not find %s'%(tmp))
			break
		targetfile = '%s%s.unique.map'%(btmapdir,sampleID)
		if os.access(targetfile, os.F_OK): break
		pipeit('- Linking map file %s to %s'%(sourcefile.split('/')[-1], targetfile.split('/')[-1]),1)
		os.system('ln -s "%s" "%s"'%(sourcefile,targetfile))
		PESECOMBO = True

#
if INDEXMAPFILES or INDEXUNI:
	# sort unique read maps
	pipeit('- Sorting and indexing unique maps...', 1)
	ANYSAMPLES = False
	for sampleID in sorted(seqRuns.keys()):
		for pese in ['PE.', 'SE.', '']:
			base = btmapdir+sampleID+'.'+pese+'unique.map'
			if not os.access(base, os.F_OK): continue
			if not os.access(base+'.index', os.F_OK) or RESORT:
				ANYSAMPLES = True
				pipeit('  %s...'%(sampleID+'.'+pese+'unique.map'))
				if not INDEXONLY:
					# os.system('mv "%s" "%s"'%(base, base+'.unsort'))
					# os.system('sort -k 3,3 -k 4,4n "%s" > "%s"'%(base+'.unsort', base))
					os.system('sort -T "%s" -k 3,3 -k 4,4n "%s" > "%s"'%(sorttmpdir, base, base+'.sorted'))
					os.system('mv "%s" "%s"'%(base+'.sorted', base))
					# os.system('rm -f "%s"'%(base+'.unsort')) # let user delete these
				indexFile(base, base+'.index')
				pipeit('done',1)
	if not ANYSAMPLES: pipeit('  No samples need indexing.',1)
	else: pipeit('- Done sorting and indexing unique maps. Manually remove *.unsort files if desired.', 1)

if INDEXMAPFILES or INDEXDEG:
	pipeit('- Indexing degenerate maps...', 1)
	ANYSAMPLES = False
	for sampleID in sorted(seqRuns.keys()):
		for pese in ['PE.', 'SE.', '']:
			base = btmapdir+sampleID+'.'+pese+'degenerate.map'
			if not os.access(base, os.F_OK): continue
			if not os.access(base+'.index', os.F_OK) or RESORT:
				pipeit('  %s...'%(sampleID+'.'+pese+'degenerate.map'))
				ANYSAMPLES = True
				badIDs = {}
				if REDUNDANCYFILTER:
					pipeit('load PCR duplicates...')
					afile = badReadDir+'%s_degenerate_redundantIDs.txt'%(sampleID)
					if not os.access(afile, os.F_OK): pipeit('Warning: redundancy filter ON but file not found: %s'%(afile),1)
					badIDs = loadPCRduplicates(afile)
				indexDegFile(base, sampleID+'.'+pese+'degenerate.map', badIDs)
				pipeit('done.',1)
	if not ANYSAMPLES: pipeit('  No samples need indexing.',1)
	else: pipeit('- Done indexing degenerate maps.', 1)
	


if ESTIMATEGLOBALERROR:
	log = lambda x: math.log(x, 10)
	Q2P = lambda e: (e == 0 and 1-1e-16) or 10**(-e/10.) # avoid 1
	P2Q = lambda p: -10*log(p)
	A2P = dict( [ascii, Q2P(A2Q[ascii])] for ascii in A2Q.keys() )
	
	pipeit('Estimating global per-nucleotide sequencing error rate...',1)
	errorsum = 0
	ecount = 0
	# errors = []
	for mapfile in getFiles(btmapdir, include='unique'):
		pipeit('- Loading', mapfile,1)
		fh = open(mapfile)
		count = 1
		for line in fh:
			if count % 1000000 == 0: pipeit('     reached', count,1)
			count+=1
			
			try:
				lst = line[:-1].split('\t')
				chrom = lst[2]
				if chrom == '*': continue
				
				for x in map(lambda x: A2Q[x], lst[10]):
					errorsum += x
					ecount += 1
					
			# exception if a line beginning with @
			except IndexError: pass
			except ValueError: pass
		fh.close()
	
	print 'Mean error over %s positions is'%(ecount), errorsum/float(ecount), 'or', Q2P(errorsum/float(ecount))
	print
	sys.exit()
# DONE

GR = None # we need the reference genome here...
flg = None
NEEDTOLOADREFERENCE = True

sampleIDs = sorted(seqRuns.keys())
newSampleIDs = []
if len(singlesampleid):
	for sid in singlesampleid:
		if sid in sampleIDs: newSampleIDs += [sid]
		else: sys.exit('Error: sample "%s" specified by --sampleids/--ids not found in map.'%(sampleID))
if len(newSampleIDs): sampleIDs = newSampleIDs; del newSampleIDs


# ====================================================
# Part III: Identify SNPs using bayesian SNP detection
# ====================================================
if SNPCALLING:
	completedir = zdir+'completed/'; createdir(completedir)
	createdir(resdir); createdir(snpdir); createdir(tmpdir)
	thesnpdir = snpdir
	if DRYRUN: thesnpdir = tmpdir
	
	
	if not bfaref:
		sys.exit('Reference genome must be specified using --ref/-r <path/to/reference>')
	
	
	for sid in sampleIDs:
		if PESECOMBO:
			if not os.access(btmapdir+sid+'.unique.map.index', os.F_OK):
				sys.exit('Error: no index files found for sample "%s". Run the --index flag before genotyping.'%(sid))
		else:
			if not os.access(btmapdir+sid+'.SE.unique.map.index', os.F_OK) or not os.access(btmapdir+sid+'.PE.unique.map.index', os.F_OK):
				sys.exit('Error: no index files found. Run the --index flag before genotyping.')
	
	# Only load reference if re-estimating...
	if GR == None: NEEDTOLOADREFERENCE = True
	for sampleID in sampleIDs:
		motherfile = snpdir+sampleID+'.mother'   # genotype at all informative positions
		motherfile += boundslabel+'.txt'
		if not os.access(motherfile, os.F_OK) or REDOSNPCALLING: NEEDTOLOADREFERENCE = True
	
	# ---------------------------
	# which sampleIDs do we actually need to execute on?
	sids = []
	mofs = {}
	noms = {}
	for sid in sampleIDs:
		motherfile = thesnpdir+sid+'.mother'+boundslabel+'.txt'
		nominalfile = thesnpdir+sid+'.Q%s'%(QvalueRange[0])+boundslabel+'.txt'
		
		# need to check if file exists...............
		if not os.access(motherfile, os.F_OK) or REDOSNPCALLING:
			# pipeit('- Genotyping sample %s...'%(sid),1)
			sids += [sid]
			
			if SAVENOMINAL:
				noms[sid] = nominalfile
				mofs[sid] = ''
			else:
				mofs[sid] = motherfile
				noms[sid] = ''
			
			# mofs[sid] = motherfile
			# noms[sid] = (SAVENOMINAL and nominalfile) or ''
		else: 
			pipeit(' - Already processed/processing SNPs for %s, SKIPPING.'%(sid),1)
	
	pipeit('- Sniper will genotype samples %s (nprocs=%s, xflag=%s)...'%(', '.join(sids),nsubprocs,xflag), 1)
	
	# true multiprocessing
	# only run this section if we need to schedule sniper jobs for mother file
	
	# add chrom breakdown and sge calling
	# idea is that if a boundsfile was provided and -x is used, we parallelize over each entry in the boundsfile
	# the other thing is we may want to sequence say 5 chroms on one process (notably for SGE) so how to handle this?
	# print 'xflag', xflag, 'ss', ISSUBPROC, nsubprocs, USESGE
	# xflag groupchrom ss False 1 False
	# xflag groupchrom ss False 2 False
	
	if not ISSUBPROC and (nsubprocs > 1 or USESGE or xflag == 'bychrom'):
		# pipeit('PART III: Genotyping',1)
		jobs = []
		cct = 1
		logdir = zdir+'log_files/'
		createdir(logdir)
		
		for sampleID in sids:
			pipeit('',1)
			
			exe = sys.argv[0]
			if callprefix: exe = 'python "%s"'%(callprefix)
			call = exe+' '+' '.join(map(lambda x: x[0]=='-' and x or '"%s"'%(x), sys.argv[1:]))
			
			call = call.replace('-x %s'%(nsubprocs),'')
			# if meant as an execute job, remove --sge flag!
			call = call.replace('--sge', '').replace('-sge','')
			# need -x 1 to get sniper to actually run
			call += ' -x 1 --ids %s --issubproc'%(sampleID)
			
			if '-o' not in call and '-out' not in call: call += ' -o "%s"'%(outdir)
			
			# add flags: --bychrom --bysample --auto (default)
			# --bychrom is --auto at the moment
			# no need for --by sample really
			
			# parallelize within chromosome for each chromosome for each sampleID
			if xflag == 'bysampleid' or xflag =='bysample' or xflag == 'groupchrom':
				logstr = ' >> "%s"'%(logdir+'%s_%s.log'%(sampleID,tmpstr))
				pipeit('   Call %s: %s'%(cct, call+logstr),1)
				jobs += [call+logstr]
				cct += 1
			else:
				# get the chroms from header in map file
				chrmlen = {}
				if PESECOMBO:
					try: chrmlen = getChromsFromHeader(btmapdir+sampleID+'.unique.map')
					except IOError: pipeit('Error: cannot find %s'%(btmapdir+sampleID+'.unique.map'))
				else:
					try: chrmlen = getChromsFromHeader(btmapdir+sampleID+'.SE.unique.map')
					except IOError: 
						try: chrmlen = getChromsFromHeader(btmapdir+sampleID+'.PE.unique.map')
						except IOError: pipeit('Error: cannot find %s'%(btmapdir+sampleID+'.SE|PE.unique.map'))
				copydir = chrmlen.keys()
				
				# if user provides bounds, distribute them
				if len(seqbounds):
					pipeit('- Distributing processes by user-supplied boundaries file, per sampleID...',1)
					copydir = seqbounds.keys()
					
					# if user only provided a single scaffold/chrom then change flag to subchrom
					if len(seqbounds) == 1:
						xflag = 'subchrom'
						
						
				# now proceed to genotype each reference genomic position
				npieces = nsubprocs
				nchroms = len(copydir)
				
				if xflag == 'bychrom': npieces = nchroms
				elif xflag != 'subchrom' and npieces > nchroms: npieces = nchroms
				copydir.sort() # must be sorted!
				
				if nchroms == 0:
					sys.exit('Cannot find header. Check map directory: %s'%(btmapdir+sampleID))
				# elif nchroms > 1000: xflag = 'bychrom'
				
				# ideally we want to split up evenly by summing the genome length
				# and then breaking up individual chromosomes
				# now that read maps are not loaded into memory we can do this easily - only factor is reference genome
				
				if xflag == 'subchrom':
					# pipeit('subchrom',1)
					# continue to break up each chromosome into pieces
					# this is straightforward if we are not using SGE, so start with local multiprocessing
					# take each scaffold/chromosome at a time and break up into x pieces
					chromgroups = []
					for achr in copydir:
						# determine piece length
						intervals = makeRange(0,chrmlen[achr],n=npieces+1)
						pieces = [[int(intervals[i-1]),int(intervals[i])-1] for i in xrange(1,len(intervals))]
						# just make sure first and last positions are correct
						pieces[0][0] = 0; pieces[-1][1] = chrmlen[achr]
						# print 'P', achr, pieces, chrmlen[achr]
						for p in pieces: chromgroups += [[achr]+list(p)]
					
					pipeit('- Automatically distributing into -x %s pieces x %s chromosomes x %s sampleIDs = %s jobs...'%(npieces,nchroms,len(sids), npieces*len(copydir)*len(sids)),1)
				
					# generate a job per chromosome interval
					for achr,start,stop in chromgroups:
						boundslabeltmp = '%s,%s,%s'%(achr,start,stop)
						boundstr = ' --rest "%s" --nosig'%(boundslabeltmp)
						logstr = ' >> "%s"'%(logdir+'%s_%s_%s.log'%(sampleID,boundslabeltmp,tmpstr))
						pipeit('   Call %s: %s'%(cct, call+boundstr+logstr),1)
						jobs += [call+boundstr+logstr]
						cct += 1
				
				else:
					# pipeit('not subchrom',1)
					# parallelize across chromosomes
					restdir = zdir+'boundsfiles/'
					createdir(restdir)
				
					if npieces > nchroms: npieces = nchroms
					# randomize order to even out the memory/time requirements
					if SHUFFLEBOUNDS: random.shuffle(copydir)
					pipeit('- Automatically distributing into -x %s jobs per sampleID (%s sampleIDs)...'%(npieces,len(sids)),1)
				
					# split chromosomes into LIPieces loading groups
					chromgroups = []
					try: itemspergroup = math.ceil(nchroms / float(npieces))
					except ZeroDivisionError: itemspergroup = 1
					while len(copydir):
						currgrp = []
						while len(currgrp) < itemspergroup and len(copydir):
							currgrp += [copydir.pop(0)]
						chromgroups += [currgrp]
				
					# generate a job per chromosome group
					for chromgrp in chromgroups:
						# create a bounds file with these rows
						# (can't just concatenate and call because there is a character limit for unix shell commands)
					
						boundslabeltmp = ''
						boundstr = ''
					
						if nsubprocs > 1 and xflag != 'bychrom':
							boundslabeltmp = '%s--%s'%(chromgrp[0],chromgrp[-1])
							if len(chromgrp) == 1: boundslabeltmp = '%s'%(chromgrp[0])
							boundsfile = restdir+'%s.txt'%(boundslabeltmp)
							printList(chromgrp, file=boundsfile)
							boundstr = ' --rest "%s" --nosig'%(boundsfile)
					
						elif xflag == 'bychrom':
							boundslabeltmp = chromgrp[0].strip()
							boundstr = ' --rest "%s" --nosig'%(chromgrp[0].strip())
					
						elif boundsfile:
							# just reprint the string
							boundslabeltmp = '%s'%('.'.join(boundsfile.split('/')[-1].split(".")[0:-1]))
							boundstr = '--rest "%s"'%(boundsfile)
					
						logstr = ' >> "%s"'%(logdir+'%s_%s_%s.log'%(sampleID,boundslabeltmp,tmpstr))
						pipeit('   Call %s: %s'%(cct, call+boundstr+logstr),1)
						jobs += [call+boundstr+logstr]
						cct += 1
			
		pipeit('',1)
		
		# set up 
		# if not ISSUBPROC and os.access(completedir, os.F_OK) and len(getFiles(completedir)) > 0:
		# 	os.system('rm -rf "%s"'%(completedir))
		# 	createdir(completedir)
		
		# this should be totally modular
		njobs = len(jobs) # initial job count
		nlaunched = 0
		running_jobs = []
		ncomplete = 0
		nkilled = 0
		if USESGE:
			while len(jobs):
				pipeit('- Scheduling job %s (%s completed, %s killed | x%s processors)'%(nlaunched+1, ncomplete, nkilled, nsubprocs),1)
				# and we can launch another job
				# fetch the next job
				call = jobs.pop(0)
				# and execute it
				nlaunched += 1
				os.system('qsub -b y %s -cwd %s' % (SGEMEM, call))
		else:
			while len(jobs):
				# as long as there are more jobs to be processed
				if nlaunched - ncomplete - nkilled < nsubprocs:
					pipeit('- Scheduling job %s (%s completed, %s killed | x%s processors)'%(nlaunched+1, ncomplete, nkilled, nsubprocs),1)
					# and we can launch another job
					# fetch the next job
					call = jobs.pop(0)
					# and execute it
					nlaunched += 1
					job = subprocess.Popen(call, shell=True)
					running_jobs += [job]
					time.sleep(.1)
				else:
					nkilled = 0
					marked = []
					for ji in xrange(len(running_jobs)): 
						if running_jobs[ji].poll() != None: 
							# print 'poll killed', rj[ji].poll(), len(running_jobs)
							ncomplete += 1
							marked += [running_jobs[ji]]
					for m in marked:
						running_jobs.pop(running_jobs.index(m))
					
					# check to see if an existing job has finished
					# ncomplete = len(getFiles(completedir, include='finished'))
					time.sleep(1)
		
		pipeit('',1)
		
		if USESGE: sys.exit('Finished scheduling SGE jobs. \nWhen they have finished processing, execute sniper.py without -s or --sge flags to perform standard SNP filtering, or add --fwer P<FLOAT>,readlength<INT> for adaptive SNP calling. ')
		for job in running_jobs: job.wait() # halt code until all jobs are finished.
	else:
		# call sniper for this single sampleid
		if NEEDTOLOADREFERENCE:
			sys.stdout.write('- Loading reference genome... '); sys.stdout.flush()
			try: GR = readFasta(bfaref, TOUPPER=True)
			except IOError: sys.exit("CANNOT READ REFERENCE FILE %s"%(bfaref))
			flg = sum(map(lambda x: len(x['seq']), GR.values()))
			sys.stdout.write('N=%s loci\n'%(flg))
			
		# def dummyLooper(jobs=[]):
		# 	for kwargs in jobs:
		# 		job = threading.Thread(target=sniper, kwargs=kwargs)
		# 		job.run()
		# 	return 1
		
		
		ct = 0
		for sampleID in sids:
			# set up SNP calling for PE vs PE+SE / unique reads vs unique + deg reads
			pref = btmapdir+sampleID
			rf = {'PEuni':'', 'PEdeg':'', 'SEuni':'', 'SEdeg':''}
			
			pipeit('\n- Processing sample %s...'%(sampleID),1)
			
			if PESECOMBO:
				rf['SEuni'] = pref+'.unique.map'
				if STRATEGY!='unique' and STRATEGY!='bestno': 
					rf['SEdeg'] = pref+'.degenerate.map'
			else:
				rf = {'PEuni':pref+'.PE.unique.map', 'PEdeg':'', 'SEuni':'', 'SEdeg':''}
			
				if USESINGLETONS:
					rf['SEuni'] = pref+'.SE.unique.map'
					if STRATEGY!='unique' and STRATEGY!='bestno': 
						rf['PEdeg'] = pref+'.PE.degenerate.map'
						rf['SEdeg'] = pref+'.SE.degenerate.map'
				else:
					if STRATEGY!='unique' and STRATEGY!='bestno':
						rf['PEdeg'] = pref+'.PE.degenerate.map'
			
			kwargs = {'readfiles':rf, 'GR':GR, 'saveall':mofs[sampleID], 'outfile':noms[sampleID], 'thesnpdir':thesnpdir, 'expPhred':expPhred, 'maxreadlen':maxreadlen, 'Lambda':Lambda, 'prior':ssPrior[sampleID], 'theta':ssTheta[sampleID], 'tmpdir':tmpdir, 'bounds':seqbounds, 'LIPieces':LIPieces, 'mindepth':mindepth, 'maxdepth':maxdepth}
			
			nh = sniper(**kwargs)
			
			# finished processing, so write to an outfile so parent process knows to launch another
			fin = open(completedir+'%s.finished'%(sampleID+boundslabel), 'w'); fin.close()
			
		sys.exit('Results for %s saved to %s'%(sampleID, thesnpdir))
	# END SNIPER LOGIC

# first rejoin subchrom pieces
if JOINPIECES and xflag == 'subchrom' and IDENTIFYSIGNIFICANTS:
	pipeit('- Joining mother subchrom pieces...',1)
	for sampleID in sampleIDs:
		pipeit('  %s...'%(sampleID))
		# figure out the files
		# get the chroms from header in map file
		chrmlen = {}
		if PESECOMBO:
			try: chrmlen = getChromsFromHeader(btmapdir+sampleID+'.unique.map')
			except IOError: pipeit('Error: cannot find %s'%(btmapdir+sampleID+'.unique.map'))
		else:
			try: chrmlen = getChromsFromHeader(btmapdir+sampleID+'.SE.unique.map')
			except IOError: 
				try: chrmlen = getChromsFromHeader(btmapdir+sampleID+'.PE.unique.map')
				except IOError: pipeit('Error: cannot find %s'%(btmapdir+sampleID+'.SE|PE.unique.map'))
		copydir = chrmlen.keys()
		if len(seqbounds): copydir = seqbounds.keys()
		
		chromfiles = ['%s.mother.%s.txt'%(sampleID,achr) for achr in sorted(copydir)]
		
		for achr in sorted(copydir):
			pipeit('%s,'%(achr))
			motherfile = '%s.mother.%s.txt'%(sampleID,achr)
			chrpieces = getFiles(snpdir,include='%s.mother.%s'%(sampleID,achr), exclude=motherfile)
			
			if os.access(snpdir+motherfile, os.F_OK) and len(chrpieces)==0: continue
			elif len(chrpieces) == 0: sys.exit('No files with prefix %s.mother.%s'%(sampleID,achr))
			
			# put these into sorted order - by chrom then by position
			parts = map(lambda x: [int(x.split('/')[-1].split('.')[2].split(',')[1]),x], chrpieces)
			
			chrmother = snpdir+'%s.mother.%s.txt'%(sampleID,achr)
			fh = open(chrmother,'w'); fh.close()
			
			for start,piece in sorted(parts):
				# print 'concatenating', piece.split('/')[-1]
				os.system('cat "%s" >> "%s"'%(piece,chrmother))
				# os.system('rm "%s"'%(piece))
		
		pipeit('final merge...')
		# merge chromosomes into single file
		finalfile = resdir+'%s.mother.txt'%(sampleID)
		fh = open(finalfile,'w'); fh.close()
		for achr in sorted(copydir):
			chromfile = '%s%s.mother.%s.txt'%(resdir,sampleID,achr)
			os.system('cat "%s" >> "%s"'%(chromfile, finalfile))
			os.system('rm -f "%s"'%(chromfile))
		pipeit('done',1)
		
	pipeit('- Finished merging the "--subchrom" mother likelihood files',1)
	pipeit('- NB, you must manually delete the partial likelihood files (samplex.mother.chrx.start,stop,+.txt).',1)

if ISSUBPROC: sys.exit('')

if IDENTIFYSIGNIFICANTS:
	allfiles = []
	for sampleID in sampleIDs:
		allfiles += [getFiles(snpdir,include=sampleID+'.mother')]
	if sum(map(lambda x: len(x), allfiles)) == 0: sys.exit('No files to filter for significant SNPs. Exiting.')
	
	pipeit('- Identifying significant SNP loci from distributed mother files...',1)
	for i in xrange(len(allfiles)):
		basename = sampleIDs[i]+boundslabel
		if not len(allfiles[i]): pipeit('- Cannot find mother SNP file for %s...SKIP.'%(basename),1)
		filterSignificantSNPs(basename, allfiles[i], ADAPTIVEQ=ADAPTIVEQ, QvalueRange=QvalueRange, prior=ssPrior[sampleID])
		

os.system('rm -rf "%s"'%(tmpdir))


# ====================================================
# Part IV: Diagnostics
# ====================================================
if DIAGNOSTICS:
	pipeit('GENERATING DIAGNOSTICS...',1)
	
	# use the alloutfile results from SNP calling to gather diagnostics
	# including the following:
	# number of reference positions for which we have no read data
	# histogram of total positions with x read depth
	# histogram of posterior probabilities for all positions and significant positions
	
	if ADAPTIVEQ: QvalueRange = [ADAPTIVEQ]
	
	stats = { 'Ns':{}, 'numHypos':{}, 'Qns':{}, 'DH':{}, 'Davg':{}, 'alpha':{}}
	
	desc = {'sigG':'Distribution of significant genotypes',
			'Gtype':'Distribution of significant categorical genotypes',
			'Ns':  'Reference positions with no read data',
			'numHypos':  'Positions with no information (depth=0 or ref=N)',
			'Qns': 'Distribution of all posterior probabilities',
			'Qs':  'Distribution of significant posterior probabilities',
			'DH':  'Distribution of genome-wide read depth',
			'Davg': 'Genome-wide average read depth (avg. depth per position)',
			'sigDavg': 'Average read depth (avg. depth per position) for significant positions',
			'alpha': 'Genome-wide degenerate/unique alignment ratio statistics',
			'sigalpha': 'Significant SNP degenerate/unique alignment ratio statistics'}
	
	# finish initializing sig entries
	Pstats = {}
	for minPvalue in map(str,QvalueRange):
		Pstats[minPvalue] = {'sigG':{}, 'Qs':{}, 'sigDavg':{}, 'sigalpha':{}, 'Gtype':{}}
	
	diagfile = resdir+'Summary-diagnostics'
	diagfile += ubamap
	if POSTPROCESSING: diagfile += '-rcf'
	diagfile += '.txt'
	
	fh = open(diagfile, 'w')
	fh.close()
	
	for sampleID in sorted(seqRuns.keys()):
		
		Ns = {}
		numHypos = {} # number of positions with information, by chromosome
		DH = {0:0, 1:0, 2:0, 3:0, 4:0, 5:0, 6:0, 7:0, 8:0, 9:0, 10:0,
		 	  11:0, 12:0, 13:0, 14:0, 15:0, 16:0, 17:0, 18:0, 19:0, 20:0,
			  40:0, 60:0, 80:0, 100:0, 200:0, 300:0, 400:0, 500:0}
		DHints = sorted(DH.keys())
		Dsum = 0; Dcount = 0; allDvals = []
		alphasum = 0; allalpha = []
		Qns = {0:0, 5:0, 10:0, 15:0, 20:0, 25:0, 30:0, 35:0, 40:0,
			   44:0, 50:0, 55:0, 60:0, 65:0, 70:0, 75:0, 80:0,
			   85:0, 90:0, 95:0, 100:0, 105:0, 110:0, 115:0, 120:0,
			   125:0, 130:0, 135:0, 140:0, 145:0, 150:0, 155:0, 160:0}
		Qints = sorted(Qns.keys())
		
		for motherfile in getFiles(snpdir,include=sampleID+'.mother'):
			label = motherfile.split('/')[-1]
			print ' - Parsing %s...'%(label)
			fh = open(motherfile)
			for line in fh:
				row = line[:-1].split('\t')
			
				if len(row) != 11: continue
			
				C = row[0]
				P = int(row[1])
				R = row[2]
				G = row[3]
				Q = row[4]=='NA' and row[4] or int(row[4])
				D = int(row[5])
				M = int(row[6])
				TM = int(row[7])
				alpha = float(row[8])
				S = row[9] # Y/N
			
				Dsum += D
				Dcount += 1
				allDvals += [D]
				# ratio of unique alignments at this position against total degenerate alignments
				alphasum += alpha
				allalpha += [alpha]
			
				try: numHypos[C] += 1
				except KeyError: numHypos[C] = 1
			
				try: DH[D] += 1
				except KeyError: 
					# then it is between intervals
					idx = 0
					while idx < len(DHints) and D>DHints[idx]:idx += 1
					if idx == len(DHints): idx -= 1
					DH[DHints[idx]] += 1
			
				if G == 'NN':
					try: Ns[C] += 1
					except KeyError: Ns[C] = 1
			
				if Q != 'NA':
					try: Qns[Q] += 1
					except KeyError: 
						try:
							idx = 0
							while Q>Qints[idx]: idx += 1
							Qns[Qints[idx]] += 1
						except IndexError:
							Qns[Qints[len(Qints)-1]] += 1
			fh.close()
		
		
		allalpha = filterstr(nastr)(map(float,allalpha))
		allDvals = filterstr(nastr)(map(float,allDvals))
		
		# save genome-wide statistics for this sampleID
		stats['Ns'][sampleID] = Ns
		stats['numHypos'][sampleID] = numHypos
		stats['Qns'][sampleID] = Qns
		stats['DH'][sampleID] = DH
		stats['Davg'][sampleID] = (divna(Dsum,Dcount), sd(allDvals), getQuants(allDvals))
		stats['alpha'][sampleID] = (divna(alphasum,Dcount), sd(allalpha), getQuants(allalpha))
		
		# collect stats for significant snps
		for minPvalue in QvalueRange:
			snpfile = snpdir+sampleID+boundslabel+'.Q%s.txt'%(minPvalue)
			if ADAPTIVEQ: snpfile = snpdir+sampleID+boundslabel+'.aQ%s,%d.txt'%(minPvalue,readlength)
			
			sigG = {}; Gtype = {'h0':0, 't1':0, 'h2':0, 't2':0}
			sigDsum = 0; sigDcount = 0; sigDvals = []
			sigalphasum = 0
			allsigalpha = []
			Qs = {0:0, 5:0, 10:0, 15:0, 20:0, 25:0, 30:0, 35:0, 40:0,
				   44:0, 50:0, 55:0, 60:0, 65:0, 70:0, 75:0, 80:0,
				   85:0, 90:0, 95:0, 100:0, 105:0, 110:0, 115:0, 120:0,
				   125:0, 130:0, 135:0, 140:0, 145:0, 150:0, 155:0, 160:0}
			
			pipeit(' - Collecting statistics from significant SNPs file at FWER < %s...'%(minPvalue),1)
			# now pass through significant snps
			fh = open(snpfile)
			for line in fh:
				row = line[:-1].split('\t')
				
				if len(row) != 11: continue
				
				C = row[0]
				P = int(row[1])
				R = row[2]
				G = row[3]
				Q = row[4]=='NA' and row[4] or int(row[4])
				D = int(row[5])
				M = int(row[6])
				TM = int(row[7])
				alpha = float(row[8])
				S = row[9] # Y/N
				
				gg = list(G)
				gj = ''.join(sorted(gg))
				try: sigG[gj] += 1
				except KeyError: sigG[gj] = 1
				
				# homo/het/triallele
				if gg[0]==gg[1]:
					if gg[0]==R:
						Gtype['h0'] += 1
					else:
						Gtype['h2'] += 1
				else:
					if gg[0]!=R and gg[1]!=R:
						Gtype['t2'] += 1
					else:
						Gtype['t1'] += 1
				
				try: Qs[Q] += 1
				except KeyError: 
					try:
						idx = 0
						while Q>Qints[idx]: idx += 1
						Qs[Qints[idx]] += 1
					except IndexError:
						Qs[Qints[len(Qints)-1]] += 1
				
				sigalphasum += alpha
				allsigalpha += [alpha]
				
				sigDsum += D
				sigDcount += 1
				sigDvals += [D]
				
			fh.close()
			
			# save significant statistics for this sampleID at Pvalue
			PP = str(minPvalue)
			
			Pstats[PP]['Qs'][sampleID] = Qs
			Pstats[PP]['sigG'][sampleID] = sigG
			Pstats[PP]['Gtype'][sampleID] = Gtype
			Pstats[PP]['sigDavg'][sampleID] = (divna(sigDsum,sigDcount), sd(sigDvals), getQuants(sigDvals))
			Pstats[PP]['sigalpha'][sampleID] = (divna(sigalphasum, sigDcount), sd(allsigalpha), getQuants(allsigalpha))
			
		
		# print out a summary diagnostics file
		fh = open(diagfile, 'a')
		
		for k in sorted(stats.keys()):
			# print >>fh, 'Diagnostic:', desc[k]
			print >>fh, desc[k]
			for sid in sorted(stats[k].keys()):
				info = None
				# if k == 'Davg' or k == 'sigDavg':
					# info = str(round(stats[k][sid],3))
				if k == 'alpha' or k == 'Davg':
					tags = ('average', 'sd', 'quantiles')
					info = [str(roundna(3)(stats[k][sid][0])), str(roundna(3)(stats[k][sid][1])), stats[k][sid][2]]
					info = ', '.join([tags[i]+': '+info[i] for i in range(len(info))])
					print >>fh, '\t'.join([sampleID, k, info])
				else:
					info = ', '.join(map(lambda x: ':'.join(map(str, x)), sorted(stats[k][sid].items())))
					print >>fh, '\t'.join([sampleID, k, info])
					
			print >> fh
		
		print >> fh, 'Statistics for significant SNPs'
		print >> fh
		tags = ('average', 'sd', 'quantiles')
		for P in map(str,QvalueRange):
			for k in sorted(Pstats[P].keys()):
				info = None
				if k == 'sigalpha' or k == 'sigDavg':
					info = [str(roundna(3)(Pstats[P][k][sid][0])), str(roundna(3)(Pstats[P][k][sid][1])), Pstats[P][k][sid][2]]
					info = ', '.join([tags[i]+': '+info[i] for i in range(len(info))])
					# print >>fh, '\t'.join([sampleID, str(P), k, info])
				elif k == 'Qs' or k=='sigG' or k=='Gtype':
					info = ', '.join(map(lambda x: ':'.join(map(str, x)), sorted(Pstats[P][k][sid].items())))
				print >>fh, '\t'.join([sampleID, str(P), k, info])
			print >> fh
		
		fh.close()
		
		print 'DONE DIAGNOSTICS.\n'
	

# ================
# Part V: ASSEMBLY
# ================
if SAVEASSEMBLY:
	print 'PART V: Building new genome assembly.'
	
	createdir(phylodir); createdir(findir)
	
	sortedSampleIDs = sorted(seqRuns.keys())
	
	# master copy
	# G = readFasta(bfaref, TOUPPER=True)
	
	if 1:
		# create a new assembly sequence including novel significant SNPs
		for sampleID in sortedSampleIDs:
			
			checkups = []
			if ADAPTIVEQ:
				# multiple-test corrected snp variants
				snpfile = snpdir+sampleID+boundslabel+'.aQ%s,%s.txt'%(ADAPTIVEQ,int(readlength))
				assfile = findir+sampleID+boundslabel+'.aQ%s,%s.fasta'%(ADAPTIVEQ,int(readlength))
				checkups += [(ADAPTIVEQ,snpfile,assfile)]
			else:
				for minPvalue in QvalueRange:
					# multiple-test corrected snp variants
					snpfile = snpdir+sampleID+boundslabel+'.Q%s.txt'%(minPvalue)
					assfile = findir+sampleID+boundslabel+'.Q%s.fasta'%(minPvalue)
					checkups += [(minPvalue,snpfile,assfile)]
			checkups.sort()
			
			
			for minPvalue,snpfile,assfile in checkups:
				# snpfile = snpdir+sampleID+boundslabel+'.Q%s.txt'%(minPvalue)
				if ADAPTIVEQ:
					pipeit('- Saving new assembly genome %s at FWER < %s'%(sampleID,ADAPTIVEQ),1)
				else:
					pipeit('- Saving new assembly genome %s at Q > %s'%(sampleID,minPvalue),1)
					
				# load reference genome - G will be editied by new SNPs
				Gtmp = readFasta(bfaref, TOUPPER=True)
				
				# load significant SNPs
				# by right, this should be processed line by line
				if os.access(snpfile, os.F_OK):
					dersnps,foo,foo = readTable(snpfile, rownames=False, header=False, comments=False)
					dct = {}
					for i in range(len(dersnps)):
						# row = dersnps[i][0]
						if dersnps[i][0][0] != '#':
							chrom = dersnps[i][0]
							pos = int(dersnps[i][1])
							refsnp = dersnps[i][2]
							cons = dersnps[i][3]
							allele = nts2iupac(list(cons)).upper()
							try: dct[chrom] += [(pos,allele)]
							except KeyError: dct[chrom] = [(pos,allele)]
						
					# edit reference genome using these SNPs
					for chrom in dct.keys():
						seq = list(Gtmp[chrom]['seq'])
					
						# change significant SNP positions
						try:
							for pos,allele in dct[chrom]: seq[pos] = allele
						except IndexError:
							print 'Can\'t fetch position: %s,%s | seq length is only %s'%(chrom,pos,len(seq))
						
						Gtmp[chrom]['seq'] = ''.join(seq)
					printFasta(Gtmp, assfile)
				else:
					print '   - Cannot access snp file:', snpfile
					
		
		
	if ADAPTIVEQ: QvalueRange = [ADAPTIVEQ]
	
	for minPvalue in QvalueRange:
		cpd = {} # chrom,pos dict: keyed by chrom,pos pairs returning the ref snp at that position
		cpdhits = {}
		
		# record positions for all SNPs across all samples
		for sampleID in sortedSampleIDs:
			if ADAPTIVEQ:
				print ' - Generating SNP string...', sampleID, 'at FWER <', minPvalue
				snpfile = snpdir+sampleID+boundslabel+'.aQ%s,%s.txt'%(ADAPTIVEQ,int(readlength))
			else:
				print ' - Generating SNP string...', sampleID, 'at Q >', minPvalue
				snpfile = snpdir+sampleID+boundslabel+'.Q%s.txt'%(minPvalue)
			
			if os.access(snpfile, os.F_OK):
				dersnps,foo,foo = readTable(snpfile, rownames=False, header=False, comments=False)
				for i in xrange(len(dersnps)):
					if dersnps[i][0][0] != '#':
						chrom = dersnps[i][0]
						pos = int(dersnps[i][1])
						refsnp = dersnps[i][2].upper()
						cpd[(chrom,pos)] = refsnp # store this as a position of interest
						
						# record number of lines with a SNP here
						try: cpdhits[(chrom,pos)] += 1
						except KeyError: cpdhits[(chrom,pos)] = 1
						
		# now we have all the positions that are hit by some line
		# save a concatenated snp string
		skeys = cpd.keys(); skeys.sort() # sorting same for all taxa
		# skeys = [(chrom,pos) for (chrom,pos) in sorted(cpd.keys())] # sorting same for all taxa
		
		snpstrings = []
		keepsids = [] # final list of sampleIDs (confirmed SNP file access)
		
		# first add the reference string
		snpstrings += [''.join(map(lambda k: cpd[k], skeys))] # cpd points to reference allele
		keepsids += ['Reference']
		
		# add alleles from a second external file
		if specialReferenceFiles:
			for srf in specialReferenceFiles.split(','):
				G2 = readFasta(srf, TOUPPER=True)
				snpstrings += [''.join([G2[chrom]['seq'][pos] for (chrom,pos) in skeys])]
				srname = srf.split('/')[-1].split('.')[0]
				keepsids += [srname]
			del G2
		
		for sampleID in sortedSampleIDs:
			snpfile = snpdir+sampleID+boundslabel+'.Q%s.txt'%(minPvalue)
			if ADAPTIVEQ: snpfile = snpdir+sampleID+boundslabel+'.aQ%s,%s.txt'%(ADAPTIVEQ,int(readlength))
				
			if os.access(snpfile, os.F_OK):
				keepsids += [sampleID]
				
				dersnps,foo,foo = readTable(snpfile, rownames=False, header=False, comments=False)
				tmp = {} # record SNP positions for this sample
				for i in xrange(len(dersnps)):
					if dersnps[i][0][0] != '#':
						chrom = dersnps[i][0]
						pos = int(dersnps[i][1])
						dersnp = nts2iupac(list(dersnps[i][3])).upper()
						tmp[(chrom,pos)] = dersnp
						
				snpstr = ''
				for cp in skeys: # recall, this is list of ALL SNP positions across samples
					refsnp = cpd[cp]
					# if derived snp in this line, use it, otherwise use the reference
					if cp in tmp: snpstr += tmp[cp]
					else: snpstr += refsnp
				snpstrings += [snpstr]
				
				snpstringfile = phylodir+sampleID+boundslabel+'.SNPstring.Q%s.txt'%(minPvalue)
				if ADAPTIVEQ: snpstringfile = phylodir+sampleID+boundslabel+'.SNPstring.aQ%s,%s.txt'%(ADAPTIVEQ,int(readlength))
				fh = open(snpstringfile, 'w')
				print >> fh, snpstr
				fh.close()
				
		# now save a multi-fasta and phylip file at this stringency level
		snpfasta = list2fasta(snpstrings, ids=keepsids)
		if ADAPTIVEQ:
			printFasta(snpfasta, file=phylodir+'SNPstring.aQ%s,%s.fasta'%(ADAPTIVEQ,int(readlength)))
			fasta2Phylip(snpfasta, file=phylodir+'SNPstring.aQ%s,%s.phylip.txt'%(ADAPTIVEQ,int(readlength)), infile='', verbose=False)
		else:
			printFasta(snpfasta, file=phylodir+'SNPstring.Q%s.fasta'%(minPvalue))
			fasta2Phylip(snpfasta, file=phylodir+'SNPstring.Q%s.phylip.txt'%(minPvalue), infile='', verbose=False)
	print 'DONE PART V: BUILDING NEW GENOME ASSEMBLY.\n'
print '\nDONE SNIPER.PY'

